/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include "DeviceOrientation.hpp"
#include "StringResource.hpp"
#include <stdio.h>
#include <string.h>

/* Function prototypes...
*/
static int    omniNameToID          (char             *pszId);
static int    standardNameToID      (char             *pszId);

#define STRING_OF(x) (apszOrientationNames[x - DeviceOrientation::ORIENTATION_UNLISTED])

static char *apszOrientationNames[] = {
   "ORIENTATION_UNLISTED",
   "ORIENTATION_NONE",
   "ORIENTATION_PORTRAIT",
   "ORIENTATION_LANDSCAPE",
   "ORIENTATION_REVERSE_PORTRAIT",
   "ORIENTATION_REVERSE_LANDSCAPE"
};

#ifdef USE_STANDARD_NAMES

static char *apszStandardOrientationNames[] = {
   "unlisted",
   "none",
   "portrait",
   "landscape",
   "reverse-portrait",
   "reverse-landscape"
};

#endif

DeviceOrientation::
DeviceOrientation (Device *pDevice,
                   int     id)
{
   pDevice_d = pDevice;
   id_d      = id;
}

DeviceOrientation::
~DeviceOrientation ()
{
   pDevice_d = 0;
   id_d      = 0;
}

char * DeviceOrientation::
getName (Device *pDevice, int id)
{
   return StringResource::getString (pDevice->getLanguageResource (),
                                     StringResource::STRINGGROUP_ORIENTATIONS,
                                     id);
}

char * DeviceOrientation::
getName ()
{
   return StringResource::getString (pDevice_d->getLanguageResource (),
                                     StringResource::STRINGGROUP_ORIENTATIONS,
                                     id_d);
}

bool DeviceOrientation::
isID (int id)
{
   return id == id_d;
}

int DeviceOrientation::
getID ()
{
   return id_d;
}

char * DeviceOrientation::
getDeviceID ()
{
   return 0;
}

DeviceOrientation * DeviceOrientation::
create (char *pszId)
{
   return this->createV (nameToID (pszId));
}

int DeviceOrientation::
nameToID (char *pszId)
{
   if (  !pszId
      || !*pszId
      )
   {
      return ORIENTATION_UNLISTED;
   }

   int iOrientationID = omniNameToID (pszId);

   if (ORIENTATION_UNLISTED != iOrientationID)
      return iOrientationID;

   iOrientationID = standardNameToID (pszId);

   return iOrientationID;
}

int
omniNameToID (char *pszId)
{
   static short int asiEntries[] = {
      DeviceOrientation::ORIENTATION_LANDSCAPE,
      DeviceOrientation::ORIENTATION_NONE,
      DeviceOrientation::ORIENTATION_PORTRAIT,
      DeviceOrientation::ORIENTATION_REVERSE_LANDSCAPE,
      DeviceOrientation::ORIENTATION_REVERSE_PORTRAIT
   };

   int iLow  = 0;
   int iMid  = (int)dimof (asiEntries) / 2;
   int iHigh = (int)dimof (asiEntries) - 1;
   int iResult;

   while (iLow <= iHigh)
   {
      iResult = strcmp (pszId, STRING_OF (asiEntries[iMid]));

      if (0 == iResult)
      {
         return asiEntries[iMid];
      }
      else if (0 > iResult)
      {
         // str1 < str2
         iHigh = iMid - 1;
         iMid  = iLow + (iHigh - iLow) / 2;
      }
      else // (0 < iResult)
      {
         // str1 > str2
         iLow  = iMid + 1;
         iMid  = iLow + (iHigh - iLow) / 2;
      }
   }

   return DeviceOrientation::ORIENTATION_UNLISTED;
}

typedef struct _OrientationMap {
   int   iOmniId;
   char *pszStandardName;
} ORIENTATIONMAP, *PORIENTATIONMAP;

static ORIENTATIONMAP vaStandardOrientationMapping[] = {
   { DeviceOrientation::ORIENTATION_LANDSCAPE,         "landscape"         },
   { DeviceOrientation::ORIENTATION_NONE,              "none"              },
   { DeviceOrientation::ORIENTATION_PORTRAIT,          "portrait"          },
   { DeviceOrientation::ORIENTATION_REVERSE_LANDSCAPE, "reverse_landscape" },
   { DeviceOrientation::ORIENTATION_REVERSE_PORTRAIT,  "reverse_portrait"  },
   { DeviceOrientation::ORIENTATION_UNLISTED,          "unlisted"          }
};

int
standardNameToID (char *pszId)
{
   int   iLow    = 0;
   int   iMid    = (int)dimof (vaStandardOrientationMapping) / 2;
   int   iHigh   = (int)dimof (vaStandardOrientationMapping) - 1;
   int   iResult;

   while (iLow <= iHigh)
   {
      iResult = strcmp (pszId, vaStandardOrientationMapping[iMid].pszStandardName);

      if (0 == iResult)
      {
         return vaStandardOrientationMapping[iMid].iOmniId;
      }
      else if (0 > iResult)
      {
         // str1 < str2
         iHigh = iMid - 1;
         iMid  = iLow + (iHigh - iLow) / 2;
      }
      else // (0 < iResult)
      {
         // str1 > str2
         iLow  = iMid + 1;
         iMid  = iLow + (iHigh - iLow) / 2;
      }
   }

   return DeviceOrientation::ORIENTATION_UNLISTED;
}

char * DeviceOrientation::
IDToName (int id)
{
   id -= ORIENTATION_UNLISTED;

   if (  0 <= id
      && id < (int)dimof (apszOrientationNames)
      )
   {
#ifdef USE_STANDARD_NAMES
      return apszStandardOrientationNames[id];
#else
      return apszOrientationNames[id];
#endif
   }
   else
   {
      static char achUnknown[21];

      sprintf (achUnknown, "Unknown (%d)", id + ORIENTATION_PORTRAIT);

      return achUnknown;
   }
}

string DeviceOrientation::
toString (ostringstream& oss)
{
   oss << "{"
       << STRING_OF (id_d)
       << "}";

   return oss.str ();
}

/* Provide a way to print out class data
*/
ostream&
operator<< (ostream& os, const DeviceOrientation& const_self)
{
   DeviceOrientation& self = const_cast<DeviceOrientation&>(const_self);
   ostringstream      oss;

   os << self.toString (oss);

   return os;
}
