/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include <stdio.h>
#include <stdlib.h>
#include <sstream>
#include <fstream>

#include "Device.hpp"
#include "Omni.hpp"
#include "omni2foo.hpp"

#include <sys/stat.h>
#include <sys/types.h>
#include <glib.h>
#include <gmodule.h>

string *
getCleanName (string *rawName, int operation)
{
   // Clean up Foomatic names to get valid file names/Foomatic IDs or
   // the correct make/model pair
   //
   // operation = 0: Return string suitable as Foomatic ID/XML file name
   // Mangle the device name to be Make-Model_Name
   // First space (after manufacturer's name) needs to be a -, and then
   // all spaces after that need to be a _
   // Ex:  Canon-BJC-2000, Epson-Stylus_Color_740
   //
   // operation = 1: Return Make and Model, separated by a "|", names
   // cleaned up from certain weirdnesses
   string device = *rawName;

   // Find the first space, ie, the gap between the make and model
   // and in case of operation 0 replace the " " with a "-", in case
   // of operation 1 with a "|"
   string::size_type pos = device.find_first_of(" ",0);

   if (operation == 0)
   {
      device.replace (pos,1,"-");
   }
   else
   {
      device.replace (pos,1,"|");
   }

   // The first word of the device name is not necessarily the
   // manufacturer's name, so determine the manufacturer here and
   // also do a clean-up of the model name
   string::size_type idx;

   if (device.find ("OkiP") == 0)
   {
      string make = "Okidata";

      // Replace the old Make/Model separator by a space and
      // build the output string according to the operation
      if (operation == 0)
      {
         if ((idx = device.find ("-")) != string::npos)
         {
            device.replace (idx, 1, " ");
         }

         device = make + "-" + device;
      }
      else
      {
         if ((idx = device.find ("|")) != string::npos)
         {
            device.replace (idx, 1, " ");
         }

         device = make + "|" + device;
      }
   }
   else if (device.find ("HP") == 0)
   {
      // Replace roman numbers in model names of HP LaserJet printers
      if ((idx = device.find ("III")) != string::npos)
      {
         device.replace (idx, 3, "3");
      }
      else if ((idx = device.find ("II")) != string::npos)
      {
         device.replace (idx, 2, "2");
      }
   }
   else if (device.find ("KS") == 0)
   {
      // "KS" is not really a manufacturer, but the name of a subdriver
      // of Omni for printers of several manufacturers.
      // Remove "KS "
      string s     = device.substr (3);
      string make;
      string model;

      if (  (  (s[0] >= '0')
            && (s[0] <= '9')
            )
         || (s.find ("KS") == 0)
         )
      {
         // Generic printers
         make = "Generic";

         if (s.find ("KS") == 0)
         {
            model = s.substr (3);
         }
         else
         {
            model = s;
         }

         if (model.find ("Generic") == 0)
         {
            model = model.substr (8);
         }
      }
      else if (s.find ("BJ") == 0)
      {
         // This is probably not the Canon BJ-230 because the "Canon"
         // subdriver already supports the Canon BJ-230
         make = "Others";
         model = s;
      }
      else if (s.find ("Gold Star") == 0)
      {
         make = "GoldStar";
         model = s.substr (10);
      }
      else if (  (s.find ("HDMF") == 0)
              || (s.find ("VP-") == 0)
              || (s.find ("LG ") == 0)
              )
      {
         make = "Others";
         model = s;
      }
      else
      {
         string::size_type idx = s.find_first_of (' ',0);

         make = s.substr (0, idx);
         make[0] = toupper (make[0]);

         for (unsigned int i = 1;
              i < make.size ();
              i ++)
         {
            make[i] = tolower (make[i]);
         }

         model = s.substr (idx+1);
      }

      if (operation == 0)
      {
         device = make + "-" + model;
      }
      else
      {
         device = make + "|" + model;
      }
   }

   // Remove trailing spaces
   while (device[device.size () - 1] == ' ')
   {
      device = device.substr (0, device.size () - 1);
   }

   if (operation == 0)
   {
      // Replace all slashes in the model name (breaks file names)
      while ((idx = device.find ("/")) != string::npos)
      {
         device.replace (idx, 1, " ");
      }

      // Replace "plus" by "+"
      while ((idx = device.find ("plus")) != string::npos)
      {
         device.replace (idx, 4, "+");
      }

      while ((idx = device.find ("Plus")) != string::npos)
      {
         device.replace (idx, 4, "+");
      }

      while ((idx = device.find ("PLUS")) != string::npos)
      {
         device.replace (idx, 4, "+");
      }

      while ((idx = device.find (" +")) != string::npos)
      {
         device.replace (idx, 2, "+");
      }

      // Remove "(", ")", ",", and " "
      // thanks to jason@alteredminds.com for pointing out a bug
      while ((idx = device.find_first_of ("(), ")) != string::npos)
      {
         device.replace (idx, 1, "_");
      }
   }

   return new string (device);
}

string *
getDeviceName (Device *pDevice, omni2foo* db, int operation)
{
   // operation = 0: Return Foomatic ID (also used as the XML file name)
   // operation = 1: Return Make and Model (for printer's DB entry, only
   //                if Foomatic ID does not match)
   //
   //                If a Foomatic ID of the printer exists, no printer
   //                XML file will be created because Foomatic provides
   //                already one (empty string returned in operation 1).
   //
   //                If a printer entry does not exist, Omni must provide
   //                one, here for the manufacturer and model entry the
   //                returned string of operation 1 is used. The returned
   //                string of operation 0 provides the Foomatic ID and
   //                the file name.
   ostringstream oss;
   string        devName = pDevice->getDeviceName ();

   // check the db
   string cleanname = *getCleanName (&devName, 0);
   string id        = db->getFooIDFromString (cleanname);

   // if there is no match, the string that is returned is empty
   if (id.length () > 0)
   {
      if (operation == 0)
      {
         // operation = 0:
         // found a match, return the fooid string
         return new string (id);
      }
      else
      {
         // operation = 1:
         // match, return empty string
         return new string ("");
      }
   }
   else
   {
      // operation = 0:
      // no match, return Foomatic-compliant device name
      // Mangle the device name to be Make-Model_Name
      // First space (after manufacturer's name) needs to be a -, and then
      // all spaces after that need to be a _
      // Ex:  Canon-BJC-2000, Epson-Stylus_Color_740
      //
      // operation = 1:
      // Return cleaned up make and model names

      string device = *getCleanName (&devName, operation);

      return new string (device);
   }
}

// Derive command-line suitable shortnames out of the longnames. Most of
// them are also equal to the shortnames used by the Foomatic options of
// other drivers (as GIMP-Print, ljet4, ...)

// Remove spaces and '#' out of a C-style string, make letters after
// former spaces capital (as long as the character before the space is not a
// number)

const char *
generateShortName (const char *str)
{
   // Allocate memory of same length of original string
   char *result;

   result = (char *)malloc (sizeof (char)*(strlen (str)+1));

   int j          = 0;
   int afterspace = 0;
   int lastnumber = 0;

   for (unsigned int i = 0; i < strlen (str); i++)
   {
      if (  (str[i] != ' ')
         && (str[i] != '#')
         )
      {
         if (afterspace)
         {
            result[j] = toupper (str[i]);
            afterspace = 0;
         }
         else
         {
            result[j] = str[i];
         }

         if (  (str[i] >= '0')
            && (str[i] <= '9')
            )
         {
            lastnumber = 1;
         }
         else
         {
            lastnumber = 0;
         }

         j++;
      }
      else
      {
         if (  (str[i] == ' ')
            && (!lastnumber)
            )
         {
            afterspace = 1;
         }
      }
   }

   result[j] = 0;

   return result;
}

int
main (int argc, char *argv[])
{
   string basedir    = "";
   string foomaticDB = "bin/newDB";
   bool   fUseBuild  = false;

   // Command Line parsing.  2 options.
   // -b : base dir, the location where the foomatic-db dir will be made
   // -d : the foomatic id db.  it is a space separated file, specify where that file is, or use the default
   // -l : use the build root
   for (int i = 1; i < argc; i++)
   {
      if (argv[i][0] == '-')
      {
         switch (argv[i][1])
         {
         case 'D':
         case 'd':
         {
            foomaticDB = argv[i + 1];
            break;
         }

         case 'b':
         case 'B':
         {
            basedir = argv[i + 1];
            break;
         }

         case 'l':
         case 'L':
         {
            fUseBuild = true;
            break;
         }
         }
      }
   }

   // Check basedir arg for trailing slash, if not present, add
   if (  basedir.length () > 0
      && basedir[basedir.length ()-1] != '/'
      )
      basedir = basedir + "/";

   typedef struct _OmniDeviceHandle {
      GModule *hmodDevice;
      Device  *pDevice;
      string  *pstringDeviceName;
      string  *pstringMakeModel;
   } OMNIHANDLE, *POMNIHANDLE;

   POMNIHANDLE  pohDevices = 0;
   ofstream    *pofstream;

   if (!g_module_supported ())
   {
      cerr << "This program needs glib's module routines!" << endl;

      return 1;
   }

   // Now make the omni2foo DB
   omni2foo::omni2foo DB (foomaticDB);

   mode_t mode      = S_IRUSR | S_IWUSR | S_IXUSR | S_IRGRP | S_IXGRP;
   string directory;

   directory = basedir + "foomatic-db";
   mkdir (directory.c_str (), mode);

   directory = basedir + "foomatic-db/db";
   mkdir (directory.c_str (), mode);

   directory = basedir + "foomatic-db/db/source";
   mkdir (directory.c_str (), mode);

   directory = basedir + "foomatic-db/db/source/driver";
   mkdir (directory.c_str (), mode);

   directory = basedir + "foomatic-db/db/source/printer";
   mkdir (directory.c_str (), mode);

   directory = basedir + "foomatic-db/db/source/opt";
   mkdir (directory.c_str (), mode);

   cout << "Counting devices..." << endl;

   Enumeration *pEnum          = Omni::listDevices (fUseBuild);
   int          iNumDevices    = 0;
   int          iCurrentDevice = 0;

   ofstream ofstreamOmni ("foomatic-db/db/source/driver/omni.xml");
   ofstream ofstreamPrintModel ("foomatic-db/db/source/opt/omni-print-model.xml");
   ofstream ofstreamForm ("foomatic-db/db/source/opt/omni-form.xml");
   ofstream ofstreamTray ("foomatic-db/db/source/opt/omni-tray.xml");
   ofstream ofstreamInputSlots ("foomatic-db/db/source/opt/omni-media.xml");
   ofstream ofstreamResolutions ("foomatic-db/db/source/opt/omni-resolution.xml");
   ofstream ofstreamPrintModes ("foomatic-db/db/source/opt/omni-printmode.xml");
   ofstream ofstreamDithers ("foomatic-db/db/source/opt/omni-dither.xml");
   ofstream ofstreamOrientation ("foomatic-db/db/source/opt/omni-orientation.xml");
   ofstream ofstreamMonoDither ("foomatic-db/db/source/opt/omni-monodither.xml");

   while (pEnum->hasMoreElements ())
   {
      pEnum->nextElement ();

      iNumDevices++;
   }

   delete pEnum;

   cout << "There are " << iNumDevices << " devices." << endl;

   if (!iNumDevices)
   {
      goto done;
   }

   cout << "Initializing devices...";
   cout.flush ();

   pohDevices = new OMNIHANDLE [iNumDevices];
   pEnum      = Omni::listDevices (fUseBuild);

   while (pEnum->hasMoreElements ())
   {
      char         *pszLibName   = (char *)pEnum->nextElement ();
      PFNNEWDEVICE  pfnNewDevice = 0;

      pohDevices[iCurrentDevice].hmodDevice = g_module_open (pszLibName, (GModuleFlags)0);

      if (pohDevices[iCurrentDevice].hmodDevice)
      {
         g_module_symbol (pohDevices[iCurrentDevice].hmodDevice, "newDeviceW_Advanced", (gpointer *)&pfnNewDevice);

         if (pfnNewDevice)
         {
            Device *pDevice = 0;

            pDevice                                      = pfnNewDevice (true);
            pohDevices[iCurrentDevice].pDevice           = pDevice;
            pohDevices[iCurrentDevice].pstringDeviceName = getDeviceName (pDevice, &DB, 0);
            pohDevices[iCurrentDevice].pstringMakeModel  = getDeviceName (pDevice, &DB, 1);
         }
      }
      else
      {
         pohDevices[iCurrentDevice].pDevice           = 0;
         pohDevices[iCurrentDevice].pstringDeviceName = 0;
      }

      iCurrentDevice++;
   }

   delete pEnum;

   cout << "done." << endl;

   for (iCurrentDevice = 0; iCurrentDevice < iNumDevices; iCurrentDevice++)
   {
      Device   *pDevice           = pohDevices[iCurrentDevice].pDevice;
      string   *pstringDeviceName = pohDevices[iCurrentDevice].pstringDeviceName;
      string   *pstringMakeModel  = pohDevices[iCurrentDevice].pstringMakeModel;
      char      achFilename[128];
      ofstream  ofstreamPrinter (achFilename);

      // Produce only printer XML files which Foomatic does not already
      // provide.

      if (pstringMakeModel->empty ())
         continue;

      sprintf (achFilename, "foomatic-db/db/source/printer/%s.xml", pstringDeviceName->c_str ());

      pofstream = &ofstreamPrinter;

      sprintf (achFilename, "printer/%s.xml", pstringDeviceName->c_str ());
      cout << "Generating " << achFilename << endl;

      *pofstream << "<printer id=\"printer/" << *pstringDeviceName << "\">" << endl;

      string::size_type idx   = pstringMakeModel->find_first_of ('|',0);
      string            make  = pstringMakeModel->substr (0, idx);
      string            model = pstringMakeModel->substr (idx+1);

      *pofstream << "  <make>" << make << "</make>" << endl;
      *pofstream << "  <model>" << model << "</model>" << endl;
      *pofstream << "  <!-- " << pDevice->getShortName () << " -->" << endl;
      *pofstream << "  <mechanism>" << endl;
      *pofstream << "    <inkjet/>" << endl;
      *pofstream << "    <color/>" << endl;
      *pofstream << "    <resolution>" << endl;
      *pofstream << "      <dpi>" << endl;
      //  Added by raharper@us.ibm.com, get the resolution for each device
      *pofstream << "        <x>" << pDevice->getCurrentResolution ()->getXRes () << "</x>" << endl;
      *pofstream << "        <y>" << pDevice->getCurrentResolution ()->getYRes () << "</y>" << endl;
      *pofstream << "      </dpi>" << endl;
      *pofstream << "    </resolution>" << endl;
      *pofstream << "    <consumables>" << endl;
      //  The <consumables> section can be left blank if the consumables
      //  for this printer are not known.
      *pofstream << "      <!-- no \"comments\" -->" << endl;
      *pofstream << "    <!--one or more \"partno\" elements.--></consumables>" << endl;
      *pofstream << "  </mechanism>" << endl;
      //  Added by raharper@us.ibm.com, changed URL
      *pofstream << "  <url>http://www.sourceforge.net/projects/omniprint</url>" << endl;
      *pofstream << "  <lang>" << endl;
      *pofstream << "    <text>" << endl;
      *pofstream << "      <charset>us-ascii</charset>" << endl;
      *pofstream << "    </text>" << endl;
      *pofstream << "  </lang>" << endl;
      *pofstream << "  <autodetect>" << endl;
      *pofstream << "    <!--no known parport probe information-->" << endl;
      *pofstream << "  </autodetect>" << endl;
      *pofstream << "  <functionality>D</functionality><!--unknown preferred \"driver\"-->" << endl;
      *pofstream << "  <unverified/><!--no \"contrib_url\"-->" << endl;
      *pofstream << "  <comments>" << endl;
      *pofstream << "    <en>" << endl;
      *pofstream << "    Foomatic entry automatically generated by Omni.&lt;p&gt;" << endl;
      *pofstream << "    Printer type, resolution, and printer language information" << endl;
      *pofstream << "    may be incorrect.&lt;p&gt;" << endl;
      *pofstream << "    </en>" << endl;
      *pofstream << "  </comments>" << endl;
      *pofstream << "</printer>" << endl;
   }

   /**
    * Create Omni
    *
    * list of device names
    */
   pofstream = &ofstreamOmni;

   cout << "Generating omni.xml" << endl;

   *pofstream << "<driver id=\"driver/omni\">" << endl;
   *pofstream << " <name>omni-print</name>" << endl;
   *pofstream << " <url>http://www.ibm.com/linux/ltc/projects/omni/</url>" << endl;
   *pofstream << " <execution>" << endl;
   *pofstream << "  <ghostscript />" << endl;
   // Edited by raharper@us.ibm.com, changed the prototype to what omni wants it to look like
   *pofstream << "  <prototype>gs -q -dSAFER -dBATCH -dNOPAUSE -sOutputFile=- -sDEVICE=omni%B -sproperties=&quot;%A%Z%C -f -</prototype>" << endl; // @TBD
   *pofstream << " </execution>" << endl;
   *pofstream << " <comments>" << endl;
   *pofstream << "  <en>" << endl;
   *pofstream << "&lt;b&gt;";
   *pofstream << "This page provides data for the Omni driver version ";
   if (0 < iNumDevices)
   {
      Device *pDevice = pohDevices[0].pDevice;

      *pofstream << pDevice->getVersion ();
   }
   else
   {
      *pofstream << "Unknown";
   }
   *pofstream << "." << endl;
   *pofstream << "&lt;b&gt;";
   *pofstream << "The Omni printer driver provides support for over 400 printers " << endl;
   *pofstream << "using the Ghostscript framework. In addition, it provides a " << endl;
   *pofstream << "model for dynamically loading printer drivers, creating new " << endl;
   *pofstream << "devices by editing device description files, and simplifies " << endl;
   *pofstream << "new printer driver development by allowing for the subclassing " << endl;
   *pofstream << "of previous device features." << endl;
   *pofstream << "  </en>" << endl;
   *pofstream << " </comments>" << endl;
   *pofstream << " <printers>" << endl;

   for (iCurrentDevice = 0; iCurrentDevice < iNumDevices; iCurrentDevice++)
   {
      string *pstringDeviceName = pohDevices[iCurrentDevice].pstringDeviceName;

      *pofstream << "  <printer><id>printer/" << *pstringDeviceName << "</id></printer>" << endl;
   }

   *pofstream << " </printers>" << endl;
   *pofstream << "</driver>" << endl;

   /**
    * Create Print Model
    *
    * list of device names
    */
   pofstream = &ofstreamPrintModel;

   cout << "Generating omni-print-model.xml" << endl;

   *pofstream << "<option type=\"enum\" id=\"opt/omni-print-model\">" << endl;
   *pofstream << "  <!-- A multilingual <comments> block can appear here, too;" << endl;
   *pofstream << "       it should be treated as documentation for the user. -->" << endl;
   *pofstream << "  <arg_longname>" << endl;
   *pofstream << "   <en>Printer Model</en>" << endl;
   *pofstream << "  </arg_longname>" << endl;
   *pofstream << "  <arg_shortname>" << endl;
   *pofstream << "   <en>Model</en><!-- backends only know <en> shortnames! -->" << endl;
   *pofstream << "  </arg_shortname>" << endl;
   *pofstream << "  <arg_execution>" << endl;
   *pofstream << "   <arg_order>10</arg_order>" << endl;
   *pofstream << "   <arg_spot>B</arg_spot>" << endl;
   *pofstream << "   <arg_required />" << endl;
   *pofstream << "   <arg_substitution />" << endl;
   *pofstream << "   <arg_proto> -sDeviceName=%s</arg_proto>" << endl;
   *pofstream << "  </arg_execution>" << endl;
   *pofstream << "" << endl;
   *pofstream << "  <constraints>" << endl;

   for (iCurrentDevice = 0; iCurrentDevice < iNumDevices; iCurrentDevice++)
   {
      string *pstringDeviceName = pohDevices[iCurrentDevice].pstringDeviceName;

      *pofstream << "    <constraint sense='true'>" << endl;
      *pofstream << "      <driver>omni</driver>" << endl;
      *pofstream << "      <printer>printer/" << *pstringDeviceName << "</printer>" << endl;
      *pofstream << "      <arg_defval>" << *pstringDeviceName << "</arg_defval>" << endl; // @TBD
      *pofstream << "    </constraint>" << endl;
   }

   *pofstream << "  </constraints>" << endl;
   *pofstream << endl;
   *pofstream << "  <enum_vals>" << endl;

   for (iCurrentDevice = 0; iCurrentDevice < iNumDevices; iCurrentDevice++)
   {
      Device *pDevice           = pohDevices[iCurrentDevice].pDevice;
      string *pstringDeviceName = pohDevices[iCurrentDevice].pstringDeviceName;

      //raharper@us.ibm.com took out the ev/omni-
      *pofstream << "    <enum_val id='" << *pstringDeviceName << "'>" << endl;
      *pofstream << "      <ev_longname><en>" << *pstringDeviceName << "</en></ev_longname>" << endl;
      *pofstream << "      <ev_shortname><en>" << pDevice->getShortName () << "</en></ev_shortname>" << endl;
      *pofstream << "      <ev_driverval>" << pDevice->getShortName () << "</ev_driverval>" << endl;
      *pofstream << "      <constraints>" << endl;
      *pofstream << "        <!-- Assume the option doesn't apply... -->" << endl;
      *pofstream << "        <constraint sense='false'>" << endl;
      *pofstream << "          <driver>omni</driver>" << endl;
      *pofstream << "        </constraint>" << endl;
      *pofstream << "        <!-- ...except to these: -->" << endl;
      *pofstream << "        <constraint sense='true'>" << endl;
      *pofstream << "          <driver>omni</driver>" << endl;
      *pofstream << "          <printer>printer/" << *pstringDeviceName << "</printer>" << endl;
      *pofstream << "        </constraint>" << endl;
      *pofstream << "      </constraints>" << endl;
      *pofstream << "    </enum_val>" << endl;
   }

   *pofstream << "  </enum_vals>" << endl;
   *pofstream << "</option>" << endl;

   /**
    * Create Paper Sizes (Forms)
    *
    * list of forms
    */
   // raharper@us.ibm.com, changed filename to form
   pofstream = &ofstreamForm;

   cout << "Generating omni-form.xml" << endl;

   *pofstream << "<option type=\"enum\" id=\"opt/omni-form\">" << endl;
   *pofstream << "  <arg_longname>" << endl;
   // Edited by raharper@us.ibm.com, changing the output to match omni driver options
   *pofstream << "   <en>Page Size</en>" << endl;
   *pofstream << "  </arg_longname>" << endl;
   *pofstream << "  <arg_shortname>" << endl;
   // Edited by raharper@us.ibm.com, changing the output to match omni driver options
   *pofstream << "   <en>PageSize</en><!-- backends only know <en> shortnames! -->" << endl;
   *pofstream << "  </arg_shortname>" << endl;
   *pofstream << "  <arg_execution>" << endl;
   *pofstream << "   <arg_order>100</arg_order>" << endl;
   *pofstream << "   <arg_spot>C</arg_spot>" << endl;
   *pofstream << "   <arg_required />" << endl;
   // raharper@us.ibm.com, changed the arg_xxxx type to subsitutiuon to get the style=C in the perl dat structure generator, db->getdat ()
   *pofstream << "   <arg_substitution />" << endl;
   // Edited by raharper@us.ibm.com, changing the prototype to what omni uses for commandline options
   *pofstream << "   <arg_proto> %s</arg_proto>" << endl;
   *pofstream << "  </arg_execution>" << endl;
   *pofstream << endl;

   *pofstream << "  <constraints>" << endl;

   for (iCurrentDevice = 0; iCurrentDevice < iNumDevices; iCurrentDevice++)
   {
      Device     *pDevice           = pohDevices[iCurrentDevice].pDevice;
      DeviceForm *pDF               = pDevice->getCurrentForm ();
      string     *pstringDeviceName = pohDevices[iCurrentDevice].pstringDeviceName;

      *pofstream << "    <constraint sense='true'>" << endl;
      *pofstream << "      <driver>omni</driver>" << endl;
      *pofstream << "      <printer>printer/" << *pstringDeviceName << "</printer>" << endl;
      // raharper@us.ibm.com, changing the arg_defval output
      *pofstream << "      <arg_defval>" << DeviceForm::IDToName (pDF->getID ()) << "</arg_defval>" << endl;
      *pofstream << "    </constraint>" << endl;
   }

   *pofstream << "  </constraints>" << endl;
   *pofstream << endl;
   *pofstream << "  <enum_vals>" << endl;

   for (int iForm = DeviceForm::FORM_NONE; iForm < DeviceForm::COUNT; iForm++)
   {
      bool fFirstTime = true;

      for (iCurrentDevice = 0; iCurrentDevice < iNumDevices; iCurrentDevice++)
      {
         Device     *pDevice           = pohDevices[iCurrentDevice].pDevice;
         DeviceForm *pDF               = pDevice->getCurrentForm ();
         string     *pstringDeviceName = pohDevices[iCurrentDevice].pstringDeviceName;

         if (pDF->isFormSupported (iForm))
         {
            if (fFirstTime)
            {
               char        *pcharLongName;
               const char  *pcharShortName;
               DeviceForm  *pDFDevice       = 0;
               HardCopyCap *pHCCDevice;
               int          iCx;
               int          iCy;

               fFirstTime = false;

               pDFDevice  = pDF->create (DeviceForm::IDToName (iForm));
               pHCCDevice = pDFDevice->getHardCopyCap ();

               // Generate the short name out of the long name
               pcharLongName  = pDFDevice->getName ();
               pcharShortName = generateShortName (pcharLongName);

               // @TBD user defined form is (0, 0) in size
               iCx = (int)(((double)pHCCDevice->getCx ()) / 25400.0 * 72.0 + 0.5);
               iCy = (int)(((double)pHCCDevice->getCy ()) / 25400.0 * 72.0 + 0.5);

               //raharper@us.ibm.com took out the ev/omni-
               *pofstream << "    <enum_val id='" << DeviceForm::IDToName (iForm) << "'>" << endl;
               *pofstream << "      <ev_longname><en>" << pcharLongName << "</en></ev_longname>" << endl;
               *pofstream << "      <ev_shortname><en>" << pcharShortName << "</en></ev_shortname>" << endl;
               *pofstream << "      <ev_driverval>form=" << DeviceForm::IDToName (iForm) << "&quot; -c &quot;&lt;&lt;/PageSize[" << iCx << " " << iCy << "]/ImagingBBox null&gt;&gt;setpagedevice&quot;</ev_driverval>" << endl;
               *pofstream << "      <constraints>" << endl;
               *pofstream << "        <!-- Assume the option doesn't apply... -->" << endl;
               *pofstream << "        <constraint sense='false'>" << endl;
               *pofstream << "          <driver>omni</driver>" << endl;
               *pofstream << "        </constraint>" << endl;

               free ((void *)pcharShortName);
               delete pDFDevice;
            }

            *pofstream << "        <constraint sense='true'>" << endl;
            *pofstream << "          <driver>omni</driver>" << endl;
            *pofstream << "          <printer>printer/" << *pstringDeviceName << "</printer>" << endl;
            *pofstream << "        </constraint>" << endl;
         }
      }

      if (!fFirstTime)
      {
         *pofstream << "      </constraints>" << endl;
         *pofstream << "    </enum_val>" << endl;
      }
   }

   *pofstream << "  </enum_vals>" << endl;
   *pofstream << "</option>" << endl;

   /**
    * Create Input Slots
    *
    * list of trays
    */
   // raharper@us.ibm.com, changed file name to tray
   pofstream = &ofstreamTray;

   cout << "Generating omni-tray.xml" << endl;

   *pofstream << "<option type=\"enum\" id=\"opt/omni-tray\">" << endl;
   *pofstream << "  <!-- A multilingual <comments> block can appear here, too;" << endl;
   *pofstream << "       it should be treated as documentation for the user. -->" << endl;
   *pofstream << "  <arg_longname>" << endl;
   // Edited by raharper@us.ibm.com, changing the output to match omni driver options
   *pofstream << "   <en>Media Source</en>" << endl;
   *pofstream << "  </arg_longname>" << endl;
   *pofstream << "  <arg_shortname>" << endl;
   // Edited by raharper@us.ibm.com, changing the output to match omni driver options
   *pofstream << "   <en>InputSlot</en><!-- backends only know <en> shortnames! -->" << endl;
   *pofstream << "  </arg_shortname>" << endl;
   *pofstream << "  <arg_execution>" << endl;
   *pofstream << "   <arg_order>100</arg_order>" << endl;
   *pofstream << "   <arg_spot>A</arg_spot>" << endl;
   *pofstream << "   <arg_substitution />" << endl;
   // Edited by raharper@us.ibm.com, changing the output to match omni driver options
   *pofstream << "   <arg_proto> tray=%s</arg_proto>" << endl;
   *pofstream << "  </arg_execution>" << endl;
   *pofstream << endl;

   *pofstream << "  <constraints>" << endl;

   for (iCurrentDevice = 0; iCurrentDevice < iNumDevices; iCurrentDevice++)
   {
      Device     *pDevice           = pohDevices[iCurrentDevice].pDevice;
      DeviceTray *pDT               = pDevice->getCurrentTray ();
      string     *pstringDeviceName = pohDevices[iCurrentDevice].pstringDeviceName;

      *pofstream << "    <constraint sense='true'>" << endl;
      *pofstream << "      <driver>omni</driver>" << endl;
      *pofstream << "      <printer>printer/" << *pstringDeviceName << "</printer>" << endl;
      // raharper@us.ibm.com, changing what the defval output looks like
      *pofstream << "      <arg_defval>" << DeviceTray::IDToName (pDT->getID ()) << "</arg_defval>" << endl;
      *pofstream << "    </constraint>" << endl;
   }

   *pofstream << "  </constraints>" << endl;
   *pofstream << endl;
   *pofstream << "  <enum_vals>" << endl;

   for (int iTray = DeviceTray::TRAY_NONE; iTray < DeviceTray::COUNT; iTray++)
   {
      bool fFirstTime = true;

      for (iCurrentDevice = 0; iCurrentDevice < iNumDevices; iCurrentDevice++)
      {
         Device     *pDevice           = pohDevices[iCurrentDevice].pDevice;
         DeviceTray *pDT               = pDevice->getCurrentTray ();
         string     *pstringDeviceName = pohDevices[iCurrentDevice].pstringDeviceName;

         if (pDT->isTraySupported (iTray))
         {
            if (fFirstTime)
            {
               const char *pcharLongName;
               const char *pcharShortName;

               fFirstTime = false;

               // Generate the short name out of the long name
               pcharLongName  = DeviceTray::getName (pDevice, iTray);
               pcharShortName = generateShortName (pcharLongName);

               //raharper@us.ibm.com took out the ev/omni-
               *pofstream << "    <enum_val id='" << DeviceTray::IDToName (iTray) << "'>" << endl;
               *pofstream << "      <ev_longname><en>" << pcharLongName << "</en></ev_longname>" << endl;
               *pofstream << "      <ev_shortname><en>" << pcharShortName << "</en></ev_shortname>" << endl;
               *pofstream << "      <ev_driverval>" << DeviceTray::IDToName (iTray) << "</ev_driverval>" << endl;
               *pofstream << "      <constraints>" << endl;
               *pofstream << "        <!-- Assume the option doesn't apply... -->" << endl;
               *pofstream << "        <constraint sense='false'>" << endl;
               *pofstream << "          <driver>omni</driver>" << endl;
               *pofstream << "        </constraint>" << endl;

               free ((void *)pcharShortName);
            }

            *pofstream << "        <constraint sense='true'>" << endl;
            *pofstream << "          <driver>omni</driver>" << endl;
            *pofstream << "          <printer>printer/" << *pstringDeviceName << "</printer>" << endl;
            *pofstream << "        </constraint>" << endl;
         }
      }

      if (!fFirstTime)
      {
         *pofstream << "      </constraints>" << endl;
         *pofstream << "    </enum_val>" << endl;
      }
   }

   *pofstream << "  </enum_vals>" << endl;
   *pofstream << "</option>" << endl;

   /**
    * Create MediaTypes
    *
    * list of medias
    */
   // raharper@us.ibm.com, Added Media section
   pofstream = &ofstreamInputSlots;

   cout << "Generating omni-media.xml" << endl;

   *pofstream << "<option type=\"enum\" id=\"opt/omni-media\">" << endl;
   *pofstream << "  <!-- A multilingual <comments> block can appear here, too;" << endl;
   *pofstream << "       it should be treated as documentation for the user. -->" << endl;
   *pofstream << "  <arg_longname>" << endl;
   // Edited by raharper@us.ibm.com, changing the output to match omni driver options
   *pofstream << "   <en>Media Type</en>" << endl;
   *pofstream << "  </arg_longname>" << endl;
   *pofstream << "  <arg_shortname>" << endl;
   // Edited by raharper@us.ibm.com, changing the output to match omni driver options
   *pofstream << "   <en>MediaType</en><!-- backends only know <en> shortnames! -->" << endl;
   *pofstream << "  </arg_shortname>" << endl;
   *pofstream << "  <arg_execution>" << endl;
   *pofstream << "   <arg_order>100</arg_order>" << endl;
   *pofstream << "   <arg_spot>A</arg_spot>" << endl;
   *pofstream << "   <arg_substitution />" << endl;
   // Edited by raharper@us.ibm.com, changing the output to match omni driver options
   *pofstream << "   <arg_proto> media=%s</arg_proto>" << endl;
   *pofstream << "  </arg_execution>" << endl;
   *pofstream << endl;

   *pofstream << "  <constraints>" << endl;

   for (iCurrentDevice = 0; iCurrentDevice < iNumDevices; iCurrentDevice++)
   {
      Device      *pDevice           = pohDevices[iCurrentDevice].pDevice;
      DeviceMedia *pDT               = pDevice->getCurrentMedia ();
      string      *pstringDeviceName = pohDevices[iCurrentDevice].pstringDeviceName;

      *pofstream << "    <constraint sense='true'>" << endl;
      *pofstream << "      <driver>omni</driver>" << endl;
      *pofstream << "      <printer>printer/" << *pstringDeviceName << "</printer>" << endl;
      // raharper@us.ibm.com, changing what the defval output looks like
      *pofstream << "      <arg_defval>" << DeviceMedia::IDToName (pDT->getID ()) << "</arg_defval>" << endl;
      *pofstream << "    </constraint>" << endl;
   }

   *pofstream << "  </constraints>" << endl;
   *pofstream << endl;
   *pofstream << "  <enum_vals>" << endl;

   for (int iMedia = DeviceMedia::MEDIA_NONE; iMedia < DeviceMedia::COUNT; iMedia++)
   {
      bool fFirstTime = true;

      for (iCurrentDevice = 0; iCurrentDevice < iNumDevices; iCurrentDevice++)
      {
         Device      *pDevice           = pohDevices[iCurrentDevice].pDevice;
         DeviceMedia *pDT               = pDevice->getCurrentMedia ();
         string      *pstringDeviceName = pohDevices[iCurrentDevice].pstringDeviceName;

         if (pDT->isMediaSupported (iMedia))
         {
            if (fFirstTime)
            {
               const char *pcharLongName;
               const char *pcharShortName;

               fFirstTime = false;

               // Generate the short name out of the long name
               pcharLongName  = DeviceMedia::getName (pDevice, iMedia);
               pcharShortName = generateShortName (pcharLongName);

               //raharper@us.ibm.com took out the ev/omni-
               *pofstream << "    <enum_val id='" << DeviceMedia::IDToName (iMedia) << "'>" << endl;
               *pofstream << "      <ev_longname><en>" << pcharLongName << "</en></ev_longname>" << endl;
               *pofstream << "      <ev_shortname><en>" << pcharShortName << "</en></ev_shortname>" << endl;
               *pofstream << "      <ev_driverval>" << DeviceMedia::IDToName (iMedia) << "</ev_driverval>" << endl;
               *pofstream << "      <constraints>" << endl;
               *pofstream << "        <!-- Assume the option doesn't apply... -->" << endl;
               *pofstream << "        <constraint sense='false'>" << endl;
               *pofstream << "          <driver>omni</driver>" << endl;
               *pofstream << "        </constraint>" << endl;

               free ((void *)pcharShortName);
            }

            *pofstream << "        <constraint sense='true'>" << endl;
            *pofstream << "          <driver>omni</driver>" << endl;
            *pofstream << "          <printer>printer/" << *pstringDeviceName << "</printer>" << endl;
            *pofstream << "        </constraint>" << endl;
         }
      }

      if (!fFirstTime)
      {
         *pofstream << "      </constraints>" << endl;
         *pofstream << "    </enum_val>" << endl;
      }
   }

   *pofstream << "  </enum_vals>" << endl;
   *pofstream << "</option>" << endl;

   /**
    * Create Resolution
    *
    * list of resolutions
    */
   pofstream = &ofstreamResolutions;

   cout << "Generating omni-resolution.xml" << endl;
   // raharper@us.ibm.com, changed GSResolution to Resolution
   *pofstream << "<option type=\"enum\" id=\"opt/omni-resolution\">" << endl;
   *pofstream << "  <!-- A multilingual <comments> block can appear here, too;" << endl;
   *pofstream << "       it should be treated as documentation for the user. -->" << endl;
   *pofstream << "  <arg_longname>" << endl;
   *pofstream << "   <en>Resolution</en>" << endl;
   *pofstream << "  </arg_longname>" << endl;
   *pofstream << "  <arg_shortname>" << endl;
   *pofstream << "   <en>Resolution</en><!-- backends only know <en> shortnames! -->" << endl;
   *pofstream << "  </arg_shortname>" << endl;
   *pofstream << "  <arg_execution>" << endl;
   *pofstream << "   <arg_order>100</arg_order>" << endl;
   *pofstream << "   <arg_spot>A</arg_spot>" << endl;
   *pofstream << "   <arg_substitution />" << endl;
   // edited by raharper@us.ibm.com, changed output to match omni driver options
   *pofstream << "   <arg_proto> resolution=%s</arg_proto>" << endl;
   *pofstream << "  </arg_execution>" << endl;
   *pofstream << endl;

   *pofstream << "  <constraints>" << endl;

   for (iCurrentDevice = 0; iCurrentDevice < iNumDevices; iCurrentDevice++)
   {
      Device           *pDevice           = pohDevices[iCurrentDevice].pDevice;
      DeviceResolution *pDR               = pDevice->getCurrentResolution ();
      string           *pstringDeviceName = pohDevices[iCurrentDevice].pstringDeviceName;

      *pofstream << "    <constraint sense='true'>" << endl;
      *pofstream << "      <driver>omni</driver>" << endl;
      *pofstream << "      <printer>printer/" << *pstringDeviceName << "</printer>" << endl;
      // raharper@us.ibm.com, changed the defval output
      *pofstream << "      <arg_defval>" << DeviceResolution::IDToName (pDR->getID ()) << "</arg_defval>" << endl;
      *pofstream << "    </constraint>" << endl;
   }

   *pofstream << "  </constraints>" << endl;
   *pofstream << endl;
   *pofstream << "  <enum_vals>" << endl;

   for (int iResolution = DeviceResolution::RESOLUTION_NONE; iResolution < DeviceResolution::COUNT; iResolution++)
   {
      bool fFirstTime = true;

      for (iCurrentDevice = 0; iCurrentDevice < iNumDevices; iCurrentDevice++)
      {
         Device           *pDevice           = pohDevices[iCurrentDevice].pDevice;
         DeviceResolution *pDR               = pDevice->getCurrentResolution ();
         string           *pstringDeviceName = pohDevices[iCurrentDevice].pstringDeviceName;

         if (pDR->isResolutionSupported (iResolution))
         {
            if (fFirstTime)
            {
               char       *pcharLongName;
               const char *pcharShortName;

               fFirstTime = false;

               // Generate the short name out of the long name
               pcharLongName = DeviceResolution::getName (pDevice, iResolution);

               if (pcharLongName == NULL)
               {
                  pcharLongName = strdup (DeviceResolution::IDToName (iResolution));
                  pcharLongName += 11;
                  for (unsigned int i = 1;
                       i < strlen (pcharLongName);
                       i ++)
                  {
                     pcharLongName[i] = tolower (pcharLongName[i]);
                  }
               }

               pcharShortName = generateShortName (pcharLongName);

               //raharper@us.ibm.com took out the ev/omni-
               *pofstream << "    <enum_val id='" << DeviceResolution::IDToName (iResolution) << "'>" << endl;
               *pofstream << "      <ev_longname><en>" << pcharLongName << "</en></ev_longname>" << endl;
               *pofstream << "      <ev_shortname><en>" << pcharShortName << "</en></ev_shortname>" << endl;
               *pofstream << "      <ev_driverval>" << DeviceResolution::IDToName (iResolution) << "</ev_driverval>" << endl;
               *pofstream << "      <constraints>" << endl;
               *pofstream << "        <!-- Assume the option doesn't apply... -->" << endl;
               *pofstream << "        <constraint sense='false'>" << endl;
               *pofstream << "          <driver>omni</driver>" << endl;
               *pofstream << "        </constraint>" << endl;

               free ((void *)pcharShortName);
            }

            *pofstream << "        <constraint sense='true'>" << endl;
            *pofstream << "          <driver>omni</driver>" << endl;
            *pofstream << "          <printer>printer/" << *pstringDeviceName << "</printer>" << endl;
            *pofstream << "        </constraint>" << endl;
         }
      }

      if (!fFirstTime)
      {
         *pofstream << "      </constraints>" << endl;
         *pofstream << "    </enum_val>" << endl;
      }
   }

   *pofstream << "  </enum_vals>" << endl;
   *pofstream << "</option>" << endl;

   /**
    * Create Ink Type
    *
    * list of printmodes
    */
   // raharper@us.ibm.com changed filename to printmode
   pofstream = &ofstreamPrintModes;

   cout << "Generating omni-printmode.xml" << endl;

   *pofstream << "<option type=\"enum\" id=\"opt/omni-inktype\">" << endl;
   *pofstream << "  <!-- A multilingual <comments> block can appear here, too;" << endl;
   *pofstream << "       it should be treated as documentation for the user. -->" << endl;
   *pofstream << "  <arg_longname>" << endl;
   // raharper@us.ibm.com, changed longname
   *pofstream << "   <en>Printing Mode</en>" << endl;
   *pofstream << "  </arg_longname>" << endl;
   *pofstream << "  <arg_shortname>" << endl;
   *pofstream << "   <en>Mode</en><!-- backends only know <en> shortnames! -->" << endl;
   *pofstream << "  </arg_shortname>" << endl;
   *pofstream << "  <arg_execution>" << endl;
   *pofstream << "   <arg_order>100</arg_order>" << endl;
   *pofstream << "   <arg_spot>A</arg_spot>" << endl;
   *pofstream << "   <arg_substitution />" << endl;
   // raharper@us.ibm.com, changed prototype
   *pofstream << "   <arg_proto> printmode=%s</arg_proto>" << endl;
   *pofstream << "  </arg_execution>" << endl;
   *pofstream << endl;

   *pofstream << "  <constraints>" << endl;

   for (iCurrentDevice = 0; iCurrentDevice < iNumDevices; iCurrentDevice++)
   {
      Device          *pDevice           = pohDevices[iCurrentDevice].pDevice;
      DevicePrintMode *pDR               = pDevice->getCurrentPrintMode ();
      string          *pstringDeviceName = pohDevices[iCurrentDevice].pstringDeviceName;

      *pofstream << "    <constraint sense='true'>" << endl;
      *pofstream << "      <driver>omni</driver>" << endl;
      *pofstream << "      <printer>printer/" << *pstringDeviceName << "</printer>" << endl;
      // raharper@us.ibm.com, changed defval
      *pofstream << "      <arg_defval>" << DevicePrintMode::IDToName (pDR->getID ()) << "</arg_defval>" << endl;
      *pofstream << "    </constraint>" << endl;
   }

   *pofstream << "  </constraints>" << endl;
   *pofstream << endl;
   *pofstream << "  <enum_vals>" << endl;

   for (int iPrintMode = DevicePrintMode::PRINT_MODE_NONE; iPrintMode < DevicePrintMode::COUNT; iPrintMode++)
   {
      bool fFirstTime = true;

      for (iCurrentDevice = 0; iCurrentDevice < iNumDevices; iCurrentDevice++)
      {
         Device          *pDevice           = pohDevices[iCurrentDevice].pDevice;
         DevicePrintMode *pDPM              = pDevice->getCurrentPrintMode ();
         string          *pstringDeviceName = pohDevices[iCurrentDevice].pstringDeviceName;

         if (pDPM->isPrintModeSupported (iPrintMode))
         {
            if (fFirstTime)
            {
               const char *pcharLongName;
               const char *pcharShortName;

               fFirstTime = false;

               // Generate the short name out of the long name
               pcharLongName  = DevicePrintMode::getName (pDevice, iPrintMode);
               pcharShortName = generateShortName (pcharLongName);

               //raharper@us.ibm.com took out the ev/omni-
               *pofstream << "    <enum_val id='" << DevicePrintMode::IDToName (iPrintMode) << "'>" << endl;
               *pofstream << "      <ev_longname><en>" << pcharLongName << "</en></ev_longname>" << endl;
               *pofstream << "      <ev_shortname><en>" << pcharShortName << "</en></ev_shortname>" << endl;
               *pofstream << "      <ev_driverval>" << DevicePrintMode::IDToName (iPrintMode) << "</ev_driverval>" << endl;
               *pofstream << "      <constraints>" << endl;
               *pofstream << "        <!-- Assume the option doesn't apply... -->" << endl;
               *pofstream << "        <constraint sense='false'>" << endl;
               *pofstream << "          <driver>omni</driver>" << endl;
               *pofstream << "        </constraint>" << endl;

               free ((void *)pcharShortName);
            }

            *pofstream << "        <constraint sense='true'>" << endl;
            *pofstream << "          <driver>omni</driver>" << endl;
            *pofstream << "          <printer>printer/" << *pstringDeviceName << "</printer>" << endl;
            *pofstream << "        </constraint>" << endl;
         }
      }

      if (!fFirstTime)
      {
         *pofstream << "      </constraints>" << endl;
         *pofstream << "    </enum_val>" << endl;
      }
   }

   *pofstream << "  </enum_vals>" << endl;
   *pofstream << "</option>" << endl;

   /**
    * Create Image Type
    *
    * list of dithers
    */
   pofstream = &ofstreamDithers;

   cout << "Generating omni-dither.xml" << endl;

   *pofstream << "<option type=\"enum\" id=\"opt/omni-dither\">" << endl;
   *pofstream << "  <!-- A multilingual <comments> block can appear here, too;" << endl;
   *pofstream << "       it should be treated as documentation for the user. -->" << endl;
   *pofstream << "  <arg_longname>" << endl;
   *pofstream << "   <en>Dither Algorithm</en>" << endl;
   *pofstream << "  </arg_longname>" << endl;
   *pofstream << "  <arg_shortname>" << endl;
   *pofstream << "   <en>Dither</en><!-- backends only know <en> shortnames! -->" << endl;
   *pofstream << "  </arg_shortname>" << endl;
   *pofstream << "  <arg_execution>" << endl;
   *pofstream << "   <arg_order>100</arg_order>" << endl;
   *pofstream << "   <arg_spot>A</arg_spot>" << endl;
   *pofstream << "   <arg_substitution />" << endl;
   // raharper@us.ibm.com, changed prototype
   *pofstream << "   <arg_proto> dither=%s</arg_proto>" << endl;
   *pofstream << "  </arg_execution>" << endl;
   *pofstream << endl;

   *pofstream << "  <constraints>" << endl;

   for (iCurrentDevice = 0; iCurrentDevice < iNumDevices; iCurrentDevice++)
   {
      Device *pDevice           = pohDevices[iCurrentDevice].pDevice;
      string *pstringDeviceName = pohDevices[iCurrentDevice].pstringDeviceName;

      *pofstream << "    <constraint sense='true'>" << endl;
      *pofstream << "      <driver>omni</driver>" << endl;
      *pofstream << "      <printer>printer/" << *pstringDeviceName << "</printer>" << endl;
      // raharper@us.ibm.com, changed defval
      *pofstream << "      <arg_defval>" << pDevice->getCurrentDitherID () << "</arg_defval>" << endl;
      *pofstream << "    </constraint>" << endl;
   }

   *pofstream << "  </constraints>" << endl;
   *pofstream << endl;
   *pofstream << "  <enum_vals>" << endl;

   pEnum = DeviceDither::getEnumeration ();
   while (pEnum->hasMoreElements ())
   {
      Device     *pDevice        = pohDevices[0].pDevice;
      char       *pszID          = (char *)pEnum->nextElement ();
      const char *pcharLongName;
      const char *pcharShortName;

      // Generate the short name out of the long name
      pcharLongName  = DeviceDither::getName (pDevice, pszID);
      pcharShortName = generateShortName (pcharLongName);

      //raharper@us.ibm.com took out the ev/omni-
      *pofstream << "    <enum_val id='" << pszID << "'>" << endl;
      *pofstream << "      <ev_longname><en>" << pcharLongName << "</en></ev_longname>" << endl;
      *pofstream << "      <ev_shortname><en>" << pcharShortName << "</en></ev_shortname>" << endl;
      *pofstream << "      <ev_driverval>" << pszID << "</ev_driverval>" << endl;
      *pofstream << "      <constraints>" << endl;
      *pofstream << "        <!-- Assume the option doesn't apply... -->" << endl;
      *pofstream << "        <constraint sense='false'>" << endl;
      *pofstream << "          <driver>omni</driver>" << endl;
      *pofstream << "        </constraint>" << endl;
      free ((void *)pcharShortName);

      for (iCurrentDevice = 0; iCurrentDevice < iNumDevices; iCurrentDevice++)
      {
         string *pstringDeviceName = pohDevices[iCurrentDevice].pstringDeviceName;

         *pofstream << "        <constraint sense='true'>" << endl;
         *pofstream << "          <driver>omni</driver>" << endl;
         *pofstream << "          <printer>printer/" << *pstringDeviceName << "</printer>" << endl;
         *pofstream << "        </constraint>" << endl;
      }

      *pofstream << "      </constraints>" << endl;
      *pofstream << "    </enum_val>" << endl;
   }

   *pofstream << "  </enum_vals>" << endl;
   *pofstream << "</option>" << endl;

   /**
    * Make Orientation section
    *
    * List orientations
    */
   pofstream = &ofstreamOrientation;

   cout << "Generating omni-orientation.xml" << endl;

   *pofstream << "<option type=\"enum\" id=\"opt/omni-orientation\">" << endl;
   *pofstream << "  <!-- A multilingual <comments> block can appear here, too;" << endl;
   *pofstream << "       it should be treated as documentation for the user. -->" << endl;
   *pofstream << "  <arg_longname>" << endl;
   // Edited by raharper@us.ibm.com, changing the output to match omni driver options
   *pofstream << "   <en>Orientation</en>" << endl;
   *pofstream << "  </arg_longname>" << endl;
   *pofstream << "  <arg_shortname>" << endl;
   // Edited by raharper@us.ibm.com, changing the output to match omni driver options
   *pofstream << "   <en>Orientation</en><!-- backends only know <en> shortnames! -->" << endl;
   *pofstream << "  </arg_shortname>" << endl;
   *pofstream << "  <arg_execution>" << endl;
   *pofstream << "   <arg_order>100</arg_order>" << endl;
   *pofstream << "   <arg_spot>A</arg_spot>" << endl;
   *pofstream << "   <arg_substitution />" << endl;
   // Edited by raharper@us.ibm.com, changing the output to match omni driver options
   *pofstream << "   <arg_proto> orientation=%s</arg_proto>" << endl;
   *pofstream << "  </arg_execution>" << endl;
   *pofstream << endl;

   *pofstream << "  <constraints>" << endl;

   for (iCurrentDevice = 0; iCurrentDevice < iNumDevices; iCurrentDevice++)
   {
      Device            *pDevice           = pohDevices[iCurrentDevice].pDevice;
      DeviceOrientation *pDT               = pDevice->getCurrentOrientation ();
      string            *pstringDeviceName = pohDevices[iCurrentDevice].pstringDeviceName;

      *pofstream << "    <constraint sense='true'>" << endl;
      *pofstream << "      <driver>omni</driver>" << endl;
      *pofstream << "      <printer>printer/" << *pstringDeviceName << "</printer>" << endl;
      // raharper@us.ibm.com, changing what the defval output looks like
      *pofstream << "      <arg_defval>" << DeviceOrientation::IDToName (pDT->getID ()) << "</arg_defval>" << endl;
      *pofstream << "    </constraint>" << endl;
   }

   *pofstream << "  </constraints>" << endl;
   *pofstream << endl;
   *pofstream << "  <enum_vals>" << endl;

   for (int iOrient = DeviceOrientation::ORIENTATION_NONE; iOrient < DeviceOrientation::COUNT; iOrient++)
   {
      bool fFirstTime = true;

      for (iCurrentDevice = 0; iCurrentDevice < iNumDevices; iCurrentDevice++)
      {
         Device            *pDevice           = pohDevices[iCurrentDevice].pDevice;
         DeviceOrientation *pDT               = pDevice->getCurrentOrientation ();
         string            *pstringDeviceName = pohDevices[iCurrentDevice].pstringDeviceName;

         if (pDT->isOrientationSupported (iOrient))
         {
            if (fFirstTime)
            {
               const char     *pcharLongName;
               const char     *pcharShortName;

               fFirstTime = false;

               // Generate the short name out of the long name
               pcharLongName  = DeviceOrientation::getName (pDevice, iOrient);
               pcharShortName = generateShortName (pcharLongName);

               //raharper@us.ibm.com took out the ev/omni-
               *pofstream << "    <enum_val id='" << DeviceOrientation::IDToName (iOrient) << "'>" << endl;
               *pofstream << "      <ev_longname><en>" << pcharLongName << "</en></ev_longname>" << endl;
               *pofstream << "      <ev_shortname><en>" << pcharShortName << "</en></ev_shortname>" << endl;
               *pofstream << "      <ev_driverval>" << DeviceOrientation::IDToName (iOrient) << "</ev_driverval>" << endl;
               *pofstream << "      <constraints>" << endl;
               *pofstream << "        <!-- Assume the option doesn't apply... -->" << endl;
               *pofstream << "        <constraint sense='false'>" << endl;
               *pofstream << "          <driver>omni</driver>" << endl;
               *pofstream << "        </constraint>" << endl;

               free ((void *)pcharShortName);
            }

            *pofstream << "        <constraint sense='true'>" << endl;
            *pofstream << "          <driver>omni</driver>" << endl;
            *pofstream << "          <printer>printer/" << *pstringDeviceName << "</printer>" << endl;
            *pofstream << "        </constraint>" << endl;
         }
      }

      if (!fFirstTime)
      {
         *pofstream << "      </constraints>" << endl;
         *pofstream << "    </enum_val>" << endl;
      }
   }

   *pofstream << "  </enum_vals>" << endl;
   *pofstream << "</option>" << endl;

   // Make an option to turn on or off GhostScript dithering. It will apply
   // to all printer models ("-smonodither=GSMONO")

   pofstream = &ofstreamMonoDither;;

   cout << "Generating omni-monodither.xml" << endl;

   *pofstream << "<option type=\"enum\" id=\"opt/omni-monodither\">" << endl;
   *pofstream << "  <!-- A multilingual <comments> block can appear here, too;" << endl;
   *pofstream << "       it should be treated as documentation for the user. -->" << endl;
   *pofstream << "  <arg_longname>" << endl;
   *pofstream << "   <en>Monochrome Dithering</en>" << endl;
   *pofstream << "  </arg_longname>" << endl;
   *pofstream << "  <arg_shortname>" << endl;
   *pofstream << "   <en>MonoDither</en><!-- backends only know <en> shortnames! -->" << endl;
   *pofstream << "  </arg_shortname>" << endl;
   *pofstream << "  <arg_execution>" << endl;
   *pofstream << "   <arg_order>90</arg_order>" << endl;
   *pofstream << "   <arg_spot>B</arg_spot>" << endl;
   *pofstream << "   <arg_substitution />" << endl;
   *pofstream << "   <arg_proto>%s</arg_proto>" << endl;
   *pofstream << "  </arg_execution>" << endl;
   *pofstream << endl;

   *pofstream << "  <constraints>" << endl;
   *pofstream << "    <constraint sense='true'>" << endl;
   *pofstream << "      <driver>omni</driver>" << endl;
   *pofstream << "      <arg_defval>MONODITHER_GS</arg_defval>" << endl;
   *pofstream << "    </constraint>" << endl;
   *pofstream << "  </constraints>" << endl;
   *pofstream << endl;
   *pofstream << "  <enum_vals>" << endl;
   *pofstream << "    <enum_val id='MONODITHER_GS'>" << endl;
   *pofstream << "      <ev_longname><en>GhostScript (Faster)</en></ev_longname>" << endl;
   *pofstream << "      <ev_shortname><en>GS</en></ev_shortname>" << endl;
   *pofstream << "      <ev_driverval> -smonodither=GSMONO</ev_driverval>" << endl;
   *pofstream << "    </enum_val>" << endl;
   *pofstream << "    <enum_val id='MONODITHER_OMNI'>" << endl;
   *pofstream << "      <ev_longname><en>Omni (Better quality)</en></ev_longname>" << endl;
   *pofstream << "      <ev_shortname><en>Omni</en></ev_shortname>" << endl;
   *pofstream << "      <ev_driverval></ev_driverval>" << endl;
   *pofstream << "    </enum_val>" << endl;
   *pofstream << "  </enum_vals>" << endl;
   *pofstream << "</option>" << endl;

   /*
   ** Make Option section
   **
   ** List Omni driver options
   **
   ** Here is what needs to be done.  There are additional options, that are in
   ** the Omni driver that each printer supports, or rather, Omni does for each
   ** printer, regardless.  For each of these options (to be defined in a file),
   ** this method must generate an omni-<option>.xml describing the properites
   ** of the option in question.  The way this will work is the Foomatic.cpp
   ** code will look for a specific file ( there will be a command line override
   ** added for the user to specify by hand the list of options they want
   ** generated) and Foomatic will read the options from the file and generate
   ** an xml file for each of the options in the file
   **
   ** The reading of the file will be one method, and then there will be a
   ** skeleton function that will take the name of the option and generate the
   ** omni-<option>.xml file with the specifics.  Each option will have a
   ** boolean flag, either on or off.  The only thing left is to determine
   ** whether the default it on or off, that can be specified in the file.
   **
   **
   ** Variables so far:
   ** String OptionLongname
   ** String OptionShortname
   ** String prototype

   ofstream ofstreamOption ("foomatic-db/db/source/opt/omni-myoption.xml");
   pofstream = &ofstreamOption;

   string OptionLongname ("OptionLongname");
   string OptionShortname ("OptionShortname");
   string prototype ("myoption");

   cout << "Generating omni-myoption.xml" << endl;

   *pofstream << "<option type=\"enum\" id=\"opt/omni-myoption\">" << endl;
   *pofstream << "  <!-- A multilingual <comments> block can appear here, too;" << endl;
   *pofstream << "       it should be treated as documentation for the user. -->" << endl;
   *pofstream << "  <arg_longname>" << endl;
   *pofstream << "   <en>" << OptionLongname << "</en><!-- backends only know <en> shortnames! -->" << endl;
   *pofstream << "  <arg_shortname>" << endl;
   *pofstream << "   <en>" << OptionShortname << "</en>" << endl;
   *pofstream << "  </arg_shortname>" << endl;
   *pofstream << "  <arg_execution>" << endl;
   *pofstream << "   <arg_order>100</arg_order>" << endl;
   *pofstream << "   <arg_spot>A</arg_spot>" << endl;
   *pofstream << "   <arg_substitution />" << endl;
   *pofstream << "   <arg_proto> " << prototype << "=%s</arg_proto>" << endl;
   *pofstream << "  </arg_execution>" << endl;
   *pofstream << endl;

   // Put in default value for each printer
   // Note to self:  find out whether this option defaults to 1 or 0
   *pofstream << "  <constraints>" << endl;

   for (iCurrentDevice = 0; iCurrentDevice < iNumDevices; iCurrentDevice++)
   {
      Device            *pDevice           = pohDevices[iCurrentDevice].pDevice;
      string            *pstringDeviceName = pohDevices[iCurrentDevice].pstringDeviceName;

      *pofstream << "    <constraint sense='true'>" << endl;
      *pofstream << "      <driver>omni</driver>" << endl;
      *pofstream << "      <printer>printer/" << *pstringDeviceName << "</printer>" << endl;
      *pofstream << "      <arg_defval>" << 1 << "</arg_defval>" << endl;
      *pofstream << "    </constraint>" << endl;
   }

   *pofstream << "  </constraints>" << endl;
   *pofstream << endl;
   *pofstream << "  <enum_vals>" << endl;


   // Here is where we print out what printers support what options.  This can be stripped down to just
   // producing what the option is, and that all devices support the current option.

   // Iterate over all the possible option values (0 or 1)
   for (int iCurrent = 0; iCurrent < 2; iCurrent++)
   {
      bool fFirstTime = true;

      for (iCurrentDevice = 0; iCurrentDevice < iNumDevices; iCurrentDevice++)
      {
         Device            *pDevice           = pohDevices[iCurrentDevice].pDevice;
         string            *pstringDeviceName = pohDevices[iCurrentDevice].pstringDeviceName;

         if (iCurrent)
         {
            if (fFirstTime)
            {
               fFirstTime = false;
               *pofstream << "    <enum_val id='" << OptionShortname << "'>" << endl;
               *pofstream << "      <ev_longname><en>" << OptionLongname << "</en></ev_longname>" << endl;
               *pofstream << "      <ev_shortname><en>" << OptionShortname << "</en></ev_shortname>" << endl;
               *pofstream << "      <ev_driverval>" << iCurrent << "</ev_driverval>" << endl;
               *pofstream << "      <constraints>" << endl;
               *pofstream << "        <!-- Assume the option doesn't apply... -->" << endl;
               *pofstream << "        <constraint sense='false'>" << endl;
               *pofstream << "          <driver>omni</driver>" << endl;
               *pofstream << "        </constraint>" << endl;
            }

            *pofstream << "        <constraint sense='true'>" << endl;
            *pofstream << "          <driver>omni</driver>" << endl;
            *pofstream << "          <printer>printer/" << *pstringDeviceName << "</printer>" << endl;
            *pofstream << "        </constraint>" << endl;

         }
      }
      if (!fFirstTime)
      {
         *pofstream << "      </constraints>" << endl;
         *pofstream << "    </enum_val>" << endl;
      }
   }

   *pofstream << "  </enum_vals>" << endl;
   *pofstream << "</option>" << endl;

ENDOFOPTIONS
   */

done:
   // Clean up
   for (iCurrentDevice = 0; iCurrentDevice < iNumDevices; iCurrentDevice++)
   {
      delete pohDevices[iCurrentDevice].pstringDeviceName;
      delete pohDevices[iCurrentDevice].pDevice;
      g_module_close (pohDevices[iCurrentDevice].hmodDevice);
   }

   delete[] pohDevices;

   return 0;
}
