/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include "GhostscriptInterface.hpp"

#include "Omni.hpp"

#include <stdio.h>
#include <sstream>

#include <glib.h>
#include <gmodule.h>

/*--------------------------------------------------------------------------------*/
/*                          GhostscriptInferfaceInit                              */
/*--------------------------------------------------------------------------------*/
void
GhostscriptInferfaceInit (void *pOmni)
{
   core_omni_device *pCoreOmni = (core_omni_device *)pOmni;

   if (!isOmni (pCoreOmni->cSignature))
      return;

   strcpy (pCoreOmni->cVersion, GIVersion);
}

/*--------------------------------------------------------------------------------*/
/*                          GhostscriptInferfaceTerm                              */
/*--------------------------------------------------------------------------------*/
void
GhostscriptInferfaceTerm (void *pOmni)
{
   core_omni_device *pCoreOmni = (core_omni_device *)pOmni;

   if (!isOmni (pCoreOmni->cSignature))
      return;

   if (pCoreOmni->pfpErr)
   {
      fclose (pCoreOmni->pfpErr);

      pCoreOmni->pfpErr = 0;
   }
}

/*------------------------------------------------------------------------*/
/*  GetResolutionInfo                                                     */
/*------------------------------------------------------------------------*/
bool
GetResolutionInfo (void *pOmni, HWRESOLUTION *hwRes)
{
   bool              bRet      = false;
   DeviceResolution *pDevRes;
   core_omni_device *pCoreOmni = (core_omni_device *)pOmni;
   Device           *pDevice;

   if (!isOmni (pCoreOmni->cSignature))  // not an OMNI structure just a pDevice
   {
      pDevice = (Device *)pOmni;
      pDevRes = pDevice->getCurrentResolution ();  // Device Resolution
   }
   else
   {
      // Device Resolution
      pDevRes = pCoreOmni->pDevice->getCurrentResolution ();

#ifndef RETAIL
      if (DebugOutput::shouldOutputOmniInterface ()) DebugOutput::getErrorStream () << "GhostscriptInterface::" << __FUNCTION__ << ": pDev = 0x" << hex << (int)pOmni << ", hwRes = 0x" << (int)hwRes << dec << endl;
#endif
   }

   if (  pCoreOmni
      && pDevRes
      )
   {
      hwRes->xRes = pDevRes->getXRes ();
      hwRes->yRes = pDevRes->getYRes ();
      hwRes->fScanDots = pDevRes->getScanlineMultiple ();
      bRet = true;
   }

   return bRet;
}

/*------------------------------------------------------------------------*/
/*   GetMarginInfo                                                        */
/*------------------------------------------------------------------------*/
bool
GetMarginInfo (void *pOmni, HWMARGINS *hwMargins)
{
   bool              bRet      = false;
   HardCopyCap      *pHCC;                                 // Hard copy information
   core_omni_device *pCoreOmni = (core_omni_device *)pOmni;
   Device           *pDevice;

#ifndef RETAIL
   if (DebugOutput::shouldOutputOmniInterface ()) DebugOutput::getErrorStream () << "GhostscriptInterface::" << __FUNCTION__ << ": pOmni = 0x" << hex << (int)pOmni << ", hwMargins = 0x" << (int)hwMargins << dec << endl;
#endif

   if (!isOmni (pCoreOmni->cSignature))  // not an OMNI structure just a pDevice
   {
      pDevice = (Device *)pOmni;
      pHCC = pDevice->getCurrentForm ()->getHardCopyCap ();
   }
   else
   {
      pHCC = pCoreOmni->pDevice->getCurrentForm ()->getHardCopyCap ();
   }

   if (pHCC)
   {
      hwMargins->fxWidth     = pHCC->getXPels ();
      hwMargins->fyHeight    = pHCC->getYPels ();

      hwMargins->fLeftClip   = 0; // changed since Ghostscript needs to render full size of
      hwMargins->fTopClip    = 0; // the page printable area.  If we add in the margin, we
      hwMargins->fRightClip  = 0; // actually force GS to remove the margin from the printable
      hwMargins->fBottomClip = 0; // area.

//    hwMargins->fLeftClip   = pHCC->getLeftClip ();
//    hwMargins->fTopClip    = pHCC->getTopClip ();
//    hwMargins->fRightClip  = pHCC->getRightClip ();
//    hwMargins->fBottomClip = pHCC->getBottomClip ();
      bRet = true;
   }

   return bRet;
}

/*--------------------------------------------------------------------------------*/
/*                          GetPrintModeInfo                                      */
/*--------------------------------------------------------------------------------*/
bool
GetPrintModeInfo (void *pOmni, PRINTMODE *pPrtMode)
{
   bool              bRet        = false;
   DevicePrintMode  *pDevPrtMode;
   core_omni_device *pCoreOmni   = (core_omni_device *)pOmni;
   Device           *pDevice;

   if (!isOmni (pCoreOmni->cSignature))  // not an OMNI structure just a pDevice
   {
      pDevice = (Device *)pOmni;
      pDevPrtMode = pDevice->getCurrentPrintMode ();
   }
   else
   {
      pDevPrtMode = pCoreOmni->pDevice->getCurrentPrintMode ();
   }

#ifndef RETAIL
   if (DebugOutput::shouldOutputOmniInterface ()) DebugOutput::getErrorStream () << "GhostscriptInterface::" << __FUNCTION__ << ": pOmni = 0x" << hex << (int)pOmni << ", pPrtMode = 0x" << (int)pPrtMode << dec << endl;
#endif

   if (  pCoreOmni
      && pPrtMode
      && pDevPrtMode
      )
   {
      pPrtMode->iBitCount = pDevPrtMode->getLogicalCount ();
      pPrtMode->iPlanes   = pDevPrtMode->getNumPlanes ();
      bRet = true;
   }

   return bRet;
}

/*--------------------------------------------------------------------------------*/
/*                          SetOutputFunction                                     */
/*--------------------------------------------------------------------------------*/
void
SetOutputFunction (void *pOmni, PFNOUTPUTFUNCTION pfn, void *pMC)
{
   core_omni_device *pCoreOmni = (core_omni_device *)pOmni;

   if (!pCoreOmni->bPDCDevice)
      pCoreOmni->pDevice->setOutputFunction (pfn, pMC);
}

/*--------------------------------------------------------------------------------*/
/*                          BeginJob                                              */
/*--------------------------------------------------------------------------------*/
void *
BeginJob (void *pOmni, FILE *prn_stream)
{
   core_omni_device *pCoreOmni = (core_omni_device *)pOmni;
   Device           *pDevice;

#ifndef RETAIL
   if (DebugOutput::shouldOutputOmniInterface ()) DebugOutput::getErrorStream () << "GhostscriptInterface::" << __FUNCTION__ << ": pOmni = 0x" << hex << (int)pOmni << ", pfpOut = 0x" << (int)pCoreOmni->pfpOut << dec << endl;
#endif

   if (!isOmni (pCoreOmni->cSignature))  // not an OMNI structure just a pDevice
   {
      pDevice = (Device *)pOmni;
      pDevice->setOutputStream (prn_stream);
      pDevice->beginJob ();
   }
   else
   {
      if (pCoreOmni->bPDCDevice)
      {
         int fdOut = STDOUT_FILENO;
         int fdErr = STDERR_FILENO;

         // delete the previous device so that we can
         // get the correct file handle to the spawned
         // process - the first device was for queries
         delete pCoreOmni->pDevice;

         if (pCoreOmni->pfpOut)
         {
            fdOut = fileno (pCoreOmni->pfpOut);
         }
         if (pCoreOmni->pfpErr)
         {
            fdErr = fileno (pCoreOmni->pfpErr);
         }

         pCoreOmni->pDevice = new OmniPDCProxy (pCoreOmni->cServer,       // client exe to spawn
                                                pCoreOmni->cDeviceName,   // device name
                                                pCoreOmni->pszJobOptions, // job properties
                                                true,                     // is renderer
                                                fdOut,                    // stdout
                                                fdErr);                   // stderr

         pCoreOmni->pDevice->beginJob ();
      }
      else
      {
         pCoreOmni->pDevice->setOutputStream (pCoreOmni->pfpOut);
         pCoreOmni->pDevice->beginJob ();
      }
   }

   return 0;
}

/*--------------------------------------------------------------------------------*/
/*                          BeginJob                                              */
/*--------------------------------------------------------------------------------*/
void
BeginJob (void *pOmni)
{
   core_omni_device *pCoreOmni = (core_omni_device *)pOmni;

#ifndef RETAIL
   if (DebugOutput::shouldOutputOmniInterface ()) DebugOutput::getErrorStream () << "GhostscriptInterface::" << __FUNCTION__ << ": pOmni = 0x" << hex << (int)pOmni << dec << endl;
#endif

   pCoreOmni->pDevice->beginJob ();
}

/*--------------------------------------------------------------------------------*/
/*                          Rasterize                                             */
/*--------------------------------------------------------------------------------*/
void
Rasterize (void             *pOmni,
           PBYTE             pbBits,
           PBITMAPINFO2      pbmi,
           PSIZEL            psizelPage,
           PRECTL            prectlPageLocation,
           BITBLT_TYPE       eType)
{
   core_omni_device *pCoreOmni = (core_omni_device *)pOmni;
   Device           *pDevice;

   if (!isOmni(pCoreOmni->cSignature))  // not an OMNI structure just a pDevice
   {
      pDevice = (Device *)pOmni;
      pDevice->rasterize (pbBits,
                          pbmi,
                          prectlPageLocation,
                          eType);
   }
   else
   {
      pCoreOmni->pDevice->rasterize (pbBits,
                                     pbmi,
                                     prectlPageLocation,
                                     eType);
   }
}

/*--------------------------------------------------------------------------------*/
/*                          NewPage                                               */
/*--------------------------------------------------------------------------------*/
void
NewFrame (void *pOmni)
{
   core_omni_device *pCoreOmni = (core_omni_device *)pOmni;
   Device           *pDevice;

#ifndef RETAIL
   if (DebugOutput::shouldOutputOmniInterface ()) DebugOutput::getErrorStream () << "GhostscriptInterface::" << __FUNCTION__ << ": pOmni = 0x" << hex << (int)pOmni << dec << endl;
#endif

   if (!isOmni(pCoreOmni->cSignature))  // not an OMNI structure just a pDevice
   {
      pDevice = (Device *)pOmni;
      pDevice->newFrame ();
   }
   else
   {
      pCoreOmni->pDevice->newFrame ();
   }
}

/*--------------------------------------------------------------------------------*/
/*                          EndJob                                                */
/*--------------------------------------------------------------------------------*/
void
EndJob (void *pOmni)
{
   core_omni_device *pCoreOmni = (core_omni_device *)pOmni;
   Device           *pDevice;

#ifndef RETAIL
   if (DebugOutput::shouldOutputOmniInterface ()) DebugOutput::getErrorStream () << "GhostscriptInterface::" << __FUNCTION__ << ": pOmni = 0x" << hex << (int)pOmni << dec << endl;
#endif

   if (!isOmni(pCoreOmni->cSignature))  // not an OMNI structure just a pDevice
   {
       pDevice = (Device *)pOmni;
       pDevice->endJob ();
       delete pDevice;
   }
   else
   {
      pCoreOmni->pDevice->endJob ();
      delete pCoreOmni->pDevice;
   }
}

/*--------------------------------------------------------------------------------*/
/*                            CreateDevice                                        */
/*--------------------------------------------------------------------------------*/
void *
CreateDevice (void  *pOmni,
              void **pvhDevice,
              int    iUsePDC)
{
   core_omni_device *pCoreOmni = (core_omni_device *)pOmni;

   if (!isOmni (pCoreOmni->cSignature))  // not an OMNI structure
      return 0;

   char    *pszDeviceName    = pCoreOmni->cDeviceName;
   char    *pszJobProperties = pCoreOmni->pszJobOptions;
   char    *pszCerr          = pCoreOmni->cDebugFile;
   char    *pszPDCServerName = pCoreOmni->cServer;
   Device  *pDevice          = 0;
   FILE    *pfpErr           = stderr;

#ifndef RETAIL
   if (DebugOutput::shouldOutputOmniInterface ()) DebugOutput::getErrorStream () << "GhostscriptInterface::" << __FUNCTION__ << endl;
   if (DebugOutput::shouldOutputOmniInterface ()) DebugOutput::getErrorStream () << "GhostscriptInterface::" << __FUNCTION__ << ": pszDeviceName    = " << pszDeviceName << endl;
   if (DebugOutput::shouldOutputOmniInterface ()) DebugOutput::getErrorStream () << "GhostscriptInterface::" << __FUNCTION__ << ": pvhDevice        = 0x" << hex << (int)pvhDevice << dec << endl;
   if (DebugOutput::shouldOutputOmniInterface ()) DebugOutput::getErrorStream () << "GhostscriptInterface::" << __FUNCTION__ << ": pszCerr          = " << pszCerr << endl;
   if (DebugOutput::shouldOutputOmniInterface ()) DebugOutput::getErrorStream () << "GhostscriptInterface::" << __FUNCTION__ << ": pszJobProperties = " << pszJobProperties << dec << endl;
#endif

   // Added for debugging convienence
   if (  pszCerr
      && pszCerr[0]
      )
   {
      pCoreOmni->pfpErr = fopen (pszCerr, "w");

      pfpErr = pCoreOmni->pfpErr;
   }

   // NOTE: Yes.  pDevice == 0 at this point
   Omni::setErrorStream (pDevice, pfpErr);

   if (iUsePDC)
   {
      // initial device is created for setting up caller with correct option and printable area
      pDevice = new OmniPDCProxy (pszPDCServerName,      // client exe to spawn
                                  pszDeviceName,         // device name
                                  pszJobProperties,      // job properties
                                  true,                  // is renderer
                                  0,                     // stdout
                                  fileno (pfpErr));      // stderr
   }
   else
   {
      if (!g_module_supported ())
      {
         DebugOutput::getErrorStream () << "This program needs glib's module routines!" << endl;

         return 0;
      }

#define LIBDEVICENAME3(s) STRINGIZE(s)
#define LIBDEVICENAME2(x) LIBDEVICENAME3(lib##x##.so)
#define LIBDEVICENAME1(x) LIBDEVICENAME2(x)
#define LIBDEVICENAME     LIBDEVICENAME1(DEVICENAME)

typedef Device * (*PFNNEWDEVICEWARGS) (char *pszJobProperties, bool fAdvanced);
typedef Device * (*PFNNEWDEVICE)      (bool fAdvanced);

      char               cDeviceName[64];
      PFNNEWDEVICE       pfnNewDevice      = 0;
      PFNNEWDEVICEWARGS  pfnNewDeviceWArgs = 0;

      if (0 != strncmp (pszDeviceName, "lib", 3))
         sprintf (cDeviceName, "lib%s.so", pszDeviceName);
      else
         strcpy (cDeviceName, pszDeviceName);

#ifndef RETAIL
      if (DebugOutput::shouldOutputOmniInterface ())
         DebugOutput::getErrorStream () << "GhostscriptInterface::" << __FUNCTION__ << ": Trying to load " << cDeviceName << endl;
#endif

      Omni::openDeviceLibrary (cDeviceName, (GModule **)pvhDevice);

#ifndef RETAIL
      if (DebugOutput::shouldOutputOmniInterface ())
         DebugOutput::getErrorStream () << "GhostscriptInterface::" << __FUNCTION__ << ": *pvhDevice = " << hex << *pvhDevice << dec << endl;
#endif

      // nm libHP_Deskjet_1120Cxi.so

      // returns 00011e20 T newDeviceW_Advanced for
      // extern Device *newDevice (bool fAdvanced);
      g_module_symbol ((GModule *)*pvhDevice, "newDeviceW_Advanced", (gpointer *)&pfnNewDevice);

#ifndef RETAIL
      if (DebugOutput::shouldOutputOmniInterface ())
         DebugOutput::getErrorStream () << "GhostscriptInterface::" << __FUNCTION__ << ": pfnNewDevice = 0x" << hex << (int)pfnNewDevice << endl;
#endif

      if (!pfnNewDevice)
      {
         DebugOutput::getErrorStream () << "GhostscriptInterface::" << __FUNCTION__ << ": g_module_error returns " << dec << g_module_error () << endl;

         return 0;
      }

      // returns 00011d30 T newDeviceW_JobProp_Advanced for
      // extern Device *newDevice (char *pszJobProperties, bool fAdvanced);
      g_module_symbol ((GModule *)*pvhDevice, "newDeviceW_JopProp_Advanced", (gpointer *)&pfnNewDeviceWArgs);

#ifndef RETAIL
      if (DebugOutput::shouldOutputOmniInterface ())
         DebugOutput::getErrorStream () << "GhostscriptInterface::" << __FUNCTION__ << ": pfnNewDeviceWArgs = 0x" << hex << (int)pfnNewDeviceWArgs << endl;
#endif

      if (!pfnNewDeviceWArgs)
      {
         DebugOutput::getErrorStream () << "GhostscriptInterface::" << __FUNCTION__ << ": g_module_error returns " << dec << g_module_error () << endl;

         return 0;
      }

      if (pszJobProperties && pszJobProperties[0])
         pDevice = pfnNewDeviceWArgs (pszJobProperties, true);
      else
         pDevice = pfnNewDevice (true);
   }

#ifndef RETAIL
   if (pfpErr)
   {
      Omni::setErrorStream (pDevice, pfpErr);
   }

   if (DebugOutput::shouldOutputOmniInterface ())
      DebugOutput::getErrorStream () << "GhostscriptInterface::" << __FUNCTION__ << ": pDevice = " << *pDevice << endl;
#endif

   if (pDevice->hasError ())
   {
      delete pDevice;

      return 0;
   }

   strcpy (pCoreOmni->cOmniVersion, pDevice->getVersion ());

   return (void *)pDevice;
}

/*--------------------------------------------------------------------------------*/
/*                          isOmni                                                */
/*--------------------------------------------------------------------------------*/
bool
isOmni (char *pszSignature)
{
   bool iRet = false;

   if (!memcmp (pszSignature, Signature, 4))  // check for signature
      iRet = true;

   return iRet;
}
