/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include "OkiPos_Instance.hpp"
#include "defines.hpp"

#include <iostream>
#include <sstream>

const static bool fTestNoCompression = true;

OkiPos_Instance::
OkiPos_Instance (PrintDevice *pDevice)
   : DeviceInstance (pDevice)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputInstance ()) DebugOutput::getErrorStream () << "OkiPos_Instance::OkiPos_Instance ()" << endl;
#endif

   fHaveInitialized_d  = false;
   fHaveSetupPrinter_d = false;
   fUseJournal_d       = false;
   eCutMode_d          = CUT_MODE_NONE;
   ptlPrintHead_d.x    = 0;
   ptlPrintHead_d.y    = 0;
}

OkiPos_Instance::
~OkiPos_Instance ()
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputInstance ()) DebugOutput::getErrorStream () << "OkiPos_Instance::~OkiPos_Instance ()" << endl;
#endif
}

void OkiPos_Instance::
initializeInstance ()
{
   if (fHaveInitialized_d)
      return;

   fHaveInitialized_d = true;

#ifndef RETAIL
   if (DebugOutput::shouldOutputInstance ()) DebugOutput::getErrorStream () << "OkiPos_Instance::initializeInstance ()" << endl;
#endif
}

static char *apszDeviceJobPropertyKeysD[] = {
   "journal",
   "cutmode"
};
static char *apszDeviceJobPropertyKeysS[] = {
   "cutmode"
};
#define JOBPROP_USEJOURNAL apszDeviceJobPropertyKeysD[0]
#define JOBPROP_USECUTMODE apszDeviceJobPropertyKeysD[1]

static char *apszDeviceJobPropertyValues[] = {
   "true",
   "false",
   "none",
   "full",
   "partial"
};
#define JOBPROP_TRUE    apszDeviceJobPropertyValues[0]
#define JOBPROP_FALSE   apszDeviceJobPropertyValues[1]
#define JOBPROP_NONE    apszDeviceJobPropertyValues[2]
#define JOBPROP_FULL    apszDeviceJobPropertyValues[3]
#define JOBPROP_PARTIAL apszDeviceJobPropertyValues[4]

Enumeration * OkiPos_Instance::
listDeviceJobPropertyKeys ()
{
   class DeviceJobPropertyKeyEnumerator : public Enumeration
   {
   public:
      DeviceJobPropertyKeyEnumerator (char **apszKeys, int iDimOfKeys)
      {
         iWhichKey_d = iDimOfKeys;
         apszKeys_d  = apszKeys;
      }

      virtual bool hasMoreElements ()
      {
         return iWhichKey_d > 0;
      }

      virtual void * nextElement ()
      {
         if (0 < iWhichKey_d)
         {
            iWhichKey_d--;
            return apszKeys_d[iWhichKey_d];
         }

         return 0;
      }

   private:
      int    iWhichKey_d;
      char **apszKeys_d;
   };

   if (hasDeviceOption (MODEL_S))
   {
      return new DeviceJobPropertyKeyEnumerator (apszDeviceJobPropertyKeysS,
                                                 dimof (apszDeviceJobPropertyKeysS));
   }
   else if (hasDeviceOption (MODEL_D))
   {
      return new DeviceJobPropertyKeyEnumerator (apszDeviceJobPropertyKeysD,
                                                 dimof (apszDeviceJobPropertyKeysD));
   }
   else
   {
      return DeviceInstance::listDeviceJobPropertyKeys ();
   }
}

Enumeration * OkiPos_Instance::
listKeyValues (char *pszKey)
{
   class KeyEnumerator : public Enumeration
   {
   public:
      KeyEnumerator (const char *pszKeys)
      {
         pszKeys_d    = 0;
         pszCurrent_d = 0;

         if (  pszKeys
            && *pszKeys
            )
         {
            pszKeys_d = (char *)malloc (strlen (pszKeys) + 1);
            if (pszKeys_d)
            {
               strcpy (pszKeys_d, pszKeys);
               pszCurrent_d = pszKeys_d;
            }
         }
      }

      virtual ~KeyEnumerator ()
      {
         if (pszKeys_d)
         {
            free (pszKeys_d);
            pszKeys_d = 0;
         }
      }

      virtual bool hasMoreElements ()
      {
         return pszCurrent_d ? true : false;
      }

      virtual void *nextElement ()
      {
         void *pvRet = pszCurrent_d;

         if (  pszCurrent_d
            && *pszCurrent_d
            )
         {
            char *pszSpace = strchr (pszCurrent_d, ' ');

            if (pszSpace)
            {
               *pszSpace = '\0';
               pszCurrent_d = pszSpace + 1;
               while (' ' == *pszCurrent_d)
               {
                  pszCurrent_d++;
               }
               if (!*pszCurrent_d)
               {
                  pszCurrent_d = 0;
               }
            }
            else
            {
               pszCurrent_d = 0;
            }
         }
         else
         {
            pszCurrent_d = 0;
         }

         return pvRet;
      }

   private:
      char *pszKeys_d;
      char *pszCurrent_d;
   };

   ostringstream oss;

   if (0 == strcasecmp (pszKey, JOBPROP_USEJOURNAL))
   {
      oss << JOBPROP_TRUE
          << " "
          << JOBPROP_FALSE;
   }

   if (hasDeviceOption (MODEL_S))
   {
      if (0 == strcasecmp (pszKey, JOBPROP_USECUTMODE))
      {
         oss << JOBPROP_NONE
             << " "
             << JOBPROP_FULL
             << " "
             << JOBPROP_PARTIAL;
      }
   }

   return new KeyEnumerator (oss.str ().c_str ());
}

bool OkiPos_Instance::
handleDeviceJobProperty (char *pszKey,
                         char *pszValue)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputInstance ()) DebugOutput::getErrorStream () << "OkiPos_Instance::handleDeviceJobProperty ("
                          << pszKey << ", "
                          << pszValue << ")"
                          << endl;
#endif

   if (0 == strcasecmp (pszKey, JOBPROP_USEJOURNAL))
   {
      if (0 == strcmp (pszValue, JOBPROP_TRUE))
      {
         fUseJournal_d = true;
         return true;
      }
      else if (0 == strcmp (pszValue, JOBPROP_FALSE))
      {
         fUseJournal_d = false;
         return true;
      }
   }
   else if (0 == strcasecmp (pszKey, JOBPROP_USECUTMODE))
   {
      if (0 == strcmp (pszValue, JOBPROP_NONE))
      {
         eCutMode_d = CUT_MODE_NONE;
         return true;
      }
      else if (0 == strcmp (pszValue, JOBPROP_FULL))
      {
         eCutMode_d = CUT_MODE_FULL;
         return true;
      }
      else if (0 == strcmp (pszValue, JOBPROP_PARTIAL))
      {
         eCutMode_d = CUT_MODE_PARTIAL;
         return true;
      }
   }

   return false;
}

string * OkiPos_Instance::
getJobPropertyType (char *pszKey)
{
   string *pRet = 0;

   if (0 == strcasecmp (pszKey, JOBPROP_USEJOURNAL))
   {
      pRet = new string ("boolean ");
      *pRet += JOBPROP_FALSE;
   }
   else if (0 == strcasecmp (pszKey, JOBPROP_USECUTMODE))
   {
      pRet = new string ("string ");
      *pRet += JOBPROP_NONE;
   }

   return pRet;
}

string * OkiPos_Instance::
getJobProperty (char *pszKey)
{
   if (0 == strcasecmp (pszKey, JOBPROP_USEJOURNAL))
   {
      if (fUseJournal_d)
      {
         return new string (JOBPROP_TRUE);
      }
      else
      {
         return new string (JOBPROP_FALSE);
      }
   }
   else if (0 == strcasecmp (pszKey, JOBPROP_USECUTMODE))
   {
      switch (eCutMode_d)
      {
      case CUT_MODE_NONE:    return new string (JOBPROP_NONE);
      case CUT_MODE_FULL:    return new string (JOBPROP_FULL);
      case CUT_MODE_PARTIAL: return new string (JOBPROP_PARTIAL);
      }
   }

   return 0;
}

string * OkiPos_Instance::
translateKeyValue (char *pszKey,
                   char *pszValue)
{
   string *pRet = 0;

   if (0 == strcasecmp (pszKey, JOBPROP_USEJOURNAL))
   {
      char *pszRetKey   = StringResource::getString (pDevice_d->getLanguageResource (),
                                                     StringResource::STRINGGROUP_DEVICE_COMMON,
                                                     StringResource::DEVICE_COMMON_JOURNAL);
      char *pszRetValue = 0;
      int   iRetID      = StringResource::DEVICE_COMMON_UNKNOWN;

      if (0 == strcasecmp (pszValue, JOBPROP_TRUE))
         iRetID = StringResource::DEVICE_COMMON_TRUE;
      else if (0 == strcasecmp (pszValue, JOBPROP_FALSE))
         iRetID = StringResource::DEVICE_COMMON_FALSE;

      if (StringResource::DEVICE_COMMON_UNKNOWN != iRetID)
      {
         pszRetValue = StringResource::getString (pDevice_d->getLanguageResource (),
                                                  StringResource::STRINGGROUP_DEVICE_COMMON,
                                                  iRetID);
      }

      if (  pszRetKey
         && pszRetValue
         )
      {
         pRet = new string (pszRetKey);
         *pRet += "=";
         *pRet += pszValue;
      }
   }
   else if (0 == strcasecmp (pszKey, JOBPROP_USECUTMODE))
   {
      char *pszRetKey   = StringResource::getString (pDevice_d->getLanguageResource (),
                                                     StringResource::STRINGGROUP_DEVICE_COMMON,
                                                     StringResource::DEVICE_COMMON_CUTMODE);
      char *pszRetValue = 0;
      int   iRetID      = StringResource::DEVICE_COMMON_UNKNOWN;

      if (0 == strcasecmp (pszValue, "none"))
         iRetID = StringResource::DEVICE_COMMON_NONE;
      else if (0 == strcasecmp (pszValue, "full"))
         iRetID = StringResource::DEVICE_COMMON_FULL;
      else if (0 == strcasecmp (pszValue, "partial"))
         iRetID = StringResource::DEVICE_COMMON_PARTIAL;

      if (StringResource::DEVICE_COMMON_UNKNOWN != iRetID)
      {
         pszRetValue = StringResource::getString (pDevice_d->getLanguageResource (),
                                                  StringResource::STRINGGROUP_DEVICE_COMMON,
                                                  iRetID);
      }

      if (  pszRetKey
         && pszRetValue
         )
      {
         pRet = new string (pszRetKey);
         *pRet += "=";
         *pRet += pszValue;
      }
   }

   return pRet;
}

void OkiPos_Instance::
setupPrinter ()
{
   if (fHaveSetupPrinter_d)
      return;

#ifndef RETAIL
   if (DebugOutput::shouldOutputInstance ()) DebugOutput::getErrorStream () << "OkiPos_Instance::setupPrinter ()" << endl;
#endif

   fHaveSetupPrinter_d = true;

   DeviceResolution *pRes      = getCurrentResolution ();
   DeviceForm       *pDF       = getCurrentForm ();
   HardCopyCap      *pHCC      = pDF->getHardCopyCap ();
   DeviceCommand    *pCommands = getCommands ();
   BinaryData       *pCmd      = 0;

   // @TBD
///pDevice_d->sendBinaryDataToDevice (getCurrentTray ());

   pCmd = pCommands->getCommandData ("cmdSetLineSpacing144inch");
   if (pCmd)
   {
      int iSpacing = 144 * pRes->getScanlineMultiple () / pRes->getYRes ();

      pDevice_d->sendPrintfToDevice (pCmd, iSpacing);
   }

   int iLines = (int)(6.0 * (float)pHCC->getCy () / 25400.0);

#ifndef RETAIL
   if (DebugOutput::shouldOutputInstance ()) DebugOutput::getErrorStream () << "iLines = " << iLines << endl;
#endif

   pCmd = pCommands->getCommandData ("cmdSetPageLengthInLines");
   if (pCmd)
   {
      pDevice_d->sendPrintfToDevice (pCmd, iLines);
   }

   if (fUseJournal_d)
   {
      pCmd = pCommands->getCommandData ("cmdSetJournalMode");
      if (pCmd)
      {
         pDevice_d->sendPrintfToDevice (pCmd, 1);
      }
   }

   DebugOutput::getErrorStream () << "******* " << pHCC->getXPels () << endl;
}

bool OkiPos_Instance::
beginJob ()
{
   DeviceCommand    *pCommands = getCommands ();
   BinaryData       *pCmd      = 0;

#ifndef RETAIL
   if (DebugOutput::shouldOutputInstance ()) DebugOutput::getErrorStream () << "OkiPos_Instance::beginJob ()" << endl;
#endif

   pCmd = pCommands->getCommandData ("cmdInit");
   if (pCmd)
   {
#ifndef RETAIL
      if (DebugOutput::shouldOutputInstance ()) DebugOutput::getErrorStream () << "cmdInit = " << *pCmd << endl;
#endif

      pDevice_d->sendBinaryDataToDevice (pCmd);
   }

   return true;
}

bool OkiPos_Instance::
beginJob (bool fJobPropertiesChanged)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputInstance ()) DebugOutput::getErrorStream () << "OkiPos_Instance::beginJob (with props)" << endl;
   if (DebugOutput::shouldOutputInstance ()) DebugOutput::getErrorStream () << "The job properties are as follows:" << endl;
   if (DebugOutput::shouldOutputInstance ()) DebugOutput::getErrorStream () << "The orientation is = " << getCurrentOrientation() << endl;
   if (DebugOutput::shouldOutputInstance ()) DebugOutput::getErrorStream () << "iDitherID = " << getCurrentDitherID () << endl;
   if (DebugOutput::shouldOutputInstance ()) DebugOutput::getErrorStream () << "The form is = " << getCurrentForm () << endl;
   if (DebugOutput::shouldOutputInstance ()) DebugOutput::getErrorStream () << "The tray is = " << getCurrentTray () << endl;
   if (DebugOutput::shouldOutputInstance ()) DebugOutput::getErrorStream () << "The media is = " << getCurrentMedia () << endl;
   if (DebugOutput::shouldOutputInstance ()) DebugOutput::getErrorStream () << "The resolution is = " << getCurrentResolution () << endl;
   if (DebugOutput::shouldOutputInstance ()) DebugOutput::getErrorStream () << "The print mode is = " << getCurrentPrintMode () << endl;
#endif

   // @TBD - reinitialize with new job properties

   // Call common code
   return beginJob ();
}

bool OkiPos_Instance::
newFrame ()
{
   DeviceCommand    *pCommands = getCommands ();
   BinaryData       *pCmd      = 0;

#ifndef RETAIL
   if (DebugOutput::shouldOutputInstance ()) DebugOutput::getErrorStream () << "OkiPos_Instance::newFrame ()" << endl;
#endif

   pCmd = pCommands->getCommandData ("cmdPageEject");
   if (pCmd)
   {
#ifndef RETAIL
      if (DebugOutput::shouldOutputInstance ()) DebugOutput::getErrorStream () << "cmdPageEject = " << *pCmd << endl;
#endif

      pDevice_d->sendBinaryDataToDevice (pCmd);
   }

   ditherNewFrame ();

   ptlPrintHead_d.x = 0;
   ptlPrintHead_d.y = 0;

   return true;
}

bool OkiPos_Instance::
newFrame (bool fJobPropertiesChanged)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputInstance ()) DebugOutput::getErrorStream () << "OkiPos_Instance::newFrame (with props)" << endl;
   if (DebugOutput::shouldOutputInstance ()) DebugOutput::getErrorStream () << "The job properties are as follows:" << endl;
   if (DebugOutput::shouldOutputInstance ()) DebugOutput::getErrorStream () << "The orientation is = " << getCurrentOrientation() << endl;
   if (DebugOutput::shouldOutputInstance ()) DebugOutput::getErrorStream () << "iDitherID = " << getCurrentDitherID () << endl;
   if (DebugOutput::shouldOutputInstance ()) DebugOutput::getErrorStream () << "The form is = " << getCurrentForm () << endl;
   if (DebugOutput::shouldOutputInstance ()) DebugOutput::getErrorStream () << "The tray is = " << getCurrentTray () << endl;
   if (DebugOutput::shouldOutputInstance ()) DebugOutput::getErrorStream () << "The media is = " << getCurrentMedia () << endl;
   if (DebugOutput::shouldOutputInstance ()) DebugOutput::getErrorStream () << "The resolution is = " << getCurrentResolution () << endl;
   if (DebugOutput::shouldOutputInstance ()) DebugOutput::getErrorStream () << "The print mode is = " << getCurrentPrintMode () << endl;
#endif

   // @TBD - reinitialize with new job properties

   // Call common code
   return newFrame ();
}

bool OkiPos_Instance::
endJob ()
{
   DeviceCommand    *pCommands = getCommands ();
   BinaryData       *pCmd      = 0;

#ifndef RETAIL
   if (DebugOutput::shouldOutputInstance ()) DebugOutput::getErrorStream () << "OkiPos_Instance::endJob ()" << endl;
#endif

   pCmd = pCommands->getCommandData ("cmdPageEject");
   if (pCmd)
   {
#ifndef RETAIL
      if (DebugOutput::shouldOutputInstance ()) DebugOutput::getErrorStream () << "cmdPageEject = " << *pCmd << endl;
#endif

      pDevice_d->sendBinaryDataToDevice (pCmd);
   }

   // If the user wants to cut the receipt, then advance the roll and cut.
   pCmd = 0;
   switch (eCutMode_d)
   {
   case CUT_MODE_FULL:    pCmd = pCommands->getCommandData ("cmdFullCut"); break;
   case CUT_MODE_PARTIAL: pCmd = pCommands->getCommandData ("cmdPartialCut"); break;
   }
   if (pCmd)
   {
      DeviceForm  *pDF       = getCurrentForm ();
      HardCopyCap *pHCC      = pDF->getHardCopyCap ();
      int          iLength   = 0;
      BinaryData  *pCmdSpace = 0;

#ifndef RETAIL
      if (DebugOutput::shouldOutputInstance ()) DebugOutput::getErrorStream () << "cmdXXXCut = " << *pCmd << endl;
#endif

      iLength = (pHCC->getTopClip () + pHCC->getBottomClip ()) * 144 / 25400;

      DebugOutput::getErrorStream () << "iLength = " << iLength << endl;

      pCmdSpace = pCommands->getCommandData ("cmdSetLineSpacing144inch");
      if (pCmdSpace)
      {
         pDevice_d->sendPrintfToDevice (pCmdSpace, iLength);
      }
      pCmdSpace = pCommands->getCommandData ("cmdMoveToNextRasterGraphicsLine");
      if (pCmdSpace)
      {
         pDevice_d->sendBinaryDataToDevice (pCmdSpace);
      }
      pCmdSpace = pCommands->getCommandData ("cmdEndRasterGraphicsLine");
      if (pCmdSpace)
      {
         pDevice_d->sendBinaryDataToDevice (pCmdSpace);
      }

      pDevice_d->sendBinaryDataToDevice (pCmd);
   }

   pCmd = pCommands->getCommandData ("cmdTerm");
   if (pCmd)
   {
#ifndef RETAIL
      if (DebugOutput::shouldOutputInstance ()) DebugOutput::getErrorStream () << "cmdTerm = " << *pCmd << endl;
#endif

      pDevice_d->sendBinaryDataToDevice (pCmd);
   }

   return true;
}

bool OkiPos_Instance::
abortJob ()
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputInstance ()) DebugOutput::getErrorStream () << "OkiPos_Instance::endJob ()" << endl;
#endif

   int cx = getCurrentForm ()->getHardCopyCap ()->getXPels ();

   // The largest size is 24 lines
   int iSize = 24 * (cx + 7) / 8;

   PBYTE abData = new BYTE [iSize];

   memset (abData, 0, sizeof (abData));

   BinaryData data (abData, sizeof (abData));

   pDevice_d->sendBinaryDataToDevice (&data);

   delete[] abData;

   return true;
}

string OkiPos_Instance::
toString (ostringstream& oss)
{
   oss << "{ "
       << DeviceInstance::toString (oss) // Add parent's output as well
       << " }";

   return oss.str ();
}

/* Provide a way to print out class data
*/
ostream&
operator<< (ostream& os, const OkiPos_Instance& const_self)
{
   OkiPos_Instance& self = const_cast<OkiPos_Instance&>(const_self);
   ostringstream    oss;

   os << self.toString (oss);

   return os;
}
