/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Idea borrowed from http://hpinkjet.sourceforge.net
 */
#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <sstream>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/fcntl.h>
#include <sys/shm.h>

#include <glib.h>
#include <gmodule.h>

#include "Device.hpp"
#include "OmniServer.hpp"
#include "PrinterCommand.hpp"
#include "Omni.hpp"

bool
writeData (int          fd,
           PPDC_PACKET  pPacket,
           size_t       cbPacket,
           byte        *pbData,
           size_t       cbData)
{
   size_t     rc;

   // Write the header
   rc = write (fd, pPacket, cbPacket);

   if (rc != cbPacket)
   {
      DebugOutput::getErrorStream () << "server: wrote "
           << rc
           << " bytes, expecting "
           << cbPacket
           << " @ line "
           << __LINE__
           << endl;

      return false;
   }

   if (pbData)
   {
      // Write the data
      rc = write (fd, pbData, cbData);

      if (rc != cbData)
      {
         DebugOutput::getErrorStream () << "server: wrote "
              << rc
              << " bytes, expecting "
              << cbData
              << " @ line "
              << __LINE__
              << endl;

         return false;
      }
   }

   return true;
}

bool
sendResponse (int fd, PDC_CMD eCommand, char *pszLine)
{
   PDC_PACKET cmdHeader;
   size_t     cbLine    = 0;

   cmdHeader.eCommand          = eCommand;
   cmdHeader.cbLength          = sizeof (cmdHeader);
   cmdHeader.eFormat           = PDCFMT_NULL;
   cmdHeader.achCommandLine[0] = '\0';

   if (  pszLine
      && *pszLine
      )
   {
      cbLine = strlen (pszLine);

      cmdHeader.eFormat           = PDCFMT_STRING;
      cmdHeader.cbLength         += cbLine;
      cmdHeader.achCommandLine[0] = pszLine[0];

      pszLine++;
   }

   return writeData (fd,
                     &cmdHeader,
                     sizeof (cmdHeader),
                     (byte *)pszLine,
                     cbLine);
}

bool
sendResponse (int fd, PDC_CMD eCommand, string *pLine)
{
   return sendResponse (fd,
                        eCommand,
                        (char *)pLine->c_str ());
}

bool
sendResponse (int fd, PDC_CMD eCommand, char **ppszLine)
{
   PDC_PACKET cmdHeader;
   size_t     cbLine    = 0;
   byte      *pbLine    = 0;

   cmdHeader.eCommand          = eCommand;
   cmdHeader.cbLength          = sizeof (cmdHeader);
   cmdHeader.eFormat           = PDCFMT_NULL;
   cmdHeader.achCommandLine[0] = '\0';

   if (  ppszLine
      && *ppszLine
      )
   {
      char *pszLine = (char *)ppszLine;

      cmdHeader.eFormat           = PDCFMT_STRING_ARRAY;
      cmdHeader.achCommandLine[0] = *pszLine++;

      pbLine = (byte *)pszLine;

      while (*pszLine)
      {
         cbLine += strlen (pszLine) + 1;

         pszLine += strlen (pszLine) + 1;
      }

      cbLine++;

      cmdHeader.cbLength += cbLine;
   }

   return writeData (fd,
                     &cmdHeader,
                     sizeof (cmdHeader),
                     pbLine,
                     cbLine);
}

bool
sendResponse (int fd, PDC_CMD eCommand, int iValue)
{
   PDC_PACKET cmdHeader;
   byte       abData[sizeof (int)];

   cmdHeader.eCommand = eCommand;
   cmdHeader.cbLength = sizeof (cmdHeader) - 1 + sizeof (int);
   cmdHeader.eFormat  = PDCFMT_INTEGER;

   for (size_t i = 0; i < sizeof (int); i++)
   {
      abData[i] = iValue & 0xFF;

      iValue >>= 8;
   }

   return writeData (fd,
                     &cmdHeader,
                     sizeof (cmdHeader) - 1,
                     abData,
                     sizeof (abData));
}

bool
sendResponse (int fd, PDC_CMD eCommand, long int lValue)
{
   PDC_PACKET cmdHeader;
   byte       abData[sizeof (long int)];

   cmdHeader.eCommand = eCommand;
   cmdHeader.cbLength = sizeof (cmdHeader) - 1 + sizeof (long int);
   cmdHeader.eFormat  = PDCFMT_LONG;

   for (size_t i = 0; i < sizeof (long int); i++)
   {
      abData[i] = lValue & 0xFF;

      lValue >>= 8;
   }

   return writeData (fd,
                     &cmdHeader,
                     sizeof (cmdHeader) - 1,
                     abData,
                     sizeof (abData));
}

bool
sendResponse (int fd, PDC_CMD eCommand, byte *pbData, int cbData)
{
   PDC_PACKET cmdHeader;

   cmdHeader.eCommand = eCommand;
   cmdHeader.cbLength = sizeof (cmdHeader) - 1 + cbData;
   cmdHeader.eFormat  = PDCFMT_BINARY;

   return writeData (fd,
                     &cmdHeader,
                     sizeof (cmdHeader) - 1,
                     pbData,
                     cbData);
}

class DeviceReference
{
public:
   DeviceReference (char *pszDeviceName)
   {
      pszLibDeviceName_d  = 0;
      hmodDevice_d        = 0;
      pfnNewDevice_d      = 0;
      pfnNewDeviceWArgs_d = 0;
      pszJobProperties_d  = 0;
      pDevice_d           = 0;

      int cbLibDeviceName = strlen (pszDeviceName) + 1;

      // Check if it starts with lib
      if (0 != strncasecmp (pszDeviceName, "lib", 3))
         cbLibDeviceName += 6;

      pszLibDeviceName_d = (char *)malloc (cbLibDeviceName);

      if (pszLibDeviceName_d)
      {
         if (0 != strncasecmp (pszDeviceName, "lib", 3))
         {
            sprintf (pszLibDeviceName_d, "lib%s.so", pszDeviceName);
         }
         else
         {
            strcpy (pszLibDeviceName_d, pszDeviceName);
         }
      }
   }

   ~DeviceReference ()
   {
      delete pDevice_d;
      pDevice_d = 0;

      if (hmodDevice_d)
      {
         g_module_close (hmodDevice_d);
         hmodDevice_d = 0;
      }

      if (pszLibDeviceName_d)
      {
         free (pszLibDeviceName_d);
         pszLibDeviceName_d = 0;
      }

      if (pszJobProperties_d)
      {
         free (pszJobProperties_d);
         pszJobProperties_d = 0;
      }
   }

   bool
   loadDevice ()
   {
      if (!pszLibDeviceName_d)
         return false;

      if (hmodDevice_d)
         return true;

      Omni::openDeviceLibrary (pszLibDeviceName_d, &hmodDevice_d);

#ifndef RETAIL
      if (DebugOutput::shouldOutputOmniServer ()) DebugOutput::getErrorStream () << "hmodDevice = 0x" << hex << (int)hmodDevice_d << dec << endl;
#endif

      if (!hmodDevice_d)
         return false;

      // Get the constructor
      g_module_symbol (hmodDevice_d,
                       "newDeviceW_Advanced",
                       (gpointer *)&pfnNewDevice_d);

#ifndef RETAIL
      if (DebugOutput::shouldOutputOmniServer ()) DebugOutput::getErrorStream () << "pfnNewDevice = 0x" << hex << (int)pfnNewDevice_d << dec << endl;
#endif

      g_module_symbol (hmodDevice_d,
                       "newDeviceW_JopProp_Advanced",
                       (gpointer *)&pfnNewDeviceWArgs_d);

#ifndef RETAIL
      if (DebugOutput::shouldOutputOmniServer ()) DebugOutput::getErrorStream () << "pfnNewDeviceWArgs = 0x" << hex << (int)pfnNewDeviceWArgs_d << dec << endl;
#endif

      if (  !pfnNewDevice_d
         || !pfnNewDeviceWArgs_d
         )
      {
         g_module_close (hmodDevice_d);
         hmodDevice_d = 0;

         return false;
      }
      else
      {
         return true;
      }
   }

   void
   instantiateDevice (char *pszJobProperties, bool fAdvanced)
   {
      delete pDevice_d;
      pDevice_d = 0;

      if (  pszJobProperties
         && *pszJobProperties
         )
      {
         pDevice_d = pfnNewDeviceWArgs_d (pszJobProperties, fAdvanced);
      }
      else
      {
         pDevice_d = pfnNewDevice_d (fAdvanced);
      }

#ifndef RETAIL
      if (DebugOutput::shouldOutputOmniServer ()) DebugOutput::getErrorStream () << "pDevice = " << *pDevice_d << endl;
#endif
   }

   Device *
   getDevice ()
   {
      return pDevice_d;
   }

private:
   char              *pszLibDeviceName_d;
   GModule           *hmodDevice_d;
   PFNNEWDEVICE       pfnNewDevice_d;
   PFNNEWDEVICEWARGS  pfnNewDeviceWArgs_d;
   char              *pszJobProperties_d;
   Device            *pDevice_d;
};

int
main (int argc, char *argv[])
{
   char              *pszS2C;
   char              *pszC2S;
   int                fdS2C                 = -1;            /* server to client */
   int                fdC2S                 = -1;            /* client to server */
   bool               fRun                  = true;
   PDC_CMD            eCommand              = PDCCMD_NACK;
   PrinterCommand    *pCmd                  = 0;
   DeviceReference   *pDeviceRef            = 0;
   bool               fAdvanced             = false;
   char              *pszJobProperties      = 0;
   size_t             cbJobProperties       = 0;
   bool               fNewJobPropertiesSent = false;
   byte              *pbBuffer1             = 0;
   byte              *pbBuffer2             = 0;
   FILE              *pfpOut                = 0;
   FILE              *pfpErr                = 0;
   int                iCurrentLanguage      = StringResource::LANGUAGE_DEFAULT;
   int                iPageNumber           = 0;

   if (getenv ("OMNIDEBUGOUTPUT"))
   {
      char *pszString = getenv ("OMNIDEBUGOUTPUT");

      DebugOutput::applyAllDebugOutput (pszString);
   }

   if (getenv ("OMNISERVERPAUSE"))
   {
      // Handy to attach via gdb debugger.  do:
      // (gdb) file omni
      // (gdb) shell ps -efl | grep omni          to get the <pid>.
      // (gdb) attach <pid>
      // (gdb) set fPause = 0
      // (gdb) c
      bool fPause = true;

      while (fPause)
      {
      }
   }

   /* Get local copy. */
   pszS2C = getenv ("PDC_SRV_TO_CLIENT");
   pszC2S = getenv ("PDC_CLIENT_TO_SRV");

   if (  !pszS2C
      || !pszC2S
      )
   {
      DebugOutput::getErrorStream () << argv[0] << " must have PDC_SRV_TO_CLIENT and PDC_CLIENT_TO_SRV set in the environment!" << endl;

      return 1;
   }

#ifndef RETAIL
   if (DebugOutput::shouldOutputOmniServer ()) DebugOutput::getErrorStream () << "server: PDC_SRV_TO_CLIENT=" << pszS2C << endl;
   if (DebugOutput::shouldOutputOmniServer ()) DebugOutput::getErrorStream () << "server: PDC_CLIENT_TO_SRV=" << pszC2S << endl;
#endif

   /* Open fifos. */
   if ((fdS2C = open (pszS2C, O_WRONLY)) < 0)
   {
      goto BUGOUT;
   }

   if ((fdC2S = open (pszC2S, O_RDONLY)) < 0)
   {
      goto BUGOUT;
   }

   pCmd = new PrinterCommand ("OmniServer");

#ifndef RETAIL
   if (DebugOutput::shouldOutputOmniServer ()) DebugOutput::getErrorStream () << "server: ready." << endl;
#endif

   while (fRun)
   {
      static char *pszErrorNoDevice      = "Device not created";
      static char *pszErrorUnexpectedANU = "Unexpected ack/nack/unsupported command";
      static char *pszErrorMissingCmd    = "Missing command parameter";

      if (!pCmd->readCommand (fdC2S))
         goto BUGOUT;

      eCommand = PDCCMD_NACK;

      switch (pCmd->getCommandType ())
      {
      case PDCCMD_ACK:
      case PDCCMD_NACK:
      case PDCCMD_UNSUPPORTED:
      {
         if (!sendResponse (fdS2C, PDCCMD_NACK, pszErrorUnexpectedANU))
            goto BUGOUT;
         break;
      }

      case PDCCMD_INITIALIZE_SESSION:
      {
         // Send the response
         if (!sendResponse (fdS2C, PDCCMD_ACK, PDC_VERSION))
            goto BUGOUT;
         break;
      }

      case PDCCMD_CLOSE_SESSION:
      {
         fRun = false;
         break;
      }

      case PDCMD_SET_TRANSLATABLE_LANGUAGE:
      {
         if (  pCmd->getCommandString ()
            && *pCmd->getCommandString ()
            )
         {
            int iLangIn = StringResource::nameToID (pCmd->getCommandString ());

            if (StringResource::LANGUAGE_UNKNOWN != iLangIn)
            {
               Enumeration *pEnum = StringResource::getLanguages ();

               while (pEnum->hasMoreElements ())
               {
                  int iLangSupported = (int)pEnum->nextElement ();

                  if (iLangIn == iLangSupported)
                  {
                     iCurrentLanguage = iLangIn;
                     eCommand         = PDCCMD_ACK;
                  }
               }

               delete pEnum;
            }
         }

         // Send the response
         if (!sendResponse (fdS2C, eCommand, (char *)0))
            goto BUGOUT;
         break;
      }

      case PDCMD_GET_TRANSLATABLE_LANGUAGE:
      {
         // Send the response
         if (!sendResponse (fdS2C, PDCCMD_ACK, StringResource::IDToName (iCurrentLanguage)))
            goto BUGOUT;
         break;
      }

      case PDCMD_QUERY_TRANSLATABLE_LANGUAGES:
      {
         Enumeration *pEnum        = StringResource::getLanguages ();
         int          cbLanguages  = 0;
         char        *pszLanguages = 0;

         while (pEnum->hasMoreElements ())
         {
            pEnum->nextElement ();

            if (cbLanguages)
               cbLanguages++;
            cbLanguages += 2;
         }

         delete pEnum;

         pszLanguages = (char *)calloc (1, cbLanguages);

         pEnum = StringResource::getLanguages ();

         while (pEnum->hasMoreElements ())
         {
            int iLangSupported = (int)pEnum->nextElement ();

            if (*pszLanguages)
               strcat (pszLanguages, " ");
            strcat (pszLanguages, StringResource::IDToName (iLangSupported));
         }

         delete pEnum;

         // Send the response
         if (!sendResponse (fdS2C, PDCCMD_ACK, pszLanguages))
         {
            free (pszLanguages);
            goto BUGOUT;
         }

         free (pszLanguages);
         break;
      }

      case PDCCMD_IS_CMD_SUPPORTED:
      {
         switch (pCmd->getCommandInt ())
         {
         case PDCCMD_ACK:
         case PDCCMD_NACK:
         case PDCCMD_UNSUPPORTED:

         case PDCCMD_INITIALIZE_SESSION:
         case PDCCMD_CLOSE_SESSION:
         case PDCCMD_IS_CMD_SUPPORTED:
         case PDCMD_SET_TRANSLATABLE_LANGUAGE:
         case PDCMD_GET_TRANSLATABLE_LANGUAGE:
         case PDCMD_QUERY_TRANSLATABLE_LANGUAGES:

         case PDCCMD_ENUM_SHORT_DEVICES:
         case PDCCMD_ENUM_LONG_DEVICES:
         case PDCCMD_SET_DEVICE_NAME:
         case PDCCMD_IS_VALID_DEVICE_NAME:
         case PDCCMD_GET_PDL_INFO:

         case PDCCMD_SET_JOB_PROPERTIES:
         case PDCCMD_QUERY_CURRENT_JOB_PROPERTIES:
         case PDCCMD_LIST_JOB_PROPERTY_KEYS:
         case PDCCMD_LIST_DEVICE_JOB_PROPERTY_KEYS:
         case PDCCMD_LIST_JOB_PROPERTY_KEY_VALUES:
         case PDCCMD_GET_JOB_PROPERTY:
         case PDCCMD_GET_JOB_PROPERTY_TYPE:
         case PDCMD_XLATE_JOB_PROPERTY_KEY_VALUE:

         case PDCCMD_NEW_DEVICE:
         case PDCCMD_SET_OUTPUT_STREAM:
         case PDCCMD_BEGIN_JOB:
         case PDCCMD_START_PAGE:
         case PDCCMD_END_PAGE:
         case PDCCMD_END_JOB:
         case PDCCMD_ABORT_JOB:

         case PDCCMD_MODE_IS_RENDERER:
         case PDCCMD_ATTACH_BUFFER1:
         case PDCCMD_ATTACH_BUFFER2:
         case PDCCMD_DETACH_BUFFER1:
         case PDCCMD_DETACH_BUFFER2:
         case PDCCMD_RASTERIZE:

         case PDCMD_IS_COLOR_PRINTER:
         case PDCMD_HAS_HARDWARE_COPY:

         case PDCCMD_GET_VERSION:
         case PDCCMD_GET_DRIVER_NAME:
         case PDCCMD_GET_DEVICE_NAME:
         case PDCCMD_GET_SHORT_NAME:

         case PDCCMD_QUERY_CURRENT_ORIENTATION:
         case PDCCMD_QUERY_CURRENT_DITHER_ID:
         case PDCCMD_QUERY_CURRENT_FORM:
         case PDCCMD_QUERY_CURRENT_TRAY:
         case PDCCMD_QUERY_CURRENT_MEDIA:
         case PDCCMD_QUERY_CURRENT_RESOLUTION:
         case PDCCMD_QUERY_CURRENT_PRINT_MODE:
         case PDCCMD_QUERY_CURRENT_GAMMA:

         case PDCCMD_HAS_CAPABILITY:
         case PDCCMD_HAS_RASTER_CAPABILITY:
         case PDCCMD_HAS_DEVICE_OPTION:
         case PDCCMD_HAS_DEVICE_OPTION2:

         case PDCCMD_IS_ORIENTATION_SUPPORTED:
         case PDCCMD_IS_FORM_SUPPORTED:
         case PDCCMD_IS_TRAY_SUPPORTED:
         case PDCCMD_IS_MEDIA_SUPPORTED:
         case PDCCMD_IS_RESOLUTION_SUPPORTED:
         case PDCCMD_IS_PRINT_MODE_SUPPORTED:

         case PDCCMD_ENUM_ORIENTATIONS:
         case PDCCMD_ENUM_DITHER_IDS:
         case PDCCMD_ENUM_FORMS:
         case PDCCMD_ENUM_TRAYS:
         case PDCCMD_ENUM_MEDIAS:
         case PDCCMD_ENUM_RESOLUTIONS:
         case PDCCMD_ENUM_PRINT_MODES:
         {
            if (!sendResponse (fdS2C, PDCCMD_ACK, (char *)0))
               goto BUGOUT;
            break;
         }

         case PDCCMD_WARNING:
         case PDCCMD_SET_PRINTER_PROPERTIES:
         case PDCCMD_QUERY_CURRENT_PRINTER_PROPERTIES:
         case PDCCMD_LIST_PRINTER_PROPERTY_KEYS:
         case PDCCMD_LIST_DEVICE_PRINTER_PROPERTY_KEYS:
         case PDCCMD_LIST_PRINTER_PROPERTY_KEY_VALUES:
         case PDCCMD_GET_PRINTER_PROPERTY:
         case PDCCMD_GET_PRINTER_PROPERTY_TYPE:
         case PDCMD_XLATE_PRINTER_PROPERTY_KEY_VALUE:
         case PDCMD_QUERY_INPUT_FORMATS:
         case PDCMD_SET_INPUT_FORMAT:
         case PDCCMD_PUSH_CURRENT_ORIENTATION:
         case PDCCMD_PUSH_CURRENT_DITHER_ID:
         case PDCCMD_PUSH_CURRENT_FORM:
         case PDCCMD_PUSH_CURRENT_TRAY:
         case PDCCMD_PUSH_CURRENT_MEDIA:
         case PDCCMD_PUSH_CURRENT_RESOLUTION:
         case PDCCMD_PUSH_CURRENT_PRINT_MODE:
         case PDCCMD_PUSH_CURRENT_GAMMA:
         case -1:
         {
            if (!sendResponse (fdS2C, PDCCMD_NACK, pszErrorMissingCmd))
               goto BUGOUT;
            break;
         }

         default:
         {
            if (!sendResponse (fdS2C, PDCCMD_UNSUPPORTED, (char *)0))
               goto BUGOUT;
            break;
         }
         }
      }

      case PDCCMD_SET_DEVICE_NAME:
      {
         if (  pCmd->getCommandString ()
            && *pCmd->getCommandString ()
            )
         {
            delete pDeviceRef;

            pDeviceRef = new DeviceReference (pCmd->getCommandString ());

            if (pDeviceRef->loadDevice ())
            {
               eCommand = PDCCMD_ACK;
            }
         }

         // Send the response
         if (!sendResponse (fdS2C, eCommand, (char *)0))
            goto BUGOUT;
         break;
      }

      case PDCCMD_IS_VALID_DEVICE_NAME:
      {
         if (  pCmd->getCommandString ()
            && *pCmd->getCommandString ()
            )
         {
            DeviceReference *pDeviceRefTmp;

            pDeviceRefTmp = new DeviceReference (pCmd->getCommandString ());

            if (pDeviceRefTmp->loadDevice ())
            {
               eCommand = PDCCMD_ACK;
            }

            delete pDeviceRefTmp;
         }

         // Send the response
         if (!sendResponse (fdS2C, eCommand, (char *)0))
            goto BUGOUT;
         break;
      }

      case PDCCMD_MODE_IS_RENDERER:
      {
         if (pCmd->getCommandInt ())
         {
            fAdvanced = true;
         }
         else
         {
            fAdvanced = false;
         }

         // Send the response
         if (!sendResponse (fdS2C, PDCCMD_ACK, (char *)0))
            goto BUGOUT;
         break;
      }

      case PDCCMD_SET_JOB_PROPERTIES:
      {
         if (  pCmd->getCommandString ()
            && *pCmd->getCommandString ()
            )
         {
            if (cbJobProperties < strlen (pCmd->getCommandString ()) + 1)
            {
               cbJobProperties  = strlen (pCmd->getCommandString ()) + 1;
               pszJobProperties = (char *)realloc (pszJobProperties, cbJobProperties);
            }

            if (pszJobProperties)
            {
               strcpy (pszJobProperties, pCmd->getCommandString ());
               eCommand              = PDCCMD_ACK;
               fNewJobPropertiesSent = true;

               DebugOutput::applyAllDebugOutput (pszJobProperties);
            }
         }

         // Send the response
         if (!sendResponse (fdS2C, eCommand, (char *)0))
            goto BUGOUT;
         break;
      }

      case PDCCMD_NEW_DEVICE:
      {
         pDeviceRef->instantiateDevice (pszJobProperties, fAdvanced);

         Device *pDevice = pDeviceRef->getDevice ();

         if (pDevice)
         {
            eCommand              = PDCCMD_ACK;
            fNewJobPropertiesSent = false;

            if (StringResource::LANGUAGE_DEFAULT != iCurrentLanguage)
            {
               pDevice->setLanguage (iCurrentLanguage);
            }
         }

         // Send the response
         if (!sendResponse (fdS2C, eCommand, (char *)0))
            goto BUGOUT;
         break;
      }

      case PDCCMD_GET_VERSION:
      case PDCCMD_GET_DRIVER_NAME:
      case PDCCMD_GET_DEVICE_NAME:
      case PDCCMD_QUERY_CURRENT_ORIENTATION:
      case PDCCMD_QUERY_CURRENT_DITHER_ID:
      case PDCCMD_QUERY_CURRENT_FORM:
      case PDCCMD_QUERY_CURRENT_TRAY:
      case PDCCMD_QUERY_CURRENT_MEDIA:
      case PDCCMD_QUERY_CURRENT_RESOLUTION:
      case PDCCMD_QUERY_CURRENT_PRINT_MODE:
      case PDCCMD_QUERY_CURRENT_GAMMA:
      case PDCCMD_GET_PDL_INFO:
      case PDCCMD_HAS_CAPABILITY:
      case PDCCMD_HAS_RASTER_CAPABILITY:
      case PDCCMD_HAS_DEVICE_OPTION:
      case PDCCMD_HAS_DEVICE_OPTION2:
      case PDCCMD_ENUM_ORIENTATIONS:
      case PDCCMD_ENUM_DITHER_IDS:
      case PDCCMD_ENUM_FORMS:
      case PDCCMD_ENUM_TRAYS:
      case PDCCMD_ENUM_MEDIAS:
      case PDCCMD_ENUM_RESOLUTIONS:
      case PDCCMD_ENUM_PRINT_MODES:
      case PDCCMD_IS_ORIENTATION_SUPPORTED:
      case PDCCMD_IS_FORM_SUPPORTED:
      case PDCCMD_IS_TRAY_SUPPORTED:
      case PDCCMD_IS_MEDIA_SUPPORTED:
      case PDCCMD_IS_RESOLUTION_SUPPORTED:
      case PDCCMD_IS_PRINT_MODE_SUPPORTED:
      case PDCCMD_BEGIN_JOB:
      case PDCCMD_START_PAGE:
      case PDCCMD_END_PAGE:
      case PDCCMD_END_JOB:
      case PDCCMD_ABORT_JOB:
      case PDCCMD_RASTERIZE:
      case PDCCMD_QUERY_CURRENT_JOB_PROPERTIES:
      case PDCCMD_LIST_JOB_PROPERTY_KEYS:
      case PDCCMD_LIST_DEVICE_JOB_PROPERTY_KEYS:
      case PDCCMD_LIST_JOB_PROPERTY_KEY_VALUES:
      case PDCCMD_GET_JOB_PROPERTY_TYPE:
      case PDCCMD_GET_JOB_PROPERTY:
      case PDCMD_XLATE_JOB_PROPERTY_KEY_VALUE:
      case PDCCMD_SET_OUTPUT_STREAM:
      case PDCCMD_SET_ERROR_STREAM:
      case PDCMD_IS_COLOR_PRINTER:
      case PDCMD_HAS_HARDWARE_COPY:
      {
         Device *pDevice = pDeviceRef->getDevice ();

         if (!pDevice)
         {
            // Send the response
            if (!sendResponse (fdS2C, PDCCMD_NACK, pszErrorNoDevice))
               goto BUGOUT;
            break;
         }

         switch (pCmd->getCommandType ())
         {
         case PDCCMD_GET_VERSION:
         {
            // Send the response
            if (!sendResponse (fdS2C, PDCCMD_ACK, pDevice->getVersion ()))
               goto BUGOUT;
            break;
         }

         case PDCCMD_GET_DRIVER_NAME:
         {
            // Send the response
            if (!sendResponse (fdS2C, PDCCMD_ACK, pDevice->getDriverName ()))
               goto BUGOUT;
            break;
         }

         case PDCCMD_GET_DEVICE_NAME:
         {
            // Send the response
            if (!sendResponse (fdS2C, PDCCMD_ACK, pDevice->getDeviceName ()))
               goto BUGOUT;
            break;
         }

         case PDCCMD_QUERY_CURRENT_ORIENTATION:
         {
            DeviceOrientation *pDO           = 0;
            DeviceOrientation *pDOTemp       = 0;
            char               achResult[128];     // @TBD

            pDO = pDevice->getCurrentOrientation ();

            if (  pCmd->getCommandString ()
               && *pCmd->getCommandString ()
               )
            {
               pDOTemp = pDO->create (pCmd->getCommandString ());

               if (pDOTemp)
               {
                  pDO = pDOTemp;
               }
               else
               {
                  DebugOutput::getErrorStream () << "Could not create " << pCmd->getCommandString () << " named orientation" << endl;
               }
            }

            if (pDO)
               eCommand = PDCCMD_ACK;

            if (PDCCMD_ACK == eCommand)
               sprintf (achResult,
                        "%s",
                        DeviceOrientation::IDToName (pDO->getID ()));
            else
               achResult[0] = '\0';

            delete pDOTemp;

            // Send the response
            if (!sendResponse (fdS2C, eCommand, achResult))
               goto BUGOUT;
            break;
         }

         case PDCCMD_QUERY_CURRENT_DITHER_ID:
         {
            // Send the response
            if (!sendResponse (fdS2C, PDCCMD_ACK, pDevice->getCurrentDitherID ()))
               goto BUGOUT;
            break;
         }

         case PDCCMD_QUERY_CURRENT_FORM:
         {
            DeviceForm       *pDF            = 0;
            DeviceForm       *pDFTemp        = 0;
            HardCopyCap      *pHCC           = 0;
            char              achResult[128];     // @TBD

            pDF = pDevice->getCurrentForm ();
            if (pDF)
               pHCC = pDF->getHardCopyCap ();

            if (  pCmd->getCommandString ()
               && *pCmd->getCommandString ()
               )
            {
               pDFTemp = pDF->create (pCmd->getCommandString ());

               if (pDFTemp)
               {
                  pDF  = pDFTemp;
                  pHCC = pDF->getHardCopyCap ();
               }
               else
               {
                  DebugOutput::getErrorStream () << "Could not create " << pCmd->getCommandString () << " named form" << endl;
               }
            }

            if (pHCC)
               eCommand = PDCCMD_ACK;

            if (PDCCMD_ACK == eCommand)
               sprintf (achResult,
                        "%s %d %d %d %d %d",
                        DeviceForm::IDToName (pDF->getID ()),
                        pDF->getCapabilities (),
                        pHCC->getLeftClip (),
                        pHCC->getTopClip (),
                        pHCC->getRightClip (),
                        pHCC->getBottomClip ());
            else
               achResult[0] = '\0';

            delete pDFTemp;

            // Send the response
            if (!sendResponse (fdS2C, eCommand, achResult))
               goto BUGOUT;
            break;
         }

         case PDCCMD_QUERY_CURRENT_TRAY:
         {
            DeviceTray       *pDT            = 0;
            DeviceTray       *pDTTemp        = 0;
            char              achResult[128];     // @TBD

            pDT  = pDevice->getCurrentTray ();

            if (  pCmd->getCommandString ()
               && *pCmd->getCommandString ()
               )
            {
               pDTTemp = pDT->create (pCmd->getCommandString ());

               if (pDTTemp)
               {
                  pDT  = pDTTemp;
               }
               else
               {
                  DebugOutput::getErrorStream () << "Could not create " << pCmd->getCommandString () << " named tray" << endl;
               }
            }
            if (pDT)
               eCommand = PDCCMD_ACK;

            if (PDCCMD_ACK == eCommand)
               sprintf (achResult,
                        "%s %d",
                        DeviceTray::IDToName (pDT->getID ()),
                        pDT->getType ());
            else
               achResult[0] = '\0';

            delete pDTTemp;

            // Send the response
            if (!sendResponse (fdS2C, eCommand, achResult))
               goto BUGOUT;
            break;
         }

         case PDCCMD_QUERY_CURRENT_MEDIA:
         {
            DeviceMedia       *pDM            = 0;
            DeviceMedia       *pDMTemp        = 0;
            char               achResult[128];     // @TBD

            pDM  = pDevice->getCurrentMedia ();

            if (  pCmd->getCommandString ()
               && *pCmd->getCommandString ()
               )
            {
               pDMTemp = pDM->create (pCmd->getCommandString ());

               if (pDMTemp)
               {
                  pDM = pDMTemp;
               }
               else
               {
                  DebugOutput::getErrorStream () << "Could not create " << pCmd->getCommandString () << " named media" << endl;
               }
            }
            if (pDM)
               eCommand = PDCCMD_ACK;

            if (PDCCMD_ACK == eCommand)
               sprintf (achResult,
                        "%s %d %d",
                        DeviceMedia::IDToName (pDM->getID ()),
                        pDM->getColorAdjustRequired (),
                        pDM->getAbsorption ());
            else
               achResult[0] = '\0';

            delete pDMTemp;

            // Send the response
            if (!sendResponse (fdS2C, eCommand, achResult))
               goto BUGOUT;
            break;
         }

         case PDCCMD_QUERY_CURRENT_RESOLUTION:
         {
            DeviceResolution *pDR            = 0;
            DeviceResolution *pDRTemp        = 0;
            char              achResult[128];     // @TBD

            pDR = pDevice->getCurrentResolution ();

            if (  pCmd->getCommandString ()
               && *pCmd->getCommandString ()
               )
            {
               pDRTemp = pDR->create (pCmd->getCommandString ());

               if (pDRTemp)
               {
                  pDR = pDRTemp;
               }
               else
               {
                  DebugOutput::getErrorStream () << "Could not create " << pCmd->getCommandString () << " named resolution" << endl;
               }
            }
            if (pDR)
               eCommand = PDCCMD_ACK;

            if (PDCCMD_ACK == eCommand)
               sprintf (achResult,
                        "%s %d %d %d %d %d %d %d",
                        DeviceResolution::IDToName (pDR->getID ()),
                        pDR->getExternalXRes (),
                        pDR->getExternalYRes (),
                        pDR->getInternalXRes (),
                        pDR->getInternalYRes (),
                        pDR->getCapabilities (),
                        pDR->getDstBitsPerPel (),
                        pDR->getScanlineMultiple ());
            else
               achResult[0] = '\0';

            delete pDRTemp;

            // Send the response
            if (!sendResponse (fdS2C, eCommand, achResult))
               goto BUGOUT;
            break;
         }

         case PDCCMD_QUERY_CURRENT_PRINT_MODE:
         {
            DevicePrintMode *pDPM           = 0;
            DevicePrintMode *pDPMTemp       = 0;
            char             achResult[128];     // @TBD

            pDPM = pDevice->getCurrentPrintMode ();

            if (  pCmd->getCommandString ()
               && *pCmd->getCommandString ()
               )
            {
               pDPMTemp = pDPM->create (pCmd->getCommandString ());

               if (pDPMTemp)
               {
                  pDPM = pDPMTemp;
               }
               else
               {
                  DebugOutput::getErrorStream () << "Could not create " << pCmd->getCommandString () << " named print mode" << endl;
               }
            }

            if (pDPM)
               eCommand = PDCCMD_ACK;

            if (PDCCMD_ACK == eCommand)
               sprintf (achResult,
                        "%s %d %d %d",
                        DevicePrintMode::IDToName (pDPM->getID ()),
                        pDPM->getPhysicalCount (),
                        pDPM->getLogicalCount (),
                        pDPM->getNumPlanes ());
            else
               achResult[0] = '\0';

            delete pDPMTemp;

            // Send the response
            if (!sendResponse (fdS2C, eCommand, achResult))
               goto BUGOUT;
            break;
         }

         case PDCCMD_QUERY_CURRENT_GAMMA:
         {
            DeviceGamma *pDG            = 0;
            char         achResult[128];     // @TBD

            pDG = pDevice->getCurrentGamma ();

            if (pDG)
               eCommand = PDCCMD_ACK;

            if (PDCCMD_ACK == eCommand)
               sprintf (achResult,
                        "%d %d %d %d %d %d %d %d",
                        pDG->getCGamma (),
                        pDG->getMGamma (),
                        pDG->getYGamma (),
                        pDG->getKGamma (),
                        pDG->getCBias (),
                        pDG->getMBias (),
                        pDG->getYBias (),
                        pDG->getKBias ());
            else
               achResult[0] = '\0';

            // Send the response
            if (!sendResponse (fdS2C, eCommand, achResult))
               goto BUGOUT;
            break;
         }

         case PDCCMD_GET_PDL_INFO:
         {
            char achResponse[64]; // @TBD

            sprintf (achResponse,
                     "%d %d %d %d",
                     pDevice->getPDLLevel (),
                     pDevice->getPDLSubLevel (),
                     pDevice->getPDLMajorRevisionLevel (),
                     pDevice->getPDLMinorRevisionLevel ());

            // Send the response
            if (!sendResponse (fdS2C, PDCCMD_ACK, achResponse))
               goto BUGOUT;
            break;
         }

         case PDCCMD_HAS_CAPABILITY:
         {
            // Send the response
            if (!sendResponse (fdS2C, PDCCMD_ACK, pDevice->hasCapability (pCmd->getCommandLong ())))
               goto BUGOUT;
            break;
         }

         case PDCCMD_HAS_RASTER_CAPABILITY:
         {
            // Send the response
            if (!sendResponse (fdS2C, PDCCMD_ACK, pDevice->hasRasterCapability (pCmd->getCommandLong ())))
               goto BUGOUT;
            break;
         }

         case PDCCMD_HAS_DEVICE_OPTION:
         {
            // Send the response
            if (!sendResponse (fdS2C, PDCCMD_ACK, pDevice->hasDeviceOption (pCmd->getCommandLong ())))
               goto BUGOUT;
            break;
         }

         case PDCCMD_HAS_DEVICE_OPTION2:
         {
            // Send the response
            if (!sendResponse (fdS2C, PDCCMD_ACK, pDevice->hasDeviceOption2 (pCmd->getCommandLong ())))
               goto BUGOUT;
            break;
         }

         case PDCCMD_ENUM_ORIENTATIONS:
         {
            DeviceOrientation  *pDO         = pDevice->getCurrentOrientation ();
            Enumeration        *pEnum       = pDO->getEnumeration ();
            char               *pszResponse = 0;
            int                 cbResponse  = 0;

            while (pEnum->hasMoreElements ())
            {
               pDO = (DeviceOrientation *)pEnum->nextElement ();

               if (0 < cbResponse)
                  cbResponse++;

               cbResponse += strlen (DeviceOrientation::IDToName (pDO->getID ()));

               delete pDO;
            }

            delete pEnum;

            if (0 < cbResponse)
            {
               pszResponse = (char *)malloc (cbResponse + 1);

               if (pszResponse)
               {
                  pDO   = pDevice->getCurrentOrientation ();
                  pEnum = pDO->getEnumeration ();

                  *pszResponse = '\0';
                  while (pEnum->hasMoreElements ())
                  {
                     pDO = (DeviceOrientation *)pEnum->nextElement ();

                     if (*pszResponse)
                        strcat (pszResponse, " ");

                     strcat (pszResponse, DeviceOrientation::IDToName (pDO->getID ()));

                     delete pDO;
                  }

                  delete pEnum;

                  eCommand = PDCCMD_ACK;
               }
            }

            // Send the response
            if (!sendResponse (fdS2C, eCommand, pszResponse))
               goto BUGOUT;

            if (pszResponse)
            {
               free (pszResponse);
            }
            break;
         }

         case PDCCMD_ENUM_DITHER_IDS:
         {
            Enumeration *pEnum       = DeviceDither::getEnumeration ();
            char        *pszResponse = 0;
            int          cbResponse  = 0;

            while (pEnum->hasMoreElements ())
            {
               char *pszID = (char *)pEnum->nextElement ();

               if (0 < cbResponse)
                  cbResponse++;

               cbResponse += strlen (pszID);
            }

            delete pEnum;

            if (0 < cbResponse)
            {
               pszResponse = (char *)malloc (cbResponse + 1);

               if (pszResponse)
               {
                  pEnum = DeviceDither::getEnumeration ();

                  *pszResponse = '\0';
                  while (pEnum->hasMoreElements ())
                  {
                     char *pszID = (char *)pEnum->nextElement ();

                     if (*pszResponse)
                        strcat (pszResponse, " ");

                     strcat (pszResponse, pszID);
                  }

                  delete pEnum;

                  eCommand = PDCCMD_ACK;
               }
            }

            // Send the response
            if (!sendResponse (fdS2C, eCommand, pszResponse))
               goto BUGOUT;

            if (pszResponse)
            {
               free (pszResponse);
            }
            break;
         }

         case PDCCMD_ENUM_FORMS:
         {
            DeviceForm  *pDF         = pDevice->getCurrentForm ();
            Enumeration *pEnum       = pDF->getEnumeration ();
            char        *pszResponse = 0;
            int          cbResponse  = 0;

            while (pEnum->hasMoreElements ())
            {
               pDF = (DeviceForm *)pEnum->nextElement ();

               if (0 < cbResponse)
                  cbResponse++;

               cbResponse += strlen (DeviceForm::IDToName (pDF->getID ()));

               delete pDF;
            }

            delete pEnum;

            if (0 < cbResponse)
            {
               pszResponse = (char *)malloc (cbResponse + 1);

               if (pszResponse)
               {
                  pDF   = pDevice->getCurrentForm ();
                  pEnum = pDF->getEnumeration ();

                  *pszResponse = '\0';
                  while (pEnum->hasMoreElements ())
                  {
                     pDF = (DeviceForm *)pEnum->nextElement ();

                     if (*pszResponse)
                        strcat (pszResponse, " ");

                     strcat (pszResponse, DeviceForm::IDToName (pDF->getID ()));

                     delete pDF;
                  }

                  delete pEnum;

                  eCommand = PDCCMD_ACK;
               }
            }

            // Send the response
            if (!sendResponse (fdS2C, eCommand, pszResponse))
               goto BUGOUT;

            if (pszResponse)
            {
               free (pszResponse);
            }
            break;
         }

         case PDCCMD_ENUM_TRAYS:
         {
            DeviceTray  *pDT         = pDevice->getCurrentTray ();
            Enumeration *pEnum       = pDT->getEnumeration ();
            char        *pszResponse = 0;
            int          cbResponse  = 0;

            while (pEnum->hasMoreElements ())
            {
               pDT = (DeviceTray *)pEnum->nextElement ();

               if (0 < cbResponse)
                  cbResponse++;

               cbResponse += strlen (DeviceTray::IDToName (pDT->getID ()));

               delete pDT;
            }

            delete pEnum;

            if (0 < cbResponse)
            {
               pszResponse = (char *)malloc (cbResponse + 1);

               if (pszResponse)
               {
                  pDT   = pDevice->getCurrentTray ();
                  pEnum = pDT->getEnumeration ();

                  *pszResponse = '\0';
                  while (pEnum->hasMoreElements ())
                  {
                     pDT = (DeviceTray *)pEnum->nextElement ();

                     if (*pszResponse)
                        strcat (pszResponse, " ");

                     strcat (pszResponse, DeviceTray::IDToName (pDT->getID ()));

                     delete pDT;
                  }

                  delete pEnum;

                  eCommand = PDCCMD_ACK;
               }
            }

            // Send the response
            if (!sendResponse (fdS2C, eCommand, pszResponse))
               goto BUGOUT;

            if (pszResponse)
            {
               free (pszResponse);
            }
            break;
         }

         case PDCCMD_ENUM_MEDIAS:
         {
            DeviceMedia *pDM         = pDevice->getCurrentMedia ();
            Enumeration *pEnum       = pDM->getEnumeration ();
            char        *pszResponse = 0;
            int          cbResponse  = 0;

            while (pEnum->hasMoreElements ())
            {
               pDM = (DeviceMedia *)pEnum->nextElement ();

               if (0 < cbResponse)
                  cbResponse++;

               cbResponse += strlen (DeviceMedia::IDToName (pDM->getID ()));

               delete pDM;
            }

            delete pEnum;

            if (0 < cbResponse)
            {
               pszResponse = (char *)malloc (cbResponse + 1);

               if (pszResponse)
               {
                  pDM   = pDevice->getCurrentMedia ();
                  pEnum = pDM->getEnumeration ();

                  *pszResponse = '\0';
                  while (pEnum->hasMoreElements ())
                  {
                     pDM = (DeviceMedia *)pEnum->nextElement ();

                     if (*pszResponse)
                        strcat (pszResponse, " ");

                     strcat (pszResponse, DeviceMedia::IDToName (pDM->getID ()));

                     delete pDM;
                  }

                  delete pEnum;

                  eCommand = PDCCMD_ACK;
               }
            }

            // Send the response
            if (!sendResponse (fdS2C, eCommand, pszResponse))
               goto BUGOUT;

            if (pszResponse)
            {
               free (pszResponse);
            }
            break;
         }

         case PDCCMD_ENUM_RESOLUTIONS:
         {
            DeviceResolution *pDR         = pDevice->getCurrentResolution ();
            Enumeration      *pEnum       = pDR->getEnumeration ();
            char             *pszResponse = 0;
            int               cbResponse  = 0;

            while (pEnum->hasMoreElements ())
            {
               pDR = (DeviceResolution *)pEnum->nextElement ();

               if (0 < cbResponse)
                  cbResponse++;

               cbResponse += strlen (DeviceResolution::IDToName (pDR->getID ()));

               delete pDR;
            }

            delete pEnum;

            if (0 < cbResponse)
            {
               pszResponse = (char *)malloc (cbResponse + 1);

               if (pszResponse)
               {
                  pDR   = pDevice->getCurrentResolution ();
                  pEnum = pDR->getEnumeration ();

                  *pszResponse = '\0';
                  while (pEnum->hasMoreElements ())
                  {
                     pDR = (DeviceResolution *)pEnum->nextElement ();

                     if (*pszResponse)
                        strcat (pszResponse, " ");

                     strcat (pszResponse, DeviceResolution::IDToName (pDR->getID ()));

                     delete pDR;
                  }

                  delete pEnum;

                  eCommand = PDCCMD_ACK;
               }
            }

            // Send the response
            if (!sendResponse (fdS2C, eCommand, pszResponse))
               goto BUGOUT;

            if (pszResponse)
            {
               free (pszResponse);
            }
            break;
         }

         case PDCCMD_ENUM_PRINT_MODES:
         {
            DevicePrintMode *pDPM        = pDevice->getCurrentPrintMode ();
            Enumeration     *pEnum       = pDPM->getEnumeration ();
            char            *pszResponse = 0;
            int              cbResponse  = 0;

            while (pEnum->hasMoreElements ())
            {
               pDPM = (DevicePrintMode *)pEnum->nextElement ();

               if (0 < cbResponse)
                  cbResponse++;

               cbResponse += strlen (DevicePrintMode::IDToName (pDPM->getID ()));

               delete pDPM;
            }

            delete pEnum;

            if (0 < cbResponse)
            {
               pszResponse = (char *)malloc (cbResponse + 1);

               if (pszResponse)
               {
                  pDPM   = pDevice->getCurrentPrintMode ();
                  pEnum = pDPM->getEnumeration ();

                  *pszResponse = '\0';
                  while (pEnum->hasMoreElements ())
                  {
                     pDPM = (DevicePrintMode *)pEnum->nextElement ();

                     if (*pszResponse)
                        strcat (pszResponse, " ");

                     strcat (pszResponse, DevicePrintMode::IDToName (pDPM->getID ()));

                     delete pDPM;
                  }

                  delete pEnum;

                  eCommand = PDCCMD_ACK;
               }
            }

            // Send the response
            if (!sendResponse (fdS2C, eCommand, pszResponse))
               goto BUGOUT;

            if (pszResponse)
            {
               free (pszResponse);
            }
            break;
         }

         case PDCCMD_IS_ORIENTATION_SUPPORTED:
         {
            DeviceOrientation *pDO = pDevice->getCurrentOrientation ();

            if (pDO->isOrientationSupported (pCmd->getCommandInt ()))
            {
               eCommand = PDCCMD_ACK;
            }

            // Send the response
            if (!sendResponse (fdS2C, eCommand, (char *)0))
               goto BUGOUT;
            break;
         }

         case PDCCMD_IS_FORM_SUPPORTED:
         {
            DeviceForm *pDF = pDevice->getCurrentForm ();

            if (pDF->isFormSupported (pCmd->getCommandInt ()))
            {
               eCommand = PDCCMD_ACK;
            }

            // Send the response
            if (!sendResponse (fdS2C, eCommand, (char *)0))
               goto BUGOUT;
            break;
         }

         case PDCCMD_IS_TRAY_SUPPORTED:
         {
            DeviceTray *pDT = pDevice->getCurrentTray ();

            if (pDT->isTraySupported (pCmd->getCommandInt ()))
            {
               eCommand = PDCCMD_ACK;
            }

            // Send the response
            if (!sendResponse (fdS2C, eCommand, (char *)0))
               goto BUGOUT;
            break;
         }

         case PDCCMD_IS_MEDIA_SUPPORTED:
         {
            DeviceMedia *pDM = pDevice->getCurrentMedia ();

            if (pDM->isMediaSupported (pCmd->getCommandInt ()))
            {
               eCommand = PDCCMD_ACK;
            }

            // Send the response
            if (!sendResponse (fdS2C, eCommand, (char *)0))
               goto BUGOUT;
            break;
         }

         case PDCCMD_IS_RESOLUTION_SUPPORTED:
         {
            DeviceResolution *pDR = pDevice->getCurrentResolution ();

            if (pDR->isResolutionSupported (pCmd->getCommandInt ()))
            {
               eCommand = PDCCMD_ACK;
            }

            // Send the response
            if (!sendResponse (fdS2C, eCommand, (char *)0))
               goto BUGOUT;
            break;
         }

         case PDCCMD_IS_PRINT_MODE_SUPPORTED:
         {
            DevicePrintMode *pDPM = pDevice->getCurrentPrintMode ();

            if (pDPM->isPrintModeSupported (pCmd->getCommandInt ()))
            {
               eCommand = PDCCMD_ACK;
            }

            // Send the response
            if (!sendResponse (fdS2C, eCommand, (char *)0))
               goto BUGOUT;
            break;
         }

         case PDCCMD_BEGIN_JOB:
         {
            if (pDevice->beginJob ())
            {
               iPageNumber = 1;
               eCommand    = PDCCMD_ACK;
            }

            // Send the response
            if (!sendResponse (fdS2C, eCommand, (char *)0))
               goto BUGOUT;
            break;
         }

         case PDCCMD_START_PAGE:
         {
            if (!sendResponse (fdS2C, PDCCMD_ACK, (char *)0))
               goto BUGOUT;
            break;
         }

         case PDCCMD_END_PAGE:
         {
            if (1 == iPageNumber)
            {
               eCommand = PDCCMD_ACK;
            }
            else
            {
               if (fNewJobPropertiesSent)
               {
                  if (pDevice->newFrame (pszJobProperties))
                     eCommand = PDCCMD_ACK;
               }
               else
               {
                  if (pDevice->newFrame ())
                     eCommand = PDCCMD_ACK;
               }
            }

            if (PDCCMD_ACK == eCommand)
            {
               iPageNumber++;
            }

            // Send the response
            if (!sendResponse (fdS2C, eCommand, (char *)0))
               goto BUGOUT;
            break;
         }

         case PDCCMD_END_JOB:
         {
            if (pDevice->endJob ())
               eCommand = PDCCMD_ACK;

            // Send the response
            if (!sendResponse (fdS2C, eCommand, (char *)0))
               goto BUGOUT;
            break;
         }

         case PDCCMD_ABORT_JOB:
         {
            if (pDevice->abortJob ())
               eCommand = PDCCMD_ACK;

            // Send the response
            if (!sendResponse (fdS2C, eCommand, (char *)0))
               goto BUGOUT;
            break;
         }

         case PDCCMD_RASTERIZE:
         {
            if (  pCmd->getCommandString ()
               && *pCmd->getCommandString ()
               )
            {
               int          iType;
               BITBLT_TYPE  eType;
               RECTL        rectlPageLocation;

               sscanf (pCmd->getCommandString (),
                       "%d %d %d %d %d",
                       &iType,
                       &rectlPageLocation.xLeft,
                       &rectlPageLocation.yBottom,
                       &rectlPageLocation.xRight,
                       &rectlPageLocation.yTop);

               eType = (BITBLT_TYPE)iType;

               if (pDevice->rasterize (pbBuffer2,
                                       (PBITMAPINFO2)pbBuffer1,
                                       &rectlPageLocation,
                                       eType))
               {
                  eCommand = PDCCMD_ACK;
               }
            }

            // Send the response
            if (!sendResponse (fdS2C, eCommand, (char *)0))
               goto BUGOUT;
            break;
         }

         case PDCCMD_QUERY_CURRENT_JOB_PROPERTIES:
         {
            string *pStr   = 0;
            char   *pszStr = 0;

            pStr = pDevice->queryCurrentJobProperties ();

            if (pStr)
            {
               pszStr   = (char *)pStr->c_str ();
               eCommand = PDCCMD_ACK;
            }

            if (!sendResponse (fdS2C, eCommand, pszStr))
            {
               delete pStr;

               goto BUGOUT;
            }

            delete pStr;
            break;
         }

         case PDCCMD_LIST_JOB_PROPERTY_KEYS:
         {
            Enumeration *pEnum       = pDevice->listJobPropertyKeys ();
            char        *pszResponse = 0;
            int          cbResponse  = 0;

            while (pEnum->hasMoreElements ())
            {
               char *pszKey = (char *)pEnum->nextElement ();

               if (0 < cbResponse)
                  cbResponse++;

               cbResponse += strlen (pszKey);
            }

            delete pEnum;

            if (0 < cbResponse)
            {
               pszResponse = (char *)malloc (cbResponse + 1);

               if (pszResponse)
               {
                  pEnum = pDevice->listJobPropertyKeys ();

                  *pszResponse = '\0';
                  while (pEnum->hasMoreElements ())
                  {
                     char *pszKey = (char *)pEnum->nextElement ();

                     if (*pszResponse)
                        strcat (pszResponse, " ");

                     strcat (pszResponse, pszKey);
                  }

                  delete pEnum;
               }
            }

            // Send the response
            if (!sendResponse (fdS2C, PDCCMD_ACK, pszResponse))
               goto BUGOUT;

            if (pszResponse)
            {
               free (pszResponse);
            }
            break;
         }

         case PDCCMD_LIST_DEVICE_JOB_PROPERTY_KEYS:
         {
            Enumeration *pEnum       = pDevice->listDeviceJobPropertyKeys ();
            char        *pszResponse = 0;
            int          cbResponse  = 0;

            while (pEnum->hasMoreElements ())
            {
               char *pszKey = (char *)pEnum->nextElement ();

               if (0 < cbResponse)
                  cbResponse++;

               cbResponse += strlen (pszKey);
            }

            delete pEnum;

            if (0 < cbResponse)
            {
               pszResponse = (char *)malloc (cbResponse + 1);

               if (pszResponse)
               {
                  pEnum = pDevice->listDeviceJobPropertyKeys ();

                  *pszResponse = '\0';
                  while (pEnum->hasMoreElements ())
                  {
                     char *pszKey = (char *)pEnum->nextElement ();

                     if (*pszResponse)
                        strcat (pszResponse, " ");

                     strcat (pszResponse, pszKey);
                  }

                  delete pEnum;
               }
            }

            // Send the response
            if (!sendResponse (fdS2C, PDCCMD_ACK, pszResponse))
               goto BUGOUT;

            if (pszResponse)
            {
               free (pszResponse);
            }
            break;
         }

         case PDCCMD_LIST_JOB_PROPERTY_KEY_VALUES:
         {
            char *pszResponse = 0;

            if (  pCmd->getCommandString ()
               && *pCmd->getCommandString ()
               )
            {
               char        *pszKey      = pCmd->getCommandString ();
               Enumeration *pEnum       = pDevice->listKeyValues (pszKey);
               int          cbResponse  = 0;

               while (pEnum->hasMoreElements ())
               {
                  char *pszKey = (char *)pEnum->nextElement ();

                  if (0 < cbResponse)
                     cbResponse++;

                  cbResponse += strlen (pszKey);
               }

               delete pEnum;

               if (0 < cbResponse)
               {
                  pszResponse = (char *)malloc (cbResponse + 1);

                  if (pszResponse)
                  {
                     pEnum = pDevice->listKeyValues (pszKey);

                     *pszResponse = '\0';
                     while (pEnum->hasMoreElements ())
                     {
                        char *pszKey = (char *)pEnum->nextElement ();

                        if (*pszResponse)
                           strcat (pszResponse, " ");

                        strcat (pszResponse, pszKey);
                     }

                     delete pEnum;
                  }
               }

               eCommand = PDCCMD_ACK;
            }

            // Send the response
            if (!sendResponse (fdS2C, eCommand, pszResponse))
               goto BUGOUT;

            if (pszResponse)
            {
               free (pszResponse);
            }
            break;
         }

         case PDCCMD_GET_JOB_PROPERTY_TYPE:
         {
            string *pStr   = 0;
            char   *pszStr = 0;

            if (  pCmd->getCommandString ()
               && *pCmd->getCommandString ()
               )
            {
               pStr = pDevice->getJobPropertyType (pCmd->getCommandString ());

               if (pStr)
               {
                  pszStr   = (char *)pStr->c_str ();
                  eCommand = PDCCMD_ACK;
               }
            }

            if (!sendResponse (fdS2C, eCommand, pszStr))
            {
               delete pStr;

               goto BUGOUT;
            }

            delete pStr;
            break;
         }

         case PDCCMD_GET_JOB_PROPERTY:
         {
            string *pStr   = 0;
            char   *pszStr = 0;

            if (  pCmd->getCommandString ()
               && *pCmd->getCommandString ()
               )
            {
               pStr = pDevice->getJobProperty (pCmd->getCommandString ());

               if (pStr)
               {
                  pszStr   = (char *)pStr->c_str ();
                  eCommand = PDCCMD_ACK;
               }
            }

            if (!sendResponse (fdS2C, eCommand, pszStr))
            {
               delete pStr;

               goto BUGOUT;
            }

            delete pStr;
            break;
         }

         case PDCMD_XLATE_JOB_PROPERTY_KEY_VALUE:
         {
            string *pStr   = 0;
            char   *pszStr = 0;

            if (  pCmd->getCommandString ()
               && *pCmd->getCommandString ()
               )
            {
               char *pszKey    = pCmd->getCommandString ();
               char *pszValue  = 0;
               char *pszEquals = strchr (pszKey, '=');

               if (pszEquals)
               {
                  *pszEquals = '\0';
                  pszValue = pszEquals + 1;

                  pStr = pDevice->translateKeyValue (pszKey,
                                                     pszValue);

                  if (pStr)
                  {
                     pszStr   = (char *)pStr->c_str ();
                     eCommand = PDCCMD_ACK;
                  }
               }
            }

            if (!sendResponse (fdS2C, eCommand, pszStr))
            {
               delete pStr;

               goto BUGOUT;
            }

            delete pStr;
            break;
         }

         case PDCCMD_SET_OUTPUT_STREAM:
         {
            if (  pCmd->getCommandString ()
               && *pCmd->getCommandString ()
               )
            {
               int fdHandle = STDOUT_FILENO;

               sscanf (pCmd->getCommandString (), "%d", &fdHandle);

               if (pfpOut)
               {
                  fclose (pfpOut);
                  pfpOut = 0;
               }

               pfpOut = fdopen (fdHandle, "wb");

               if (pfpOut)
               {
                  Omni::setOutputStream (pDevice, pfpOut);

                  eCommand = PDCCMD_ACK;
               }
            }

            // Send the response
            if (!sendResponse (fdS2C, eCommand, (char *)0))
               goto BUGOUT;
         }

         case PDCCMD_SET_ERROR_STREAM:
         {
            if (  pCmd->getCommandString ()
               && *pCmd->getCommandString ()
               )
            {
               int fdHandle = STDERR_FILENO;

               sscanf (pCmd->getCommandString (), "%d", &fdHandle);

               if (pfpErr)
               {
                  fclose (pfpErr);
                  pfpErr = 0;
               }

               pfpErr = fdopen (fdHandle, "wb");

               if (pfpErr)
               {
                  Omni::setErrorStream (pDevice, pfpErr);

                  eCommand = PDCCMD_ACK;
               }
            }

            // Send the response
            if (!sendResponse (fdS2C, eCommand, (char *)0))
               goto BUGOUT;
         }

         case PDCMD_IS_COLOR_PRINTER:
         {
            char *pszResponse = 0;

            eCommand = PDCCMD_ACK;

            if (pDevice->hasCapability (Capability::COLOR))
            {
               pszResponse = "1";
            }
            else
            {
               pszResponse = "0";
            }

            if (!sendResponse (fdS2C, eCommand, pszResponse))
            {
               goto BUGOUT;
            }
            break;
         }

         case PDCMD_HAS_HARDWARE_COPY:
         {
            char *pszResponse = 0;

            eCommand = PDCCMD_ACK;

            if (pDevice->hasCapability (Capability::HARDWARE_COPIES))
            {
               pszResponse = "1";
            }
            else
            {
               pszResponse = "0";
            }

            if (!sendResponse (fdS2C, eCommand, pszResponse))
            {
               goto BUGOUT;
            }
            break;
         }
         }
         break;
      }

      case PDCCMD_ATTACH_BUFFER1:
      {
         if (pbBuffer1)
         {
            shmdt (pbBuffer1);
         }

         pbBuffer1 = (byte *)shmat (pCmd->getCommandInt (), 0, 0);

         if (0 < pbBuffer1)
         {
            eCommand = PDCCMD_ACK;
         }

         if (!sendResponse (fdS2C, eCommand, (char *)0))
            goto BUGOUT;
         break;
      }

      case PDCCMD_ATTACH_BUFFER2:
      {
         if (pbBuffer2)
         {
            shmdt (pbBuffer2);
         }

         pbBuffer2 = (byte *)shmat (pCmd->getCommandInt (), 0, 0);

         if (0 < pbBuffer2)
         {
            eCommand = PDCCMD_ACK;
         }

         if (!sendResponse (fdS2C, eCommand, (char *)0))
            goto BUGOUT;
         break;
      }

      case PDCCMD_DETACH_BUFFER1:
      {
         if (pbBuffer1)
         {
            int rc = shmdt (pbBuffer1);

            if (0 == rc)
            {
               eCommand = PDCCMD_ACK;
            }
         }

         if (!sendResponse (fdS2C, eCommand, (char *)0))
            goto BUGOUT;
         break;
      }

      case PDCCMD_DETACH_BUFFER2:
      {
         if (pbBuffer2)
         {
            int rc = shmdt (pbBuffer2);

            if (0 == rc)
            {
               eCommand = PDCCMD_ACK;
            }
         }

         if (!sendResponse (fdS2C, eCommand, (char *)0))
            goto BUGOUT;
         break;
      }

      case PDCCMD_ENUM_SHORT_DEVICES:
      {
         Enumeration *pEnum                           = Omni::listDevices ();
         int          cbSize                          = 0;
         char        *pszMatch                        = 0;
         bool         fMustMatchPDLLevel              = false;
         int          iPDLLevel                       = PDL::PDL_DONTCARE;
         bool         fMustMatchPDLSubLevel           = false;
         int          iPDLSubLevel                    = PDL::PDL_DONTCARE;
         bool         fMustMatchPDLMajorRevisionLevel = false;
         int          iPDLMajorRevisionLevel          = PDL::PDL_DONTCARE;
         bool         fMustMatchPDLMinorRevisionLevel = false;
         int          iPDLMinorRevisionLevel          = PDL::PDL_DONTCARE;

         if (!pEnum)
         {
            if (!sendResponse (fdS2C, PDCCMD_NACK, (char *)0))
               goto BUGOUT;
            break;
         }

         pszMatch = pCmd->getCommandString ();
         if (  pszMatch
            && *pszMatch
            )
         {
            if (1 == sscanf (pszMatch, "%d", &iPDLLevel))
            {
               fMustMatchPDLLevel = true;
               pszMatch = strchr (pszMatch, ' ');

               if (pszMatch)
               {
                  if (1 == sscanf (pszMatch + 1, "%d", &iPDLSubLevel))
                  {
                     fMustMatchPDLSubLevel = true;
                     pszMatch = strchr (pszMatch + 1, ' ');

                     if (pszMatch)
                     {
                        if (1 == sscanf (pszMatch + 1,
                                         "%d",
                                         &iPDLMajorRevisionLevel))
                        {
                           fMustMatchPDLMajorRevisionLevel = true;
                           pszMatch = strchr (pszMatch + 1, ' ');

                           if (pszMatch)
                           {
                              if (1 == sscanf (pszMatch + 1,
                                               "%d",
                                               &iPDLMinorRevisionLevel))
                              {
                                 fMustMatchPDLMinorRevisionLevel = true;
                                 pszMatch = strchr (pszMatch + 1, ' ');
                              }
                           }
                        }
                     }
                  }
               }
            }
         }

#ifndef RETAIL
         if (DebugOutput::shouldOutputOmniServer ())
         {
            DebugOutput::getErrorStream () << "fMustMatchPDLLevel              = " << fMustMatchPDLLevel << endl;
            DebugOutput::getErrorStream () << "iPDLLevel                       = " << iPDLLevel << endl;
            DebugOutput::getErrorStream () << "fMustMatchPDLSubLevel           = " << fMustMatchPDLSubLevel << endl;
            DebugOutput::getErrorStream () << "iPDLSubLevel                    = " << iPDLSubLevel << endl;
            DebugOutput::getErrorStream () << "fMustMatchPDLMajorRevisionLevel = " << fMustMatchPDLMajorRevisionLevel << endl;
            DebugOutput::getErrorStream () << "iPDLMajorRevisionLevel          = " << iPDLMajorRevisionLevel << endl;
            DebugOutput::getErrorStream () << "fMustMatchPDLMinorRevisionLevel = " << fMustMatchPDLMinorRevisionLevel << endl;
            DebugOutput::getErrorStream () << "iPDLMinorRevisionLevel          = " << iPDLMinorRevisionLevel << endl;
         }
#endif

         while (pEnum->hasMoreElements ())
         {
            char         *pszLibName   = (char *)pEnum->nextElement ();
            Device       *pDevice      = 0;
            PFNNEWDEVICE  pfnNewDevice = 0;
            GModule      *hmodDevice   = 0;

            hmodDevice = g_module_open (pszLibName, (GModuleFlags)0);

            if (hmodDevice)
            {
               g_module_symbol (hmodDevice, "newDeviceW_Advanced", (gpointer *)&pfnNewDevice);

               if (pfnNewDevice)
               {
                  pDevice = pfnNewDevice (true);

                  if (  (  (  fMustMatchPDLLevel
                           && iPDLLevel == pDevice->getPDLLevel ()
                           )
                        || !fMustMatchPDLLevel
                        )
                     && (  (  fMustMatchPDLSubLevel
                           && iPDLSubLevel == pDevice->getPDLSubLevel ()
                           )
                        || !fMustMatchPDLSubLevel
                        )
                     && (  (  fMustMatchPDLMajorRevisionLevel
                           && iPDLMajorRevisionLevel == pDevice->getPDLMajorRevisionLevel ()
                           )
                        || !fMustMatchPDLMajorRevisionLevel
                        )
                     && (  (  fMustMatchPDLMinorRevisionLevel
                           && iPDLMinorRevisionLevel == pDevice->getPDLMinorRevisionLevel ()
                           )
                        || !fMustMatchPDLMinorRevisionLevel
                        )
                     )
                  {
                     cbSize += strlen (pDevice->getShortName ())
                             + 1;                                 // end-of-string
                  }

                  delete pDevice;
               }

               g_module_close (hmodDevice);
            }
         }

         if (0 < cbSize)
            // add end-of-array
            cbSize++;

         delete pEnum;

         if (0 == cbSize)
         {
            if (!sendResponse (fdS2C, PDCCMD_NACK, (char *)0))
               goto BUGOUT;
            break;
         }

         char *pszResponse = 0;
         char *pszAppend   = 0;

         pszResponse = (char *)calloc (1, cbSize);

         if (!pszResponse)
         {
            if (!sendResponse (fdS2C, PDCCMD_NACK, (char *)0))
               goto BUGOUT;
            break;
         }

         pszAppend = pszResponse;
         pEnum     = Omni::listDevices ();

         while (pEnum->hasMoreElements ())
         {
            char         *pszLibName   = (char *)pEnum->nextElement ();
            Device       *pDevice      = 0;
            PFNNEWDEVICE  pfnNewDevice = 0;
            GModule      *hmodDevice   = 0;

            hmodDevice = g_module_open (pszLibName, (GModuleFlags)0);

            if (hmodDevice)
            {
               g_module_symbol (hmodDevice, "newDeviceW_Advanced", (gpointer *)&pfnNewDevice);

               if (pfnNewDevice)
               {
                  pDevice = pfnNewDevice (true);

                  if (  (  (  fMustMatchPDLLevel
                           && iPDLLevel == pDevice->getPDLLevel ()
                           )
                        || !fMustMatchPDLLevel
                        )
                     && (  (  fMustMatchPDLSubLevel
                           && iPDLSubLevel == pDevice->getPDLSubLevel ()
                           )
                        || !fMustMatchPDLSubLevel
                        )
                     && (  (  fMustMatchPDLMajorRevisionLevel
                           && iPDLMajorRevisionLevel == pDevice->getPDLMajorRevisionLevel ()
                           )
                        || !fMustMatchPDLMajorRevisionLevel
                        )
                     && (  (  fMustMatchPDLMinorRevisionLevel
                           && iPDLMinorRevisionLevel == pDevice->getPDLMinorRevisionLevel ()
                           )
                        || !fMustMatchPDLMinorRevisionLevel
                        )
                     )
                  {
                     strcat (pszAppend, pDevice->getShortName ());
                     pszAppend += strlen (pszAppend) + 1;
                  }

                  delete pDevice;
               }

               g_module_close (hmodDevice);
            }
         }

         *pszAppend++ = '\0'; // end-of-array

         delete pEnum;

         if (!sendResponse (fdS2C, PDCCMD_ACK, (char **)pszResponse))
         {
            free (pszResponse);
            goto BUGOUT;
         }

         free (pszResponse);

         break;
      }

      case PDCCMD_ENUM_LONG_DEVICES:
      {
         Enumeration *pEnum                           = Omni::listDevices ();
         int          cbSize                          = 0;
         char        *pszMatch                        = 0;
         bool         fMustMatchPDLLevel              = false;
         int          iPDLLevel                       = PDL::PDL_DONTCARE;
         bool         fMustMatchPDLSubLevel           = false;
         int          iPDLSubLevel                    = PDL::PDL_DONTCARE;
         bool         fMustMatchPDLMajorRevisionLevel = false;
         int          iPDLMajorRevisionLevel          = PDL::PDL_DONTCARE;
         bool         fMustMatchPDLMinorRevisionLevel = false;
         int          iPDLMinorRevisionLevel          = PDL::PDL_DONTCARE;

         if (!pEnum)
         {
            if (!sendResponse (fdS2C, PDCCMD_NACK, (char *)0))
               goto BUGOUT;
            break;
         }

         pszMatch = pCmd->getCommandString ();
         if (  pszMatch
            && *pszMatch
            )
         {
            if (1 == sscanf (pszMatch, "%d", &iPDLLevel))
            {
               fMustMatchPDLLevel = true;
               pszMatch = strchr (pszMatch, ' ');

               if (pszMatch)
               {
                  if (1 == sscanf (pszMatch + 1, "%d", &iPDLSubLevel))
                  {
                     fMustMatchPDLSubLevel = true;
                     pszMatch = strchr (pszMatch + 1, ' ');

                     if (pszMatch)
                     {
                        if (1 == sscanf (pszMatch + 1,
                                         "%d",
                                         &iPDLMajorRevisionLevel))
                        {
                           fMustMatchPDLMajorRevisionLevel = true;
                           pszMatch = strchr (pszMatch + 1, ' ');

                           if (pszMatch)
                           {
                              if (1 == sscanf (pszMatch + 1,
                                               "%d",
                                               &iPDLMinorRevisionLevel))
                              {
                                 fMustMatchPDLMinorRevisionLevel = true;
                                 pszMatch = strchr (pszMatch + 1, ' ');
                              }
                           }
                        }
                     }
                  }
               }
            }
         }

#ifndef RETAIL
         if (DebugOutput::shouldOutputOmniServer ())
         {
            DebugOutput::getErrorStream () << "fMustMatchPDLLevel              = " << fMustMatchPDLLevel << endl;
            DebugOutput::getErrorStream () << "iPDLLevel                       = " << iPDLLevel << endl;
            DebugOutput::getErrorStream () << "fMustMatchPDLSubLevel           = " << fMustMatchPDLSubLevel << endl;
            DebugOutput::getErrorStream () << "iPDLSubLevel                    = " << iPDLSubLevel << endl;
            DebugOutput::getErrorStream () << "fMustMatchPDLMajorRevisionLevel = " << fMustMatchPDLMajorRevisionLevel << endl;
            DebugOutput::getErrorStream () << "iPDLMajorRevisionLevel          = " << iPDLMajorRevisionLevel << endl;
            DebugOutput::getErrorStream () << "fMustMatchPDLMinorRevisionLevel = " << fMustMatchPDLMinorRevisionLevel << endl;
            DebugOutput::getErrorStream () << "iPDLMinorRevisionLevel          = " << iPDLMinorRevisionLevel << endl;
         }
#endif

         while (pEnum->hasMoreElements ())
         {
            char         *pszLibName   = (char *)pEnum->nextElement ();
            Device       *pDevice      = 0;
            PFNNEWDEVICE  pfnNewDevice = 0;
            GModule      *hmodDevice   = 0;

            hmodDevice = g_module_open (pszLibName, (GModuleFlags)0);

            if (hmodDevice)
            {
               g_module_symbol (hmodDevice, "newDeviceW_Advanced", (gpointer *)&pfnNewDevice);

               if (pfnNewDevice)
               {
                  pDevice = pfnNewDevice (true);

                  if (  (  (  fMustMatchPDLLevel
                           && iPDLLevel == pDevice->getPDLLevel ()
                           )
                        || !fMustMatchPDLLevel
                        )
                     && (  (  fMustMatchPDLSubLevel
                           && iPDLSubLevel == pDevice->getPDLSubLevel ()
                           )
                        || !fMustMatchPDLSubLevel
                        )
                     && (  (  fMustMatchPDLMajorRevisionLevel
                           && iPDLMajorRevisionLevel == pDevice->getPDLMajorRevisionLevel ()
                           )
                        || !fMustMatchPDLMajorRevisionLevel
                        )
                     && (  (  fMustMatchPDLMinorRevisionLevel
                           && iPDLMinorRevisionLevel == pDevice->getPDLMinorRevisionLevel ()
                           )
                        || !fMustMatchPDLMinorRevisionLevel
                        )
                     )
                  {
                     cbSize += strlen (pDevice->getDriverName ())
                             + 1                                  // '.'
                             + strlen (pDevice->getDeviceName ())
                             + 1;                                 // end-of-string
                  }

                  delete pDevice;
               }

               g_module_close (hmodDevice);
            }
         }

         if (0 < cbSize)
            // add end-of-array
            cbSize++;

         delete pEnum;

         if (0 == cbSize)
         {
            if (!sendResponse (fdS2C, PDCCMD_NACK, (char *)0))
               goto BUGOUT;
            break;
         }

         char *pszResponse = 0;
         char *pszAppend   = 0;

         pszResponse = (char *)calloc (1, cbSize);

         if (!pszResponse)
         {
            if (!sendResponse (fdS2C, PDCCMD_NACK, (char *)0))
               goto BUGOUT;
            break;
         }

         pszAppend = pszResponse;
         pEnum     = Omni::listDevices ();

         while (pEnum->hasMoreElements ())
         {
            char         *pszLibName   = (char *)pEnum->nextElement ();
            Device       *pDevice      = 0;
            PFNNEWDEVICE  pfnNewDevice = 0;
            GModule      *hmodDevice   = 0;

            hmodDevice = g_module_open (pszLibName, (GModuleFlags)0);

            if (hmodDevice)
            {
               g_module_symbol (hmodDevice, "newDeviceW_Advanced", (gpointer *)&pfnNewDevice);

               if (pfnNewDevice)
               {
                  pDevice = pfnNewDevice (true);

                  if (  (  (  fMustMatchPDLLevel
                           && iPDLLevel == pDevice->getPDLLevel ()
                           )
                        || !fMustMatchPDLLevel
                        )
                     && (  (  fMustMatchPDLSubLevel
                           && iPDLSubLevel == pDevice->getPDLSubLevel ()
                           )
                        || !fMustMatchPDLSubLevel
                        )
                     && (  (  fMustMatchPDLMajorRevisionLevel
                           && iPDLMajorRevisionLevel == pDevice->getPDLMajorRevisionLevel ()
                           )
                        || !fMustMatchPDLMajorRevisionLevel
                        )
                     && (  (  fMustMatchPDLMinorRevisionLevel
                           && iPDLMinorRevisionLevel == pDevice->getPDLMinorRevisionLevel ()
                           )
                        || !fMustMatchPDLMinorRevisionLevel
                        )
                     )
                  {
                     sprintf (pszAppend,
                              "%s.%s",
                              pDevice->getDriverName (),
                              pDevice->getDeviceName ());
                     pszAppend += strlen (pszAppend) + 1;
                  }

                  delete pDevice;
               }

               g_module_close (hmodDevice);
            }
         }

         *pszAppend++ = '\0'; // end-of-array

         delete pEnum;

         if (!sendResponse (fdS2C, PDCCMD_ACK, (char **)pszResponse))
         {
            free (pszResponse);
            goto BUGOUT;
         }

         free (pszResponse);

         break;
      }

      default:
      {
         if (!sendResponse (fdS2C, PDCCMD_UNSUPPORTED, "Unknown command"))
            goto BUGOUT;
         break;
      }
      }
   }

BUGOUT:

   // Clean up
   close (fdS2C);
   close (fdC2S);

   if (pszJobProperties)
   {
      free (pszJobProperties);
   }

   if (pbBuffer1)
   {
      shmdt (pbBuffer1);
   }
   if (pbBuffer2)
   {
      shmdt (pbBuffer2);
   }

   delete pDeviceRef;
   delete pCmd;

   if (pfpOut)
   {
      fclose (pfpOut);
      pfpOut = 0;
   }

   if (pfpErr)
   {
      fclose (pfpErr);
      pfpErr = 0;
   }

   return 0;
}
