/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include "PrintDevice.hpp"

#include <stdarg.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>

#include <fstream>
#include <vector>

#include <JobProperties.hpp>

#include "hppcl3/bitmap.hpp"

#include "version.hpp"

// warning: initializeJobProperties(), queryCurrentJobProperties(),
//          getJobProperty(), and listJobPropertyKeys() references these
static char *vapszJobPropertyKeys[] = {
   "orientation",
   "dither",
   "form",
   "tray",
   "media",
   "resolution",
   "printmode"
};
#define JOBPROP_ORIENTATION vapszJobPropertyKeys[0]
#define JOBPROP_DITHER      vapszJobPropertyKeys[1]
#define JOBPROP_FORM        vapszJobPropertyKeys[2]
#define JOBPROP_TRAY        vapszJobPropertyKeys[3]
#define JOBPROP_MEDIA       vapszJobPropertyKeys[4]
#define JOBPROP_RESOLUTION  vapszJobPropertyKeys[5]
#define JOBPROP_PRINTMODE   vapszJobPropertyKeys[6]

PrintDevice::
PrintDevice (char *pszDriverName,
             char *pszDeviceName,
             char *pszShortName,
             char *pszJobProperties)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputPrintDevice ()) DebugOutput::getErrorStream () << "PrintDevice::" << __FUNCTION__ << " ()" << endl;
#endif

   // Initialize class data
   outputStream_d              = &cout;
   outputStreamBuf_d           = 0;
   fShouldDeleteOutputStream_d = false;

   pfnOutputFunction_d         = 0;
   pMagicCookie_d              = 0;

   iLanguageID_d               = 0;
   pLanguage_d                 = 0;

   pszDriverName_d             = pszDriverName;
   pszDeviceName_d             = pszDeviceName;
   pszShortName_d              = pszShortName;
   lCapabilities_d             = 0;
   lRasterCapabilities_d       = 0;
   lDeviceOptions_d            = 0;
   lDeviceOptions2_d           = 0;

   pInstance_d                 = 0;
   pBlitter_d                  = 0;
   pPDL_d                      = 0;

   pszJobProperties_d          = 0;

   pOrientation_d              = 0;
   pszDitherID_d               = 0;
   pForm_d                     = 0;
   pTray_d                     = 0;
   pMedia_d                    = 0;
   pResolution_d               = 0;
   pCommand_d                  = 0;
   pPrintMode_d                = 0;
   pData_d                     = 0;
   pString_d                   = 0;

   pszLoadedLibrary_d          = 0;
   hModLibrary_d               = 0;

   if (   pszJobProperties
      && *pszJobProperties
      )
   {
      pszJobProperties_d = (char *)malloc (strlen (pszJobProperties) + 1);
      if (pszJobProperties_d)
         strcpy (pszJobProperties_d, pszJobProperties);
   }
}

PrintDevice::
~PrintDevice ()
{
   if (pInstance_d)
   {
      delete pInstance_d;
      pInstance_d = 0;
   }

   if (pBlitter_d)
   {
      delete pBlitter_d;
      pBlitter_d = 0;
   }

   if (pPDL_d)
   {
      delete pPDL_d;
      pPDL_d = 0;
   }

   if (fShouldDeleteOutputStream_d)
   {
      delete outputStream_d;
      delete outputStreamBuf_d;
   }

   iLanguageID_d = 0;
   delete pLanguage_d;
   pLanguage_d = 0;

   cleanupProperties ();

   if (pszLoadedLibrary_d)
   {
      free (pszLoadedLibrary_d);
      pszLoadedLibrary_d = 0;
   }

   if (hModLibrary_d)
   {
      g_module_close (hModLibrary_d);
      hModLibrary_d = 0;
   }

#ifndef RETAIL
   if (DebugOutput::shouldOutputPrintDevice ()) DebugOutput::getErrorStream () << "PrintDevice::~" << __FUNCTION__ << " ()" << endl;
#endif
}

bool PrintDevice::
hasError ()
{
   return false;
}

void PrintDevice::
initialize ()
{
   if (!pString_d)
   {
      pString_d = getDefaultString ();
   }

   iLanguageID_d = StringResource::LANGUAGE_DEFAULT;

   if (pString_d)
   {
      pString_d->setLanguage (iLanguageID_d);
   }

   pLanguage_d = StringResource::create (iLanguageID_d, pString_d);

   initializeJobProperties ();

   if (pInstance_d)
      // Initialize the device instance
      pInstance_d->initializeInstance ();

   if (pBlitter_d)
      // Initialize the device blitter
      pBlitter_d->initializeInstance ();
}

void PrintDevice::
cleanupProperties ()
{
   if (pszJobProperties_d)
   {
      free (pszJobProperties_d);
      pszJobProperties_d = 0;
   }

   if (pOrientation_d)
   {
      delete pOrientation_d;
      pOrientation_d = 0;
   }

   if (pszDitherID_d)
   {
      free (pszDitherID_d);
      pszDitherID_d = 0;
   }

   if (pForm_d)
   {
      delete pForm_d;
      pForm_d = 0;
   }

   if (pTray_d)
   {
      delete pTray_d;
      pTray_d = 0;
   }

   if (pMedia_d)
   {
      delete pMedia_d;
      pMedia_d = 0;
   }

   if (pResolution_d)
   {
      delete pResolution_d;
      pResolution_d = 0;
   }

   if (pPrintMode_d)
   {
      delete pPrintMode_d;
      pPrintMode_d = 0;
   }

   if (pCommand_d)
   {
      delete pCommand_d;
      pCommand_d = 0;
   }

   if (pData_d)
   {
      delete pData_d;
      pData_d = 0;
   }

   if (pString_d)
   {
      delete pString_d;
      pString_d = 0;
   }
}

/*8<--------8<--------8<--------8<--------8<--------8<--------8<--------*/

char * PrintDevice::
getVersion ()
{
   return OMNI_VERSION;
}

/*8<--------8<--------8<--------8<--------8<--------8<--------8<--------*/

char * PrintDevice::
getDriverName ()
{
   return pszDriverName_d;
}

char * PrintDevice::
getDeviceName ()
{
   return pszDeviceName_d;
}

char * PrintDevice::
getShortName ()
{
   return pszShortName_d;
}

/*8<--------8<--------8<--------8<--------8<--------8<--------8<--------*/

void PrintDevice::
setDeviceInstance (DeviceInstance *pInstance)
{
   pInstance_d = pInstance;
}

DeviceInstance * PrintDevice::
getDeviceInstance ()
{
   return pInstance_d;
}

void PrintDevice::
setDeviceBlitter (DeviceBlitter *pBlitter)
{
   pBlitter_d = pBlitter;
}

DeviceBlitter * PrintDevice::
getDeviceBlitter ()
{
   return pBlitter_d;
}

void PrintDevice::
setPDL (PDL *pPDL)
{
   pPDL_d = pPDL;
}

/*8<--------8<--------8<--------8<--------8<--------8<--------8<--------*/

void PrintDevice::
setCapabilities (long lCapabilities)
{
   lCapabilities_d |= lCapabilities;
}

void PrintDevice::
setRasterCapabilities (long lRasterCapabilities)
{
   lRasterCapabilities_d |= lRasterCapabilities;
}

void PrintDevice::
setDeviceOptions (long lOptions)
{
   lDeviceOptions_d |= lOptions;
}

void PrintDevice::
setDeviceOptions2 (long lOptions)
{
   lDeviceOptions2_d |= lOptions;
}

bool PrintDevice::
hasCapability (long lMask)
{
   if ((lCapabilities_d & lMask) != 0)
      return true;
   else
      return false;
}

bool PrintDevice::
hasRasterCapability (long lMask)
{
   if ((lRasterCapabilities_d & lMask) != 0)
      return true;
   else
      return false;
}

bool PrintDevice::
hasDeviceOption (long lMask)
{
   if ((lDeviceOptions_d & lMask) != 0)
      return true;
   else
      return false;
}

bool PrintDevice::
hasDeviceOption2 (long lMask)
{
   if ((lDeviceOptions2_d & lMask) != 0)
      return true;
   else
      return false;
}

/*8<--------8<--------8<--------8<--------8<--------8<--------8<--------*/

int PrintDevice::
getScanlineMultiple ()
{
   int iScanLineMultiple = pResolution_d->getScanlineMultiple ();

   if (0 == iScanLineMultiple)
   {
      iScanLineMultiple = 1;

#ifndef RETAIL
      if (DebugOutput::shouldOutputPrintDevice ()) DebugOutput::getErrorStream () << "PrintDevice::" << __FUNCTION__ << " (): Error: iScanLineMultiple == 0!" << endl;
#endif
   }

   return iScanLineMultiple;
}

/*8<--------8<--------8<--------8<--------8<--------8<--------8<--------*/

static int
parseSize (char **ppszComma)
{
   int iReturn = 0;

   while (  '0' <= **ppszComma
         && **ppszComma <= '9'
         )
   {
      iReturn = iReturn * 10 + (**ppszComma - '0');

      (*ppszComma)++;
   }

   return iReturn;
}

void PrintDevice::
initializeJobProperties ()
{
   char *pszError = 0;

   // Initialize to defaults.  This is done always since you need an
   // object to create another of the same class.
   pOrientation_d = getDefaultOrientation ();
   pForm_d        = getDefaultForm ();
   pTray_d        = getDefaultTray ();
   pMedia_d       = getDefaultMedia ();
   pResolution_d  = getDefaultResolution ();
   pPrintMode_d   = getDefaultPrintMode ();

   // Allocate a copy of the dither id since it is freed
   char *pszDitherID = getDefaultDitherID ();

   if (pszDitherID)
   {
      pszDitherID_d = (char *)malloc (strlen (pszDitherID) + 1);
      if (pszDitherID_d)
         strcpy (pszDitherID_d, pszDitherID);
   }

   // These are not job properties but initialize them with the rest
   if (!pCommand_d)
      pCommand_d = getDefaultCommands ();
   if (!pData_d)
      pData_d = getDefaultData ();

   if (  pszJobProperties_d
      && *pszJobProperties_d
      )
   {
      JobProperties jobProp (pszJobProperties_d);

#ifndef RETAIL
      if (DebugOutput::shouldOutputPrintDevice ()) DebugOutput::getErrorStream () << "PrintDevice::" << __FUNCTION__ << ": The job properties on input are: " << pszJobProperties_d << endl;
#endif

      while (jobProp.hasKeyNameLeft ())
      {
         char *pszKey   = jobProp.getCurrentKey ();
         char *pszValue = jobProp.getCurrentValue ();

#ifndef RETAIL
         if (DebugOutput::shouldOutputPrintDevice ()) DebugOutput::getErrorStream () << "PrintDevice::" << __FUNCTION__ << ": found " << pszKey << "=" << pszValue << endl;
#endif
         if (0 == strcasecmp (JOBPROP_ORIENTATION, pszKey))
         {
            DeviceOrientation *pDO = pOrientation_d->create (pszValue);
            if (pDO)
            {
               delete pOrientation_d;
               pOrientation_d = pDO;
            }
            else
            {
#ifndef RETAIL
               if (DebugOutput::shouldOutputPrintDevice ()) DebugOutput::getErrorStream () << "PrintDevice::" << __FUNCTION__ << ": " << pszValue << " is an invalid Orientation name!" << endl;
#endif
            }
         }
         else if (0 == strcasecmp (JOBPROP_DITHER, pszKey))
         {
            if (DeviceDither::ditherNameValid (pszValue))
            {
               if (pszDitherID_d)
               {
                  free (pszDitherID_d);
                  pszDitherID_d = 0;
               }

               pszDitherID_d = (char *)malloc (strlen (pszValue) + 1);
               if (pszDitherID_d)
               {
                  strcpy (pszDitherID_d, pszValue);
               }
               else
               {
                  pszError = "Error: out of memory!";
               }
            }
            else
            {
               pszError = "Error: Invalid dither name";
            }
         }
         else if (0 == strcasecmp (JOBPROP_FORM, pszKey))
         {
            char *pszComma = 0;

            pszComma = strchr (pszValue, ',');
            if (pszComma)
               *pszComma++ = '\0';

            DeviceForm *pDF = pForm_d->create (pszValue);
            if (pDF)
            {
               delete pForm_d;
               pForm_d = pDF;
            }
            else
            {
#ifndef RETAIL
               if (DebugOutput::shouldOutputPrintDevice ()) DebugOutput::getErrorStream () << "PrintDevice::" << __FUNCTION__ << ": " << pszValue << " is an invalid Form name!" << endl;
#endif

               if (pszComma)
               {
                  pszComma[-1] = ',';
                  pszComma = 0;
               }
            }

            if (pszComma)
            {
               int iX = 0,
                   iY = 0;

               iX = parseSize (&pszComma);

               if (pDF->hasCapability (DeviceForm::FORM_CAPABILITY_USERDEFINED))
               {
                  if (',' != *pszComma)
                  {
                     pszError = "Error: Form missing second size";
                  }
                  else
                  {
                     pszComma++;
                     iY = parseSize (&pszComma);

                     if (  0 == iX
                        || 0 == iY
                        )
                     {
                        pszError = "Error: Missing/Bad size for user defined form";
                     }
                  }
               }
               else if (pDF->hasCapability (DeviceForm::FORM_CAPABILITY_ROLL))
               {
                  iY = iX;
                  iX = 0;

                  if (0 == iY)
                  {
                     pszError = "Error: Missing/Bad size for roll form";
                  }
               }
               else
               {
                  pszError = "Error: unknown capability!";
               }

               if (  !pszError
                  && *pszComma
                  )
               {
                  pszError = "Error: Parse error!";
               }

               if (  !pszError
                  && iX
                  )
               {
                  pDF->setCx (iX);
               }
               if (  !pszError
                  && iY
                  )
               {
                  pDF->setCy (iY);
               }
            }
         }
         else if (0 == strcasecmp (JOBPROP_TRAY, pszKey))
         {
            DeviceTray *pDT = pTray_d->create (pszValue);
            if (pDT)
            {
               delete pTray_d;
               pTray_d = pDT;
            }
            else
            {
#ifndef RETAIL
               if (DebugOutput::shouldOutputPrintDevice ()) DebugOutput::getErrorStream () << "PrintDevice::" << __FUNCTION__ << ": " << pszValue << " is an invalid Tray name!" << endl;
#endif
            }
         }
         else if (0 == strcasecmp (JOBPROP_MEDIA, pszKey))
         {
            DeviceMedia *pDM = pMedia_d->create (pszValue);
            if (pDM)
            {
               delete pMedia_d;
               pMedia_d = pDM;
            }
            else
            {
#ifndef RETAIL
               if (DebugOutput::shouldOutputPrintDevice ()) DebugOutput::getErrorStream () << "PrintDevice::" << __FUNCTION__ << ": " << pszValue << " is an invalid Media name!" << endl;
#endif
            }
         }
         else if (0 == strcasecmp (JOBPROP_RESOLUTION, pszKey))
         {
            DeviceResolution *pDR = pResolution_d->create (pszValue);
            if (pDR)
            {
               delete pResolution_d;
               pResolution_d = pDR;
            }
            else
            {
#ifndef RETAIL
               if (DebugOutput::shouldOutputPrintDevice ()) DebugOutput::getErrorStream () << "PrintDevice::" << __FUNCTION__ << ": " << pszValue << " is an invalid Resolution name!" << endl;
#endif
            }
         }
         else if (0 == strcasecmp (JOBPROP_PRINTMODE, pszKey))
         {
            DevicePrintMode *pDPM = pPrintMode_d->create (pszValue);
            if (pDPM)
            {
               delete pPrintMode_d;
               pPrintMode_d = pDPM;
            }
            else
            {
#ifndef RETAIL
               if (DebugOutput::shouldOutputPrintDevice ()) DebugOutput::getErrorStream () << "PrintDevice::" << __FUNCTION__ << ": " << pszValue << " is an invalid Print Mode name!" << endl;
#endif
            }
         }
#ifndef RETAIL
         else if (0 == strcasecmp ("debugoutput", pszKey))
         {
            DebugOutput::setDebugOutput (pszValue);
            pInstance_d->handleDeviceJobProperty (pszKey, pszValue);
         }
#endif
         else
         {
#ifndef RETAIL
            bool rc =
#endif
                      pInstance_d->handleDeviceJobProperty (pszKey, pszValue);

#ifndef RETAIL
            if (DebugOutput::shouldOutputPrintDevice ())
               DebugOutput::getErrorStream () << "PrintDevice::" << __FUNCTION__
                    << ": device returned "
                    << dec << rc << " from handleDeviceJobProperty ("
                    << pszKey << ", "
                    << pszValue << ")" << endl;
#endif
         }

         jobProp.getNextKey ();
      }
   }
   else
   {
#ifndef RETAIL
      if (DebugOutput::shouldOutputPrintDevice ())
         DebugOutput::getErrorStream () << "PrintDevice::" << __FUNCTION__ << ": There are no job properties on input!" << endl;
#endif
   }

   if (  pForm_d
      && pResolution_d
      )
   {
      pForm_d->associateWith (pResolution_d);
   }

   if (pForm_d)
   {
      HardCopyCap *pHCC = pForm_d->getHardCopyCap ();

      if (pHCC)
      {
         if (  0 == pHCC->getCx ()
            || 0 == pHCC->getCy ()
            )
         {
            if (pForm_d != getDefaultForm ())
            {
               delete pForm_d;

               pszError = "Error: new form has 0 for its size.";

               pForm_d = getDefaultForm ();

               pForm_d->associateWith (pResolution_d);

               pHCC = pForm_d->getHardCopyCap ();
            }
         }
         if (  0 == pHCC->getCx ()
            || 0 == pHCC->getCy ()
            )
         {
            pszError = "Error: form has 0 for its size.";
         }
      }
   }

   if (pszError)
   {
      DebugOutput::getErrorStream () << pszError << endl;
      // @TBD - how to handle and report errors?
   }

#ifndef RETAIL
   if (DebugOutput::shouldOutputPrintDevice ())
   {
      DebugOutput::getErrorStream () << "PrintDevice::" << __FUNCTION__ << ": The new job properties are as follows:" << endl;
      if (pOrientation_d)
         DebugOutput::getErrorStream () << "PrintDevice::" << __FUNCTION__ << ": " << pOrientation_d->getName () << " = " << *pOrientation_d << endl;
      else
         DebugOutput::getErrorStream () << "PrintDevice::" << __FUNCTION__ << ": Orientation is NULL!" << endl;
      if (pszDitherID_d)
         DebugOutput::getErrorStream () << "PrintDevice::" << __FUNCTION__ << ": pszDitherID_d = " << pszDitherID_d << endl;
      else
         DebugOutput::getErrorStream () << "PrintDevice::" << __FUNCTION__ << ": Dither is NULL!" << endl;
      if (pForm_d)
         DebugOutput::getErrorStream () << "PrintDevice::" << __FUNCTION__ << ": " << pForm_d->getName () << " = " << *pForm_d << endl;
      else
         DebugOutput::getErrorStream () << "PrintDevice::" << __FUNCTION__ << ": Form is NULL!" << endl;
      if (pTray_d)
         DebugOutput::getErrorStream () << "PrintDevice::" << __FUNCTION__ << ": " << pTray_d->getName () << " = " << *pTray_d << endl;
      else
         DebugOutput::getErrorStream () << "PrintDevice::" << __FUNCTION__ << ": Tray is NULL!" << endl;
      if (pMedia_d)
         DebugOutput::getErrorStream () << "PrintDevice::" << __FUNCTION__ << ": " << pMedia_d->getName () << " = " << *pMedia_d << endl;
      else
         DebugOutput::getErrorStream () << "PrintDevice::" << __FUNCTION__ << ": Media is NULL!" << endl;
      if (pResolution_d)
         DebugOutput::getErrorStream () << "PrintDevice::" << __FUNCTION__ << ": " << pResolution_d->getName () << " = " << *pResolution_d << endl;
      else
         DebugOutput::getErrorStream () << "PrintDevice::" << __FUNCTION__ << ": Resolution is NULL!" << endl;
      if (pPrintMode_d)
         DebugOutput::getErrorStream () << "PrintDevice::" << __FUNCTION__ << ": " << pPrintMode_d->getName () << " = " << *pPrintMode_d << endl;
      else
         DebugOutput::getErrorStream () << "PrintDevice::" << __FUNCTION__ << ": PrintMode is NULL!" << endl;
   }
#endif
}

DeviceOrientation * PrintDevice::
getCurrentOrientation ()
{
   return pOrientation_d;
}

char * PrintDevice::
getCurrentDitherID ()
{
   return pszDitherID_d;
}

DeviceDither * PrintDevice::
getCurrentDither ()
{
   return pBlitter_d->getDitherInstance ();
}

DeviceForm * PrintDevice::
getCurrentForm ()
{
   return pForm_d;
}

DeviceTray * PrintDevice::
getCurrentTray ()
{
   return pTray_d;
}

DeviceMedia * PrintDevice::
getCurrentMedia ()
{
   return pMedia_d;
}

DeviceResolution * PrintDevice::
getCurrentResolution ()
{
   return pResolution_d;
}

DevicePrintMode * PrintDevice::
getCurrentPrintMode ()
{
   return pPrintMode_d;
}

DeviceCommand * PrintDevice::
getCommands ()
{
   return pCommand_d;
}

DeviceData * PrintDevice::
getDeviceData ()
{
   return pData_d;
}

DeviceString * PrintDevice::
getDeviceString ()
{
   return pString_d;
}

/*8<--------8<--------8<--------8<--------8<--------8<--------8<--------*/

int PrintDevice::
getPDLLevel ()
{
   if (pPDL_d)
      return pPDL_d->getPDLLevel ();
   else
      return 0;
}

int PrintDevice::
getPDLSubLevel ()
{
   if (pPDL_d)
      return pPDL_d->getPDLSubLevel ();
   else
      return 0;
}

int PrintDevice::
getPDLMajorRevisionLevel ()
{
   if (pPDL_d)
      return pPDL_d->getPDLMajorRevisionLevel ();
   else
      return 0;
}

int PrintDevice::
getPDLMinorRevisionLevel ()
{
   if (pPDL_d)
      return pPDL_d->getPDLMinorRevisionLevel ();
   else
      return 0;
}

/*8<--------8<--------8<--------8<--------8<--------8<--------8<--------*/

bool PrintDevice::
beginJob ()
{
   return pInstance_d->beginJob ();
}

bool PrintDevice::
beginJob (char *pszJobProperties)
{
   if (pszJobProperties)
   {
      // Free the old ones
      cleanupProperties ();

      // Allocate the new one
      pszJobProperties_d = (char *)malloc (strlen (pszJobProperties) + 1);
      if (pszJobProperties_d)
      {
         strcpy (pszJobProperties_d, pszJobProperties);

         initializeJobProperties ();
      }
   }

   return pInstance_d->beginJob (true);
}

bool PrintDevice::
newFrame ()
{
   return pInstance_d->newFrame ();
}

bool PrintDevice::
newFrame (char *pszJobProperties)
{
   if (pszJobProperties)
   {
      // Free the old ones
      cleanupProperties ();

      // Allocate the new one
      pszJobProperties_d = (char *)malloc (strlen (pszJobProperties) + 1);
      if (pszJobProperties_d)
      {
         strcpy (pszJobProperties_d, pszJobProperties);

         initializeJobProperties ();
      }
   }

   return pInstance_d->newFrame (true);
}

bool PrintDevice::
endJob ()
{
   return pInstance_d->endJob ();
}

bool PrintDevice::
abortJob ()
{
   return pInstance_d->abortJob ();
}

/*8<--------8<--------8<--------8<--------8<--------8<--------8<--------*/

bool PrintDevice::
rasterize (PBYTE        pbBits,
           PBITMAPINFO2 pbmi2,
           PRECTL       prectlPageLocation,
           BITBLT_TYPE  eType)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputPrintDevice ())
      DebugOutput::getErrorStream ()
           << "PrintDevice::rasterize (0x"
           << hex << (int)pbBits << dec << ", {"
           << pbmi2->cx << ", "
           << pbmi2->cy << ", "
           << pbmi2->cPlanes << ", "
           << pbmi2->cBitCount << "}, "
           << "{" << prectlPageLocation->xLeft << ", " << prectlPageLocation->yBottom << ", " << prectlPageLocation->xRight << ", " << prectlPageLocation->yTop << "})"
           << endl;
#endif

   char *pszDumpEnvironmentVar = getenv ("DUMP_INCOMING_BITMAPS");
   bool  fDumpIncomingBitmaps  = false;

   if (pszDumpEnvironmentVar)
   {
#ifndef RETAIL
      if (DebugOutput::shouldOutputPrintDevice ()) DebugOutput::getErrorStream () << "PrintDevice::" << __FUNCTION__ << ": (in)pszDumpEnvironmentVar = " << hex << (int)pszDumpEnvironmentVar << dec << endl;
#endif

      if (*pszDumpEnvironmentVar)
         fDumpIncomingBitmaps = true;
   }

   if (fDumpIncomingBitmaps)
   {
      static int  iNum = 0;
      char        achName[4 + 1 + 3 + 1];
      PNEUTRALRGB pColors = 0;

      sprintf (achName, "%04dIN.bmp", iNum);

      if (8 >= pbmi2->cBitCount)
      {
         int iNumColors = 1 << pbmi2->cBitCount;

         pColors = (PNEUTRALRGB)malloc (sizeof (NEUTRALRGB) * iNumColors);
         if (pColors)
         {
            PRGB2 pRGBs = pbmi2->argbColor;

            for (int i = 0; i < iNumColors; i++)
            {
               pColors[i].bRed   = pRGBs[i].bRed;
               pColors[i].bGreen = pRGBs[i].bGreen;
               pColors[i].bBlue  = pRGBs[i].bBlue;
            }
         }
      }

      Bitmap dump (achName,
                   pbmi2->cx,
                   pbmi2->cy,
                   pbmi2->cBitCount,
                   pColors);
      dump.addScanLine (pbBits, pbmi2->cy);

      if (pColors)
      {
         free (pColors);
      }

      iNum++;
      if (999 < iNum)
         iNum = 0;
   }

   return pBlitter_d->rasterize (pbBits,
                                 pbmi2,
                                 prectlPageLocation,
                                 eType);
}

/*8<--------8<--------8<--------8<--------8<--------8<--------8<--------*/

void PrintDevice::
setOutputStream (FILE *pFile)
{
#ifdef GCC_VER_3
   outputStreamBuf_d           = new filebuf (pFile, ios::binary | ios::out);
#else
   outputStreamBuf_d           = new filebuf (fileno (pFile));
#endif
   outputStream_d              = new ostream (outputStreamBuf_d);
   fShouldDeleteOutputStream_d = true;

   if (pInstance_d)
   {
      pInstance_d->setOutputStream (pFile);
   }
}

void PrintDevice::
setOutputFunction (PFNOUTPUTFUNCTION  pfnOutputFunction,
                   void              *pMagicCookie)
{
   pfnOutputFunction_d = pfnOutputFunction;
   pMagicCookie_d      = pMagicCookie;
}

void PrintDevice::
setErrorStream (FILE *pFile)
{
   if (pInstance_d)
   {
      pInstance_d->setErrorStream (pFile);
   }
}

/*8<--------8<--------8<--------8<--------8<--------8<--------8<--------*/

bool PrintDevice::
setLanguage (int iLanguageID)
{
   if (iLanguageID != iLanguageID_d)
   {
      if (pInstance_d)
      {
         pInstance_d->setLanguage (iLanguageID_d);
      }

      if (pString_d)
      {
         pString_d->setLanguage (iLanguageID_d);
      }

      StringResource *pNewLanguage = StringResource::create (iLanguageID, pString_d);

      if (pNewLanguage)
      {
         delete pLanguage_d;

         iLanguageID_d = iLanguageID;
         pLanguage_d   = pNewLanguage;

         return true;
      }
      else
      {
         return false;
      }
   }

   return true;
}

int PrintDevice::
getLanguage ()
{
   return iLanguageID_d;
}

Enumeration * PrintDevice::
getLanguages ()
{
   return StringResource::getLanguages ();
}

StringResource * PrintDevice::
getLanguageResource ()
{
   return pLanguage_d;
}

/*8<--------8<--------8<--------8<--------8<--------8<--------8<--------*/

bool PrintDevice::
sendBinaryDataToDevice (DeviceForm *pForm)
{
   return sendBinaryDataToDevice (pForm->getData ());
}

bool PrintDevice::
sendBinaryDataToDevice (DeviceTray *pTray)
{
   return sendBinaryDataToDevice (pTray->getData ());
}

bool PrintDevice::
sendBinaryDataToDevice (DeviceMedia *pMedia)
{
   return sendBinaryDataToDevice (pMedia->getData ());
}

bool PrintDevice::
sendBinaryDataToDevice (DeviceResolution *pResolution)
{
   return sendBinaryDataToDevice (pResolution->getData ());
}

bool PrintDevice::
sendBinaryDataToDevice (BinaryData *pData)
{
   return sendBinaryDataToDevice (pData->getData (),
                                  pData->getLength ());
}

bool PrintDevice::
sendBinaryDataToDevice (PBYTE pbData, int iLength)
{
   if (pfnOutputFunction_d)
   {
      pfnOutputFunction_d (pMagicCookie_d, (unsigned char *)pbData, iLength);
   }
   else
   {
      outputStream_d->write ((const char *)pbData, iLength);
      outputStream_d->flush ();
   }

   return true;
}

bool PrintDevice::
sendVPrintfToDevice (BinaryData *pData,
                     va_list     list)
{
   char    achCmd[10];
   char    achOutput[512];
   int     iBytesWritten = 0;
   bool    bRC           = false;
   char   *pszCommand    = (char *)pData->getData ();
   int     iLength       = pData->getLength ();
   int     iSrc, iDst;

   for (iSrc = 0, iDst = 0; iSrc < iLength; iSrc++)
   {
      achOutput[iDst] =pszCommand[iSrc];

      if (  '%' == pszCommand[iSrc]
         && '%' != pszCommand[iSrc+1]
         )
      {
         short  sLen          = 0;
         int    iValueOnStack;
         double dValueOnStack;

         achCmd[0] = pszCommand[iSrc];
         achCmd[1] = pszCommand[iSrc + 1];
         achCmd[2] = 0;

         if (  'f' == achCmd[1]
            || 'F' == achCmd[1]
            )
         {
            iValueOnStack = 0;
            dValueOnStack = va_arg (list, double);
         }
         else
         {
            iValueOnStack = va_arg (list, int);
            dValueOnStack = 0.0;
         }

         if ('d' == achCmd[1])
         {
            /* Special format character -- double word, little endian
            */
            achOutput[iDst]     = (iValueOnStack & 0x000000FF);
            achOutput[iDst + 1] = (iValueOnStack & 0x0000FF00) >> 8;
            achOutput[iDst + 2] = (iValueOnStack & 0x00FF0000) >> 16;
            achOutput[iDst + 3] = (iValueOnStack & 0xFF000000) >> 24;
            sLen                = 4;
         }
         else if ('D' == achCmd[1])
         {
            /* Special format character -- double word, big endian
            */
            achOutput[iDst]     = (iValueOnStack & 0xFF000000) >> 24;
            achOutput[iDst + 1] = (iValueOnStack & 0x00FF0000) >> 16;
            achOutput[iDst + 2] = (iValueOnStack & 0x0000FF00) >> 8;
            achOutput[iDst + 3] = (iValueOnStack & 0x000000FF);
            sLen                = 4;
         }
         else if ('w' == achCmd[1])
         {
            /* Special format character -- single word, little endian
            */
            achOutput[iDst]     = (iValueOnStack & 0x00FF);
            achOutput[iDst + 1] = (iValueOnStack & 0xFF00) >> 8;
            sLen                = 2;
         }
         else if ('W' == achCmd[1])
         {
            /* Special format character -- single word, big endian
            */
            achOutput[iDst]     = (iValueOnStack & 0xFF00) >> 8;
            achOutput[iDst + 1] = (iValueOnStack & 0x00FF);
            sLen                = 2;
         }
         else if (  'c' == achCmd[1]
                 || 'C' == achCmd[1]
                 )
         {
            /* Special format character -- single byte
            ** Need to special case this because if you use %c and give it
            ** a 0, then strlen will return 0 instead of 1.
            */
            achOutput[iDst] = (iValueOnStack & 0xFF);
            sLen            = 1;
         }
         else if (  'n' == achCmd[1]
                 || 'N' == achCmd[1]
                 )
         {
            /* Special format character -- ASCII number string
            */
            sprintf (achOutput + iDst, "%d", iValueOnStack);
            sLen = strlen (achOutput + iDst);
         }
         else if (  'f' == achCmd[1]
                 || 'F' == achCmd[1]
                 )
         {
            /* Special format character -- ASCII number string with decimal
            */
            sprintf (achOutput + iDst, "%f", dValueOnStack);
            sLen = strlen (achOutput + iDst);
         }
         else
         {
            sprintf (achOutput + iDst,
                     achCmd,
                     iValueOnStack);

            if (iValueOnStack)
               sLen = strlen (achOutput + iDst);
            else
               sLen = 1;
         }

         iBytesWritten += sLen;
         iDst += sLen;
         iSrc++;
      }
      else if (  '%' == pszCommand[iSrc]
              && '%' == pszCommand[iSrc+1]
              )
      {
         achOutput[iDst] = '%';

         iBytesWritten++;
         iDst++;
         iSrc++;
      }
      else
      {
         iBytesWritten++;
         iDst++;
      }
   }

   if (pfnOutputFunction_d)
   {
      pfnOutputFunction_d (pMagicCookie_d, (unsigned char *)achOutput, iDst);
   }
   else
   {
      outputStream_d->write (achOutput, iDst);
      outputStream_d->flush ();
   }

   return bRC;
}

bool PrintDevice::
sendPrintfToDevice (BinaryData *pData, ...)
{
   va_list list;
   bool    rc;

   va_start (list, pData);

   rc = sendVPrintfToDevice (pData, list);

   va_end (list);

   return rc;
}

/*8<--------8<--------8<--------8<--------8<--------8<--------8<--------*/

string * PrintDevice::
queryDeviceJobProperties ()
{
   ostringstream  oss;
   char          *pszValue;

   oss << JOBPROP_ORIENTATION
       << "=";
   pszValue = pOrientation_d->getDeviceID ();
   if (!pszValue)
   {
      pszValue = DeviceOrientation::IDToName (pOrientation_d->getID ());
   }
   oss << pszValue;

   oss << " "
       << JOBPROP_DITHER
       << "="
       << pszDitherID_d;

   oss << " "
       << JOBPROP_FORM
       << "=";
   pszValue = pForm_d->getDeviceID ();
   if (pszValue)
   {
      oss << pszValue;
   }
   else
   {
      oss << DeviceForm::IDToName (pForm_d->getID ());

      if (pForm_d->getOverrideX ())
      {
         oss << "," << pForm_d->getOverrideX ();
      }
      if (pForm_d->getOverrideY ())
      {
         oss << "," << pForm_d->getOverrideY ();
      }
   }

   oss << " "
       << JOBPROP_TRAY
       << "=";
   pszValue = pTray_d->getDeviceID ();
   if (!pszValue)
   {
      pszValue = DeviceTray::IDToName (pTray_d->getID ());
   }
   oss << pszValue;

   oss << " "
       << JOBPROP_MEDIA
       << "=";
   pszValue = pMedia_d->getDeviceID ();
   if (!pszValue)
   {
      pszValue = DeviceMedia::IDToName (pMedia_d->getID ());
   }
   oss << pszValue;

   oss << " "
       << JOBPROP_RESOLUTION
       << "=";
   pszValue = pResolution_d->getDeviceID ();
   if (!pszValue)
   {
      pszValue = DeviceResolution::IDToName (pResolution_d->getID ());
   }
   oss << pszValue;

   oss << " "
       << JOBPROP_PRINTMODE
       << "=";
   pszValue = pPrintMode_d->getDeviceID ();
   if (!pszValue)
   {
      pszValue = DevicePrintMode::IDToName (pPrintMode_d->getID ());
   }
   oss << pszValue;

   string *pDeviceJobProp = pInstance_d->queryDeviceJobProperties ();

   if (pDeviceJobProp)
   {
      oss << " " << *pDeviceJobProp;

      delete pDeviceJobProp;
   }

   string *pRet = new string (oss.str ());

   return pRet;
}

string * PrintDevice::
queryCurrentJobProperties ()
{
   ostringstream oss;

   oss <<        JOBPROP_ORIENTATION << "=" << DeviceOrientation::IDToName (pOrientation_d->getID ())
       << " " << JOBPROP_DITHER      << "=" << pszDitherID_d
       << " " << JOBPROP_FORM        << "=" << DeviceForm::IDToName (pForm_d->getID ());

   if (pForm_d->getOverrideX ())
   {
      oss << "," << pForm_d->getOverrideX ();
   }
   if (pForm_d->getOverrideY ())
   {
      oss << "," << pForm_d->getOverrideY ();
   }

   oss << " " << JOBPROP_TRAY        << "=" << DeviceTray::IDToName (pTray_d->getID ())
       << " " << JOBPROP_MEDIA       << "=" << DeviceMedia::IDToName (pMedia_d->getID ())
       << " " << JOBPROP_RESOLUTION  << "=" << DeviceResolution::IDToName (pResolution_d->getID ())
       << " " << JOBPROP_PRINTMODE   << "=" << DevicePrintMode::IDToName (pPrintMode_d->getID ());

   string *pDeviceJobProp = pInstance_d->queryDeviceJobProperties ();

   if (pDeviceJobProp)
   {
      oss << " " << *pDeviceJobProp;

      delete pDeviceJobProp;
   }

   string *pRet = new string (oss.str ());

   return pRet;
}

Enumeration * PrintDevice::
listJobPropertyKeys ()
{
   class JobPropertyKeyEnumerator : public Enumeration
   {
   public:
      JobPropertyKeyEnumerator (int cKeys, char **aKeys)
      {
         iKey_d  = 0;
         cKeys_d = cKeys;
         aKeys_d = aKeys;
      }

      virtual bool hasMoreElements ()
      {
         if (iKey_d < cKeys_d)
            return true;
         else
            return false;
      }

      virtual void *nextElement ()
      {
         if (iKey_d > cKeys_d - 1)
            return 0;

         return aKeys_d[iKey_d++];
      }

   private:
      int    iKey_d;
      int    cKeys_d;
      char **aKeys_d;
   };

   return new JobPropertyKeyEnumerator (dimof (vapszJobPropertyKeys), vapszJobPropertyKeys);
}

Enumeration * PrintDevice::
listDeviceJobPropertyKeys ()
{
   class NullEnumerator : public Enumeration
   {
   public:
      NullEnumerator ()
      {
      }

      virtual bool hasMoreElements ()
      {
         return false;
      }

      virtual void *nextElement ()
      {
         return 0;
      }
   };

   if (pInstance_d)
   {
      return pInstance_d->listDeviceJobPropertyKeys ();
   }
   else
   {
      return new NullEnumerator ();
   }
}

Enumeration * PrintDevice::
listKeyValues (char *pszKey)
{
   class ListEnumerator : public Enumeration
   {
   public:
      ListEnumerator ()
      {
         iCurrentElement_d = 0;
      }

      virtual ~ListEnumerator ()
      {
      }

      void
      addElement (char *pszValue)
      {
         list_d.push_back (pszValue);
      }

      virtual bool hasMoreElements ()
      {
         return iCurrentElement_d < (int)list_d.size ();
      }

      virtual void *nextElement ()
      {
         void *pvRet = 0;

         if (iCurrentElement_d < (int)list_d.size ())
         {
            string& pValue = list_d[iCurrentElement_d];

            pvRet = (void *)pValue.c_str ();

            iCurrentElement_d++;
         }

         return pvRet;
      }

   private:
      vector<string> list_d;
      int            iCurrentElement_d;
   };

   ListEnumerator *pList = new ListEnumerator ();
   Enumeration    *pEnum = 0;

   if (0 == strcasecmp (JOBPROP_ORIENTATION, pszKey))
   {
      DeviceOrientation *pDO = getCurrentOrientation ();

      pEnum = pDO->getEnumeration ();

      while (pEnum->hasMoreElements ())
      {
         pDO = (DeviceOrientation *)pEnum->nextElement ();

         pList->addElement (DeviceOrientation::IDToName (pDO->getID ()));

         delete pDO;
      }

      delete pEnum;
   }
   else if (0 == strcasecmp (JOBPROP_DITHER, pszKey))
   {
      pEnum = DeviceDither::getEnumeration ();

      while (pEnum->hasMoreElements ())
      {
         char *pszValue = (char *)pEnum->nextElement ();

         pList->addElement (pszValue);
      }

      delete pEnum;
   }
   else if (0 == strcasecmp (JOBPROP_FORM, pszKey))
   {
      DeviceForm *pDF = getCurrentForm ();

      pEnum = pDF->getEnumeration ();

      while (pEnum->hasMoreElements ())
      {
         pDF = (DeviceForm *)pEnum->nextElement ();

         pList->addElement (DeviceForm::IDToName (pDF->getID ()));

         delete pDF;
      }

      delete pEnum;
   }
   else if (0 == strcasecmp (JOBPROP_TRAY, pszKey))
   {
      DeviceTray *pDT = getCurrentTray ();

      pEnum = pDT->getEnumeration ();

      while (pEnum->hasMoreElements ())
      {
         pDT = (DeviceTray *)pEnum->nextElement ();

         pList->addElement (DeviceTray::IDToName (pDT->getID ()));

         delete pDT;
      }

      delete pEnum;
   }
   else if (0 == strcasecmp (JOBPROP_MEDIA, pszKey))
   {
      DeviceMedia *pDM = getCurrentMedia ();

      pEnum = pDM->getEnumeration ();

      while (pEnum->hasMoreElements ())
      {
         pDM = (DeviceMedia *)pEnum->nextElement ();

         pList->addElement (DeviceMedia::IDToName (pDM->getID ()));

         delete pDM;
      }

      delete pEnum;
   }
   else if (0 == strcasecmp (JOBPROP_RESOLUTION, pszKey))
   {
      DeviceResolution *pDR = getCurrentResolution ();

      pEnum = pDR->getEnumeration ();

      while (pEnum->hasMoreElements ())
      {
         pDR = (DeviceResolution *)pEnum->nextElement ();

         pList->addElement (DeviceResolution::IDToName (pDR->getID ()));

         delete pDR;
      }

      delete pEnum;
   }
   else if (0 == strcasecmp (JOBPROP_PRINTMODE, pszKey))
   {
      DevicePrintMode *pDPM = getCurrentPrintMode ();

      pEnum = pDPM->getEnumeration ();

      while (pEnum->hasMoreElements ())
      {
         pDPM = (DevicePrintMode *)pEnum->nextElement ();

         pList->addElement (DevicePrintMode::IDToName (pDPM->getID ()));

         delete pDPM;
      }

      delete pEnum;
   }
   else
   {
      if (pInstance_d)
      {
         delete pList;

         return pInstance_d->listKeyValues (pszKey);
      }
   }

   return pList;
}

string * PrintDevice::
getJobPropertyType (char *pszKey)
{
   ostringstream oss;

   if (0 == strcasecmp (JOBPROP_ORIENTATION, pszKey))
   {
      DeviceOrientation *pDO = getDefaultOrientation ();

      oss << "string " << DeviceOrientation::IDToName (pDO->getID ());

      delete pDO;
   }
   else if (0 == strcasecmp (JOBPROP_DITHER, pszKey))
   {
      char *pszDitherID = getDefaultDitherID ();

      oss << "string " << pszDitherID;
   }
   else if (0 == strcasecmp (JOBPROP_FORM, pszKey))
   {
      DeviceForm *pDF = getDefaultForm ();

      oss << "string " << DeviceForm::IDToName (pDF->getID ());

      delete pDF;
   }
   else if (0 == strcasecmp (JOBPROP_TRAY, pszKey))
   {
      DeviceTray *pDT = getDefaultTray ();

      oss << "string " << DeviceTray::IDToName (pDT->getID ());

      delete pDT;
   }
   else if (0 == strcasecmp (JOBPROP_MEDIA, pszKey))
   {
      DeviceMedia *pDM = getDefaultMedia ();

      oss << "string " << DeviceMedia::IDToName (pDM->getID ());

      delete pDM;
   }
   else if (0 == strcasecmp (JOBPROP_RESOLUTION, pszKey))
   {
      DeviceResolution *pDR = getDefaultResolution ();

      oss << "string " << DeviceResolution::IDToName (pDR->getID ());

      delete pDR;
   }
   else if (0 == strcasecmp (JOBPROP_PRINTMODE, pszKey))
   {
      DevicePrintMode *pDPM = getDefaultPrintMode ();

      oss << "string " << DevicePrintMode::IDToName (pDPM->getID ());

      delete pDPM;
   }
   else
   {
      if (pInstance_d)
      {
         string *pDeviceRet = pInstance_d->getJobPropertyType (pszKey);

         if (!pDeviceRet)
            return 0;

         oss << *pDeviceRet;

         delete pDeviceRet;
      }
   }

   string *pRet = new string (oss.str ());

   return pRet;
}

string * PrintDevice::
getJobProperty (char *pszKey)
{
   ostringstream oss;

   if (0 == strcasecmp (JOBPROP_ORIENTATION, pszKey))
   {
      oss << DeviceOrientation::IDToName (pOrientation_d->getID ());
   }
   else if (0 == strcasecmp (JOBPROP_DITHER, pszKey))
   {
      oss << pszDitherID_d;
   }
   else if (0 == strcasecmp (JOBPROP_FORM, pszKey))
   {
      oss << DeviceForm::IDToName (pForm_d->getID ());
   }
   else if (0 == strcasecmp (JOBPROP_TRAY, pszKey))
   {
      oss << DeviceTray::IDToName (pTray_d->getID ());
   }
   else if (0 == strcasecmp (JOBPROP_MEDIA, pszKey))
   {
      oss << DeviceMedia::IDToName (pMedia_d->getID ());
   }
   else if (0 == strcasecmp (JOBPROP_RESOLUTION, pszKey))
   {
      oss << DeviceResolution::IDToName (pResolution_d->getID ());
   }
   else if (0 == strcasecmp (JOBPROP_PRINTMODE, pszKey))
   {
      oss << DevicePrintMode::IDToName (pPrintMode_d->getID ());
   }
   else
   {
      if (pInstance_d)
      {
         string *pDeviceRet = pInstance_d->getJobProperty (pszKey);

         if (!pDeviceRet)
            return 0;

         oss << *pDeviceRet;

         delete pDeviceRet;
      }
   }

   string *pRet = new string (oss.str ());

   return pRet;
}

string * PrintDevice::
translateKeyValue (char *pszKey,
                   char *pszValue)
{
   string *pRet = 0;

   if (0 == strcasecmp (JOBPROP_ORIENTATION, pszKey))
   {
      char *pszXLateKey = StringResource::getString (getLanguageResource (),
                                                     StringResource::STRINGGROUP_DEVICE_COMMON,
                                                     StringResource::DEVICE_COMMON_ORIENTATION);

      DeviceOrientation *pDO = pOrientation_d->create (pszValue);

      if (  pszXLateKey
         && pDO
         )
      {
         pRet = new string (pszXLateKey);
         *pRet += "=";
         *pRet += pDO->getName ();
      }

      delete pDO;
   }
   else if (0 == strcasecmp (JOBPROP_DITHER, pszKey))
   {
      char *pszXLateKey = StringResource::getString (getLanguageResource (),
                                                     StringResource::STRINGGROUP_DEVICE_COMMON,
                                                     StringResource::DEVICE_COMMON_DITHER);

      char *pszXLateValue = StringResource::getString (getLanguageResource (),
                                                       StringResource::STRINGGROUP_DITHERS,
                                                       pszValue);


      if (  pszXLateKey
         && pszXLateValue
         )
      {
         pRet = new string (pszXLateKey);
         *pRet += "=";
         *pRet += pszXLateValue;
      }
   }
   else if (0 == strcasecmp (JOBPROP_FORM, pszKey))
   {
      char *pszXLateKey = StringResource::getString (getLanguageResource (),
                                                     StringResource::STRINGGROUP_DEVICE_COMMON,
                                                     StringResource::DEVICE_COMMON_FORM);

      DeviceForm *pDF = pForm_d->create (pszValue);

      if (  pszXLateKey
         && pDF
         )
      {
         pRet = new string (pszXLateKey);
         *pRet += "=";
         *pRet += pDF->getName ();
      }

      delete pDF;
   }
   else if (0 == strcasecmp (JOBPROP_TRAY, pszKey))
   {
      char *pszXLateKey = StringResource::getString (getLanguageResource (),
                                                     StringResource::STRINGGROUP_DEVICE_COMMON,
                                                     StringResource::DEVICE_COMMON_TRAY);

      DeviceTray *pDT = pTray_d->create (pszValue);

      if (  pszXLateKey
         && pDT
         )
      {
         pRet = new string (pszXLateKey);
         *pRet += "=";
         *pRet += pDT->getName ();
      }

      delete pDT;
   }
   else if (0 == strcasecmp (JOBPROP_MEDIA, pszKey))
   {
      char *pszXLateKey = StringResource::getString (getLanguageResource (),
                                                     StringResource::STRINGGROUP_DEVICE_COMMON,
                                                     StringResource::DEVICE_COMMON_MEDIA);

      DeviceMedia *pDM = pMedia_d->create (pszValue);

      if (  pszXLateKey
         && pDM
         )
      {
         pRet = new string (pszXLateKey);
         *pRet += "=";
         *pRet += pDM->getName ();
      }

      delete pDM;
   }
   else if (0 == strcasecmp (JOBPROP_RESOLUTION, pszKey))
   {
      char *pszXLateKey = StringResource::getString (getLanguageResource (),
                                                     StringResource::STRINGGROUP_DEVICE_COMMON,
                                                     StringResource::DEVICE_COMMON_RESOLUTION);

      DeviceResolution *pDR = pResolution_d->create (pszValue);

      if (  pszXLateKey
         && pDR
         )
      {
         pRet = new string (pszXLateKey);
         *pRet += "=";
         *pRet += pDR->getName ();
      }

      delete pDR;
   }
   else if (0 == strcasecmp (JOBPROP_PRINTMODE, pszKey))
   {
      char *pszXLateKey = StringResource::getString (getLanguageResource (),
                                                     StringResource::STRINGGROUP_DEVICE_COMMON,
                                                     StringResource::DEVICE_COMMON_PRINTMODE);

      DevicePrintMode *pDPM = pPrintMode_d->create (pszValue);

      if (  pszXLateKey
         && pDPM
         )
      {
         pRet = new string (pszXLateKey);
         *pRet += "=";
         *pRet += pDPM->getName ();
      }

      delete pDPM;
   }
   else
   {
      if (pInstance_d)
      {
         string *pDeviceRet = pInstance_d->translateKeyValue (pszKey,
                                                              pszValue);

         return pDeviceRet;
      }
   }

   return pRet;
}

/*8<--------8<--------8<--------8<--------8<--------8<--------8<--------*/

void PrintDevice::
loadLibrary (char *pszLibraryName)
{
   if (pszLoadedLibrary_d)
   {
      free (pszLoadedLibrary_d);
      pszLoadedLibrary_d = 0;
   }

   if (  !pszLibraryName
      || !*pszLibraryName
      )
      return;

   hModLibrary_d = g_module_open (pszLibraryName, (GModuleFlags)0);

#ifndef RETAIL
   if (DebugOutput::shouldOutputPrintDevice ()) DebugOutput::getErrorStream () << "PrintDevice::loadLibrary: g_module_open (" << pszLibraryName << ") = 0x" << hex << (int)hModLibrary_d << dec << endl;
#endif

   if (!hModLibrary_d)
      // Error!
      return;

   pszLoadedLibrary_d = (char *)malloc (strlen (pszLibraryName) + 1);
   if (pszLoadedLibrary_d)
   {
      strcpy (pszLoadedLibrary_d, pszLibraryName);
   }
   else
   {
      g_module_close (hModLibrary_d);
   }
}

void * PrintDevice::
dlsym (char *pszLibraryName,
       char *pszSymbol)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputPrintDevice ()) DebugOutput::getErrorStream () << "PrintDevice::dlsym (" << pszLibraryName << ", " << pszSymbol << ")" << endl;
#endif

   if (!pszLibraryName)
      // Nothing loaded
      return 0;

   if (0 != strcmp (pszLibraryName, pszLoadedLibrary_d))
      // Wrong library
      return 0;

#ifndef RETAIL
   if (DebugOutput::shouldOutputPrintDevice ()) DebugOutput::getErrorStream () << "PrintDevice::dlsym hModLibrary_d = 0x" << hex << (int)hModLibrary_d << dec << endl;
#endif

   if (!hModLibrary_d)
      // Error!
      return 0;

   gpointer ptr = NULL;
   bool     rc;

   rc = ::g_module_symbol (hModLibrary_d, pszSymbol, &ptr);

#ifndef RETAIL
   if (DebugOutput::shouldOutputPrintDevice ()) DebugOutput::getErrorStream () << "PrintDevice::g_module_symbol returns " << rc << endl;
#endif

   return ptr;
}

/*8<--------8<--------8<--------8<--------8<--------8<--------8<--------*/

string PrintDevice::
toString (ostringstream& oss)
{
   oss << hex
       << "{ lCapabilities_d = " << lCapabilities_d << ", "
       << "lRasterCapabilities_d = " << lRasterCapabilities_d << ", "
       << "lDeviceOptions_d = " << lDeviceOptions_d << ", "
       << "lDeviceOptions2_d = " << lDeviceOptions2_d << ", "
       << dec;

   {
      ostringstream oss2;
      if (pInstance_d)
         oss << "pInstance_d = " << pInstance_d->toString (oss2) << ", ";
      else
         oss << "pInstance_d = (null), ";
   }

   {
      ostringstream oss2;
      if (pBlitter_d)
         oss << "pBlitter_d = " << pBlitter_d->toString (oss2) << ", ";
      else
         oss << "pBlitter_d = (null), ";
   }

   {
      ostringstream oss2;

      if (pOrientation_d)
         oss << "pOrientation_d = " << pOrientation_d->toString (oss2) << ", ";
      else
         oss << "pOrientation_d = (null), ";
   }

   {
      ostringstream oss2;
      if (pForm_d)
         oss << "pForm_d = " << pForm_d->toString (oss2) << ", ";
      else
         oss << "pForm_d = (null), ";
   }

   {
      ostringstream oss2;
      if (pTray_d)
         oss << "pTray_d = " << pTray_d->toString (oss2) << ", ";
      else
         oss << "pTray_d = (null), ";
   }

   {
      ostringstream oss2;
      if (pMedia_d)
         oss << "pMedia_d = " << pMedia_d->toString (oss2) << ", ";
      else
         oss << "pMedia_d = (null), ";
   }

   {
      ostringstream oss2;
      if (pResolution_d)
         oss << "pResolution_d = " << pResolution_d->toString (oss2) << ", ";
      else
         oss << "pResolution_d = (null), ";
   }

   if (pCommand_d)
      oss << "pCommand_d = " << pCommand_d << ", ";
   else
      oss << "pCommand_d = (null), ";

   {
      ostringstream oss2;
      if (pPrintMode_d)
         oss << "pPrintMode_d = " << pPrintMode_d->toString (oss2) << ", ";
      else
         oss << "pPrintMode_d = (null), ";
   }

   {
      ostringstream oss2;
      if (pData_d)
         oss << "pData_d = " << pData_d->toString (oss2) << ", ";
      else
         oss << "pData_d = (null), ";
   }

   {
      ostringstream oss2;
      if (pString_d)
         oss << "pString_d = " << pString_d->toString (oss2);
      else
         oss << "pString_d = (null)";
   }

   oss << " }";

   return oss.str ();
}

/* Provide a way to print out class data
*/
ostream&
operator<< (ostream& os, const PrintDevice& const_self)
{
   PrintDevice&  self = const_cast<PrintDevice&>(const_self);
   ostringstream oss;

   os << self.toString (oss);

   return os;
}
