/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include <UPDF_DeviceOrientation.hpp>
#include <UPDF_DeviceForm.hpp>
#include <UPDF_DeviceTray.hpp>
#include <UPDF_DeviceMedia.hpp>
#include <UPDF_DeviceResolution.hpp>
#include <UPDF_DevicePrintMode.hpp>
//#include <UPDF_DeviceCommands.hpp>
#include <UPDF_Instance.hpp>
#include <UPDF_Blitter.hpp>
#include <UPDF_Device.hpp>
#include <JobProperties.hpp>

#include <glob.h>

static bool
isDeviceConfigurationXMLFile (xmlDocPtr doc)
{
   xmlNodePtr nodeRoot = xmlDocGetRootElement (doc);
   bool       fRc      = false;

   if (nodeRoot)
   {
      xmlNodePtr nodeElm = 0;

      nodeElm = firstNode (nodeRoot);

      if (nodeElm)
      {
         if (0 == strcmp ((char *)nodeElm->name, "DeviceConfiguration"))
         {
            fRc = true;
         }
      }
   }

   return fRc;
}

static xmlDocPtr
getDeviceConfigurationFromFile (char *pszMasterXMLFile)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDevice ()) DebugOutput::getErrorStream () << "UPDF_Device: " << __FUNCTION__ << ": pszMasterXMLFile = " << pszMasterXMLFile << endl;
#endif

   xmlDocPtr  docRet   = 0;

   docRet = xmlParseFile (pszMasterXMLFile);
///xmlValidateDocument (&validator, docRet); // @TBD

   if (docRet)
   {
      if (isDeviceConfigurationXMLFile (docRet))
      {
         return docRet;
      }

      xmlFreeDoc (docRet);
   }

   return 0;
}

static xmlDocPtr
getDefaultDeviceConfiguration ()
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDevice ()) DebugOutput::getErrorStream () << "UPDF_Device: " << __FUNCTION__ << ": " << endl;
#endif

   xmlDocPtr docRet  = 0;
   glob_t    globbuf;
   int       rc;

   memset (&globbuf, 0, sizeof (globbuf));
   rc = glob ("*.xml", 0, NULL, &globbuf);

   if (0 == rc)
   {
      for (int i = 0; i < (int)globbuf.gl_pathc; i++)
      {
         docRet = xmlParseFile (globbuf.gl_pathv[i]);
/////////xmlValidateDocument (&validator, docRet); // @TBD

         if (docRet)
         {
            if (isDeviceConfigurationXMLFile (docRet))
            {
               break;
            }

            xmlFreeDoc (docRet);
         }
      }
   }

   globfree (&globbuf);

   return docRet;
}

static xmlDocPtr
getUDRFromDeviceConfiguration (xmlDocPtr docDC)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDevice ()) DebugOutput::getErrorStream () << "UPDF_Device: " << __FUNCTION__ << ": " << endl;
#endif

   xmlDocPtr   docRet     = 0;
   xmlNodePtr  nodeRootDC = 0;
   char       *pszUDR     = 0;

   nodeRootDC = xmlDocGetRootElement (docDC);

   pszUDR = (char *)xmlGetProp (nodeRootDC, (xmlChar *)"Unit_Description_Reference");

   if (pszUDR)
   {
      string stringUDR = pszUDR;

      stringUDR += ".xml";

      docRet = xmlParseFile (stringUDR.c_str ());

      if (docRet)
      {
/////////xmlValidateDocument (&validator, docRet); // @TBD
      }

      free (pszUDR);
   }

   return docRet;
}

static bool
getOmniNames (xmlDocPtr   docDC,
              char      **ppszDriverName,
              char      **ppszDeviceName,
              char      **ppszShortName)
{
   bool fRc = false;

   if (  docDC
      && ppszDriverName
      )
   {
      xmlNodePtr nodeDeviceHeader = xmlDocGetRootElement (docDC);

      nodeDeviceHeader = UPDF_Device::findEntry (nodeDeviceHeader,
                                                 "Device_Header",
                                                 DebugOutput::shouldOutputUPDFDevice ());

      if (nodeDeviceHeader)
      {
         *ppszDriverName = (char *)xmlGetProp (nodeDeviceHeader,
                                               (xmlChar *)"Manufacturer");

         *ppszDeviceName = (char *)xmlGetProp (nodeDeviceHeader,
                                               (xmlChar *)"ProductName");

         if (  *ppszDeviceName
            && 0 < strlen (*ppszDeviceName)
            )
         {
            *ppszShortName = (char *)malloc (strlen (*ppszDeviceName) + 1);

            if (*ppszShortName)
            {
               char *pszFrom = *ppszDeviceName;
               char *pszTo   = *ppszShortName;

               while (*pszFrom)
               {
                  switch (*pszFrom)
                  {
                  case '-':
                  case '+':
                  case '(':
                  case ')':
                  case ' ':
                  case '/':
                  case '\\':
                  {
                     *pszTo++ = '_';
                     pszFrom++;
                     break;
                  }
                  default:
                  {
                     *pszTo++ = *pszFrom++;
                  }
                  }
               }
            }
         }

         if (  *ppszDriverName
            && *ppszDeviceName
            && *ppszShortName
            )
         {
            fRc = true;
         }
      }
   }

#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDevice ())
      DebugOutput::getErrorStream () << "UPDF_Device: "
                                     << __FUNCTION__
                                     << ": *ppszDriverName = "
                                     << (ppszDriverName ? *ppszDriverName : "(null)")
                                     << ", *ppszDeviceName = "
                                     << (ppszDeviceName ? *ppszDeviceName : "(null)")
                                     << ", *ppszShortName = "
                                     << (ppszShortName ? *ppszShortName : "(null)")
                                     << endl;
#endif

   return fRc;
}

Device *
newDeviceW_Advanced (bool fAdvanced)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDevice ()) DebugOutput::getErrorStream () << "UPDF_Device: " << __FUNCTION__ << ": fAdvanced = " << fAdvanced << endl;
#endif

   xmlDocPtr docDC          = getDefaultDeviceConfiguration ();
   xmlDocPtr docUDR         = 0;
   char      *pszDriverName = 0;
   char      *pszDeviceName = 0;
   char      *pszShortName  = 0;

   if (docDC)
   {
      docUDR = getUDRFromDeviceConfiguration (docDC);
   }

   if (  docDC
      && docUDR
      && getOmniNames (docUDR,
                       &pszDriverName,
                       &pszDeviceName,
                       &pszShortName)
      )
   {
      PrintDevice *pPrintDevice = new UPDF_Device (docDC,
                                                   docUDR,
                                                   pszDriverName,
                                                   pszDeviceName,
                                                   pszShortName);

      pPrintDevice->initialize ();

      if (fAdvanced)
         return pPrintDevice;
      else
         return new OmniProxy (pPrintDevice);
   }

   return 0;
}

Device *
newDeviceW_JopProp_Advanced (char *pszJobProperties, bool fAdvanced)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDevice ()) DebugOutput::getErrorStream () << "UPDF_Device: " << __FUNCTION__ << ": pszJobProperties = \"" << pszJobProperties << "\", fAdvanced = " << fAdvanced << endl;
#endif

   xmlDocPtr  docDC         = 0;
   xmlDocPtr  docUDR        = 0;
   char      *pszDriverName = 0;
   char      *pszDeviceName = 0;
   char      *pszShortName  = 0;

   if (  pszJobProperties
      && *pszJobProperties
      )
   {
      JobProperties jobProp (pszJobProperties);

      while (jobProp.hasKeyNameLeft ("UPDFDeviceName"))
      {
         char *pszFileName = jobProp.getCurrentValue ();

         docDC = getDeviceConfigurationFromFile (pszFileName);

         if (docDC)
         {
            break;
         }
         else
         {
            DebugOutput::getErrorStream () << "Error: Could not load Device Configuration from " << pszFileName << endl;
         }

         jobProp.getNextKey ("UPDFDeviceName");
      }
   }

   if (!docDC)
   {
      docDC = getDefaultDeviceConfiguration ();
   }

   if (docDC)
   {
      docUDR = getUDRFromDeviceConfiguration (docDC);
   }

   if (  docDC
      && docUDR
      && getOmniNames (docUDR,
                       &pszDriverName,
                       &pszDeviceName,
                       &pszShortName)
      )
   {
      PrintDevice *pPrintDevice = new UPDF_Device (docDC,
                                                   docUDR,
                                                   pszDriverName,
                                                   pszDeviceName,
                                                   pszShortName,
                                                   pszJobProperties);

      pPrintDevice->initialize ();

      if (fAdvanced)
         return pPrintDevice;
      else
         return new OmniProxy (pPrintDevice);
   }

   if (docDC)
   {
      xmlFreeDoc (docDC);
   }

   return 0;
}

void
deleteDevice (Device *pDevice)
{
   delete pDevice;
}

static void
MyValidityErrorFunc (void *ctx, const char *msg, ...)
{
   va_list list;

   va_start (list, msg);

   printf ("Error: ");
   vprintf (msg, list);

   va_end (list);

   if (ctx)
   {
      // @TBD
   }
}

static void
MyValidityWarningFunc (void *ctx, const char *msg, ...)
{
   va_list list;

   va_start (list, msg);

   printf ("Warning: ");
   vprintf (msg, list);

   va_end (list);
}

void UPDF_Device::
commonInit (xmlDocPtr docDC,
            xmlDocPtr docUDR)
{
   xmlValidCtxt validator = {
      this,
      MyValidityErrorFunc,
      MyValidityWarningFunc
   };

   setCapabilities ( Capability::MIRROR
                   | Capability::MONOCHROME
                   );

   setRasterCapabilities ( TOP_TO_BOTTOM
                         );

   docDC_d  = docDC;
   docUDR_d = docUDR;

   if (  docDC_d
      && docUDR_d
      )
   {
      nodeRootDC_d  = xmlDocGetRootElement (docDC_d);
      nodeRootUDR_d = xmlDocGetRootElement (docUDR_d);

      xmlNodePtr nodeLocale = 0;

      if ((nodeLocale = findEntry (nodeRootDC_d, "Locale", DebugOutput::shouldOutputUPDFDevice ())) != 0)
      {
         char *pszLocale = (char *)xmlGetProp (nodeLocale, (xmlChar *)"Locale_Reference");

         if (pszLocale)
         {
            string stringLocale = pszLocale;

            stringLocale += ".xml";

            docLocale_d = xmlParseFile (stringLocale.c_str ());

            if (docLocale_d)
            {
               xmlValidateDocument (&validator, docLocale_d);

               nodeRootLocale_d = xmlDocGetRootElement (docLocale_d);
            }

            free (pszLocale);
         }
      }

      char *pszCS = (char *)xmlGetProp (nodeRootDC_d, (xmlChar *)"CommandSequences_Reference");

      if (pszCS)
      {
         string stringCS = pszCS;

         stringCS += ".xml";

         docCS_d = xmlParseFile (stringCS.c_str ());

         if (docCS_d)
         {
            xmlValidateDocument (&validator, docCS_d);

            nodeRootCS_d = xmlDocGetRootElement (docCS_d);
         }

         free (pszCS);
      }
   }

   pInstance_d = new UPDF_Instance (this);

   setDeviceInstance (pInstance_d);
   setDeviceBlitter (new UPDF_Blitter (this));

   setPDL (new PDL (PDL::PDL_Epson,PDL::LEVEL_ESCP_2,1,0)); // @TBD

   iVirtualUnits_d = setVirtualUnits ();
}

UPDF_Device::
UPDF_Device (xmlDocPtr  docDC,
             xmlDocPtr  docUDR,
             char      *pszDriverName,
             char      *pszDeviceName,
             char      *pszShortName)
   : PrintDevice (pszDriverName,
                  pszDeviceName,
                  pszShortName,
                  "orientation=ORIENTATION_PORTRAIT dither=DITHER_STUCKI_DIFFUSION form=FORM_LETTER tray=TRAY_AUTO media=MEDIA_PLAIN resolution=RESOLUTION_300_X_300 printmode=PRINT_MODE_1_ANY")
{
   pszDriverName_d = pszDriverName;
   pszDeviceName_d = pszDeviceName;
   pszShortName_d  = pszShortName;

   commonInit (docDC, docUDR);
}

UPDF_Device::
UPDF_Device (xmlDocPtr  docDC,
             xmlDocPtr  docUDR,
             char      *pszDriverName,
             char      *pszDeviceName,
             char      *pszShortName,
             char      *pszJobProperties)
   : PrintDevice (pszDriverName,
                  pszDeviceName,
                  pszShortName,
                  pszJobProperties)
{
   pszDriverName_d = pszDriverName;
   pszDeviceName_d = pszDeviceName;
   pszShortName_d  = pszShortName;

   commonInit (docDC, docUDR);
}

UPDF_Device::
~UPDF_Device ()
{
   if (docDC_d)
   {
      xmlFreeDoc (docDC_d);
      docDC_d = 0;
   }
   if (docUDR_d)
   {
      xmlFreeDoc (docUDR_d);
      docUDR_d = 0;
   }
   if (docLocale_d)
   {
      xmlFreeDoc (docLocale_d);
      docLocale_d = 0;
   }
   if (docCS_d)
   {
      xmlFreeDoc (docCS_d);
      docCS_d = 0;
   }
   if (pszDriverName_d)
   {
      free (pszDriverName_d);
      pszDriverName_d = 0;
   }
   if (pszDeviceName_d)
   {
      free (pszDeviceName_d);
      pszDeviceName_d = 0;
   }
   if (pszShortName_d)
   {
      free (pszShortName_d);
      pszShortName_d = 0;
   }
}

DeviceOrientation * UPDF_Device::
getDefaultOrientation ()
{
   if (pInstance_d)
   {
      char *pszValue = pInstance_d->getXMLObjectValue ("MediaPageOrientation",
                                                       "ID");

      if (pszValue)
      {
         DeviceOrientation *pDORet = 0;

         pDORet = UPDF_DeviceOrientation::create (this, pszValue);

         free (pszValue);

         if (pDORet)
            return pDORet;
      }
   }

   return UPDF_DeviceOrientation::create (this, DeviceOrientation::ORIENTATION_PORTRAIT);
}

char * UPDF_Device::
getDefaultDitherID ()
{
   return "DITHER_STUCKI_DIFFUSION";
}

DeviceForm * UPDF_Device::
getDefaultForm ()
{
   if (pInstance_d)
   {
      char *pszValue = pInstance_d->getXMLObjectValue ("MediaSize",
                                                       "ID");

      if (pszValue)
      {
         DeviceForm *pDFRet = 0;

         pDFRet = UPDF_DeviceForm::create (this, pszValue);

         free (pszValue);

         if (pDFRet)
            return pDFRet;
      }
   }

   return UPDF_DeviceForm::create (this, DeviceForm::FORM_LETTER);
}

DeviceTray * UPDF_Device::
getDefaultTray ()
{
   if (pInstance_d)
   {
      char *pszValue = pInstance_d->getXMLObjectValue ("MediaSource",
                                                       "ID");

      if (pszValue)
      {
         DeviceTray *pDTRet = 0;

         pDTRet = UPDF_DeviceTray::create (this, pszValue);

         free (pszValue);

         if (pDTRet)
            return pDTRet;
      }
   }

   return UPDF_DeviceTray::create (this, DeviceTray::TRAY_AUTO);
}

DeviceMedia * UPDF_Device::
getDefaultMedia ()
{
   if (pInstance_d)
   {
      char *pszValue = pInstance_d->getXMLObjectValue ("MediaType",
                                                       "ID");

      if (pszValue)
      {
         DeviceMedia *pDMRet = 0;

         pDMRet = UPDF_DeviceMedia::create (this, pszValue);

         free (pszValue);

         if (pDMRet)
            return pDMRet;
      }
   }

   return UPDF_DeviceMedia::create (this, DeviceMedia::MEDIA_PLAIN);
}

DeviceResolution * UPDF_Device::
getDefaultResolution ()
{
   if (pInstance_d)
   {
      char *pszValue = pInstance_d->getXMLObjectValue ("DeviceResolution",
                                                       "ID");

      if (pszValue)
      {
         DeviceResolution *pDRRet = 0;

         pDRRet = UPDF_DeviceResolution::create (this, pszValue);

         free (pszValue);

         if (pDRRet)
            return pDRRet;
      }
   }

   return UPDF_DeviceResolution::create (this, DeviceResolution::RESOLUTION_300_X_300);
}

DevicePrintMode * UPDF_Device::
getDefaultPrintMode ()
{
   return UPDF_DevicePrintMode::create (this, DevicePrintMode::PRINT_MODE_1_ANY);
}

DeviceCommand * UPDF_Device::
getDefaultCommands ()
{
   return 0;
///return new UPDF_DeviceCommands (this);
}

DeviceData * UPDF_Device::
getDefaultData ()
{
   // @TBD
   return 0;
}

DeviceString * UPDF_Device::
getDefaultString ()
{
   // @TBD
   return 0;
}

DeviceGamma * UPDF_Device::
getCurrentGamma ()
{
   // @TBD
   return 0;
}

UPDF_Device * UPDF_Device::
isAUPDF_Device (Device *pDevice)
{
   return dynamic_cast<UPDF_Device *>(pDevice);
}

xmlNodePtr UPDF_Device::
findDCEntry (xmlNodePtr  root,
             char       *pszName,
             bool        fShouldDebugOutput)
{
   if (root == 0)
   {
      root = nodeRootDC_d;
   }

   return findEntry (root, pszName, fShouldDebugOutput);
}

xmlNodePtr UPDF_Device::
findUDREntry (xmlNodePtr  root,
              char       *pszName,
              bool        fShouldDebugOutput)
{
   if (root == 0)
   {
      root = nodeRootUDR_d;
   }

   return findEntry (root, pszName, fShouldDebugOutput);
}

xmlNodePtr UPDF_Device::
findLocaleEntry (xmlNodePtr  root,
                 char       *pszName,
                 bool        fShouldDebugOutput)
{
   if (root == 0)
   {
      root = nodeRootLocale_d;
   }

   return findEntry (root, pszName, fShouldDebugOutput);
}

xmlNodePtr UPDF_Device::
findCSEntry (xmlNodePtr  root,
             char       *pszName,
             bool        fShouldDebugOutput)
{
   if (root == 0)
   {
      root = nodeRootCS_d;
   }

   return findEntry (root, pszName, fShouldDebugOutput);
}

xmlNodePtr UPDF_Device::
findDCEntryKeyValue (char *pszKey,
                     char *pszValue,
                     bool  fShouldDebugOutput)
{
   return findEntryKeyValue (nodeRootDC_d, pszKey, pszValue, fShouldDebugOutput);
}

xmlNodePtr UPDF_Device::
findUDREntryKeyValue (char *pszKey,
                      char *pszValue,
                      bool  fShouldDebugOutput)
{
   return findEntryKeyValue (nodeRootUDR_d, pszKey, pszValue, fShouldDebugOutput);
}

xmlNodePtr UPDF_Device::
findLocaleEntryKeyValue (char *pszKey,
                         char *pszValue,
                         bool  fShouldDebugOutput)
{
   return findEntryKeyValue (nodeRootLocale_d, pszKey, pszValue, fShouldDebugOutput);
}

xmlNodePtr UPDF_Device::
findCSEntryKeyValue (char *pszKey,
                     char *pszValue,
                     bool  fShouldDebugOutput)
{
   return findEntryKeyValue (nodeRootCS_d, pszKey, pszValue, fShouldDebugOutput);
}

xmlNodePtr UPDF_Device::
findEntry (xmlNodePtr  root,
           char       *pszName,
           bool        fShouldDebugOutput)
{
   if (root == 0)
      return 0;

   xmlNodePtr elm = firstNode (root->xmlChildrenNode);

   while (elm != 0)
   {
#ifndef RETAIL
//////if (DebugOutput::shouldOutputUPDFDevice ()) DebugOutput::getErrorStream () << "UPDF_Device::" << __FUNCTION__ << ": " << elm->name << " == " << pszName << endl;
#endif

      if (0 == strcmp ((char *)elm->name, pszName))
      {
         break;
      }

      elm = nextNode (elm);
   }

#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDevice ())
   {
      if (elm)
      {
         DebugOutput::getErrorStream () << "UPDF_Device::" << __FUNCTION__ << ": Found ";
      }
      else
      {
         DebugOutput::getErrorStream () << "UPDF_Device::" << __FUNCTION__ << ": Did not find ";
      }
      DebugOutput::getErrorStream () << pszName << endl;
   }
#endif

   return elm;
}

xmlNodePtr UPDF_Device::
findEntryKeyValue (xmlNodePtr  root,
                   char       *pszKey,
                   char       *pszValue,
                   bool        fShouldDebugOutput)
{
   if (root == 0)
      return 0;

   xmlNodePtr elm      = root;
   xmlNodePtr elmFound = 0;

   while (elm != 0)
   {
      char *pszElmValue = (char *)xmlGetProp (elm, (xmlChar *)pszKey);

      if (  pszElmValue
         && 0 == strcmp (pszElmValue, pszValue)
         )
      {
         elmFound = elm;
      }

      if (pszElmValue)
      {
         free (pszElmValue);
      }

      if (  !elmFound
         && elm->xmlChildrenNode
         )
      {
         elmFound = findEntryKeyValue (elm->xmlChildrenNode, pszKey, pszValue, false);
      }

      if (elmFound)
      {
         elm = elmFound;
         break;
      }

      elm = nextNode (elm);
   }

#ifndef RETAIL
   if (fShouldDebugOutput)
   {
      if (elm)
      {
         DebugOutput::getErrorStream () << "UPDF_Device::" << __FUNCTION__ << ": Found ";
      }
      else
      {
         DebugOutput::getErrorStream () << "UPDF_Device::" << __FUNCTION__ << ": Did not find ";
      }
      DebugOutput::getErrorStream () << pszValue << endl;
   }
#endif

   return elm;
}

int UPDF_Device::
setVirtualUnits ()
{
   xmlNodePtr nodeHeader    = 0;
   int        iVirtualUnits = 0;

   if (  ((nodeHeader = findUDREntry (nodeHeader, "Device_Features",   DebugOutput::shouldOutputUPDFDevice ())) != 0)
      && ((nodeHeader = findUDREntry (nodeHeader, "PrintCapabilities", DebugOutput::shouldOutputUPDFDevice ())) != 0)
      && ((nodeHeader = findUDREntry (nodeHeader, "Header",            DebugOutput::shouldOutputUPDFDevice ())) != 0)
      )
   {
      char *pszVirtualUnits = (char *)xmlGetProp (nodeHeader,
                                                  (xmlChar *)"VirtualUnits");
      if (pszVirtualUnits)
      {
         iVirtualUnits = atoi (pszVirtualUnits);

         free (pszVirtualUnits);
      }
   }

   return iVirtualUnits;
}

int UPDF_Device::
getVirtualUnits ()
{
   return iVirtualUnits_d;
}

xmlNodePtr UPDF_Device::
getRootDC ()
{
   return nodeRootDC_d;
}

xmlNodePtr UPDF_Device::
getRootUDR ()
{
   return nodeRootUDR_d;
}

xmlNodePtr UPDF_Device::
getRootLocale ()
{
   return nodeRootLocale_d;
}

xmlNodePtr UPDF_Device::
getRootCS ()
{
   return nodeRootCS_d;
}
