/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <UPDF_DeviceForm.hpp>

/*
*/
static int     getOmniFormID             (xmlNodePtr      node);

UPDF_DeviceForm::
UPDF_DeviceForm (Device      *pDevice,
                 int          id,
                 int          iCapabilities,
                 BinaryData  *data,
                 HardCopyCap *hcInfo,
                 xmlNodePtr   node)
   : DeviceForm (pDevice,
                 id,
                 iCapabilities,
                 data,
                 hcInfo)
{
   node_d = node;
}

UPDF_DeviceForm::
~UPDF_DeviceForm ()
{
}

DeviceForm * UPDF_DeviceForm::
createV (int id)
{
   return UPDF_DeviceForm::create (pDevice_d, id);
}

DeviceForm * UPDF_DeviceForm::
create (Device *pDevice, char *pszId)
{
   int id = DeviceForm::nameToID (pszId);

   if (DeviceForm::FORM_UNLISTED != id)
   {
      return create (pDevice, id);
   }
   else
   {
      UPDF_Device *pUPDFDevice = UPDF_Device::isAUPDF_Device (pDevice);
      xmlNodePtr   nodeForms   = 0;

      nodeForms = findForms (pUPDFDevice);

      if (nodeForms)
      {
         xmlNodePtr nodeItem  = firstNode (nodeForms->xmlChildrenNode);
         xmlNodePtr nodeFound = 0;

         while (nodeItem != 0)
         {
            char *pszNodeId = (char *)xmlGetProp (nodeItem, (xmlChar *)"ID");

            if (pszNodeId)
            {
               if (0 == strcmp (pszNodeId, pszId))
               {
                  nodeFound = nodeItem;
               }

               free (pszNodeId);
            }

            if (nodeFound)
            {
               id = getOmniFormID (nodeFound);

               if (DeviceForm::FORM_UNLISTED != id)
               {
                  return create (pDevice, id);
               }

               nodeFound = 0;
            }

            nodeItem = nextNode (nodeItem);
         }
      }

      return 0;
   }
}

static int
getOmniFormID (xmlNodePtr node)
{
   int id = DeviceForm::FORM_UNLISTED;

   if (!node)
   {
      return id;
   }

   char *pszID = (char *)xmlGetProp (node, (xmlChar *)"Predefined_ID");

   if (pszID)
   {
      id = DeviceForm::nameToID (pszID);

      free (pszID);
   }

   return id;
}

static xmlNodePtr
skipInvalidForms (xmlNodePtr node)
{
   if (!node)
   {
      return 0;
   }

   int id = getOmniFormID (node);

   while (  node
           && (  id == DeviceForm::FORM_UNLISTED
              || id == DeviceForm::FORM_NONE
              )
           )
   {
      node = nextNode (node);
      id   = getOmniFormID (node);
   }

   return node;
}

static int
converUPDFMarginToOmniMargin (char *pszMarginInfo)
{
   static char *pszMargin     = "margin_";
   int          cbMargin      = 0;
   char        *pszUnderscore = 0;
   float        flMargin      = 0.0;
   int          iReturn       = -1;

   cbMargin = strlen (pszMargin);

   if (0 != strncmp (pszMarginInfo, pszMargin, cbMargin - 1))
   {
      return -1;
   }

   pszMarginInfo += cbMargin;

   pszUnderscore = strchr (pszMarginInfo, '_');

   if (pszUnderscore)
   {
      pszMarginInfo = pszUnderscore + 1;

      if (1 == sscanf (pszMarginInfo, "%f", &flMargin))
      {
         char *pszEnd = pszMarginInfo + strlen (pszMarginInfo);

         if (0 == strcmp (pszEnd - 2, "in"))
         {
            iReturn = (int)(flMargin * 25400.0);
         }
         else if (0 == strcmp (pszEnd - 2, "mm"))
         {
            iReturn = (int)(flMargin * 1000.0);
         }
         else
         {
#ifndef RETAIL
            if (DebugOutput::shouldOutputUPDFDeviceForm ()) DebugOutput::getErrorStream () << "UPDF_DeviceForm::" << __FUNCTION__ << " Unknown margin encoding!" << endl;
#endif
         }
      }
   }

   return iReturn;
}

static DeviceForm *
createFromXMLNode (Device      *pDevice,
                   xmlNodePtr   node,
                   int          id)
{
   UPDF_Device *pUPDFDevice = UPDF_Device::isAUPDF_Device (pDevice);

   if (!pUPDFDevice)
   {
      return 0;
   }

   int                iCapabilities = 0;
   BinaryData        *data          = 0;
   int                iTopMargin    = 0;
   int                iBottomMargin = 0;
   int                iLeftMargin   = 0;
   int                iRightMargin  = 0;
   xmlNodePtr         nodeMargins   = 0;
   DeviceOrientation *pDO           = pUPDFDevice->getCurrentOrientation ();

   switch (pDO->getID ())
   {
   case DeviceOrientation::ORIENTATION_LANDSCAPE:
   case DeviceOrientation::ORIENTATION_REVERSE_LANDSCAPE:
   {
      nodeMargins = FINDUDRENTRY (pUPDFDevice,
                                  node,
                                  "MediaSizeHardwareMarginsLandscape",
                                  UPDFDeviceForm);
      break;
   }
   }

   if (!nodeMargins)
      nodeMargins = FINDUDRENTRY (pUPDFDevice,
                                  node,
                                  "MediaSizeHardwareMargins",
                                  UPDFDeviceForm);

#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDeviceForm ()) DebugOutput::getErrorStream () << "UPDF_DeviceForm::" << __FUNCTION__ << ": nodeMargins = " << nodeMargins << endl;
#endif

   if (nodeMargins)
   {
      char *pszTopMargin    = (char *)xmlGetProp (nodeMargins, (xmlChar *)"Top");
      char *pszBottomMargin = (char *)xmlGetProp (nodeMargins, (xmlChar *)"Bottom");
      char *pszLeftMargin   = (char *)xmlGetProp (nodeMargins, (xmlChar *)"Left");
      char *pszRightMargin  = (char *)xmlGetProp (nodeMargins, (xmlChar *)"Right");

      iTopMargin    = converUPDFMarginToOmniMargin (pszTopMargin);
      iBottomMargin = converUPDFMarginToOmniMargin (pszBottomMargin);
      iLeftMargin   = converUPDFMarginToOmniMargin (pszLeftMargin);
      iRightMargin  = converUPDFMarginToOmniMargin (pszRightMargin);

      switch (pDO->getID ())
      {
      case DeviceOrientation::ORIENTATION_REVERSE_PORTRAIT:
      case DeviceOrientation::ORIENTATION_REVERSE_LANDSCAPE:
      {
         int iSwap;

         iSwap         = iBottomMargin;
         iBottomMargin = iTopMargin;
         iTopMargin    = iSwap;
         iSwap         = iLeftMargin;
         iLeftMargin   = iRightMargin;
         iRightMargin  = iSwap;
         break;
      }
      }

      if (pszTopMargin)
      {
         free (pszTopMargin);
      }
      if (pszBottomMargin)
      {
         free (pszBottomMargin);
      }
      if (pszLeftMargin)
      {
         free (pszLeftMargin);
      }
      if (pszRightMargin)
      {
         free (pszRightMargin);
      }
   }

   return new UPDF_DeviceForm (pUPDFDevice,
                               id,
                               iCapabilities,
                               data,
                               new HardCopyCap (iLeftMargin,
                                                iTopMargin,
                                                iRightMargin,
                                                iBottomMargin),
                               node);
}

DeviceForm * UPDF_DeviceForm::
create (Device *pDevice, int id)
{
   UPDF_Device *pUPDFDevice = UPDF_Device::isAUPDF_Device (pDevice);

#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDeviceForm ()) DebugOutput::getErrorStream () << "UPDF_DeviceForm::" << __FUNCTION__ << ": pDevice = " << pDevice << ", id = " << id << endl;
#endif

   if (!pUPDFDevice)
   {
      return 0;
   }

   xmlNodePtr nodeForms = 0;

   nodeForms = findForms (pUPDFDevice);

   if (nodeForms)
   {
      xmlNodePtr nodeItem = firstNode (nodeForms->xmlChildrenNode);

      while (nodeItem != 0)
      {
         int iFormID = getOmniFormID (nodeItem);

         if (  id == iFormID
            && !(  id == DeviceForm::FORM_UNLISTED
                || id == DeviceForm::FORM_NONE
                )
            )
         {
            return createFromXMLNode (pDevice, nodeItem, id);
         }

         nodeItem = nextNode (nodeItem);
      }

      return 0;
   }
   else
   {
      return 0;
   }
}

char * UPDF_DeviceForm::
getDeviceID ()
{
   return 0;
}

bool UPDF_DeviceForm::
isFormSupported (int id)
{
   UPDF_Device *pUPDFDevice = UPDF_Device::isAUPDF_Device (pDevice_d);

   if (!pUPDFDevice)
   {
      return 0;
   }

   xmlNodePtr nodeForms = 0;

   nodeForms = findForms (pUPDFDevice);

   if (nodeForms)
   {
      xmlNodePtr nodeItem = firstNode (nodeForms->xmlChildrenNode);

      while (nodeItem != 0)
      {
         int iFormID = getOmniFormID (nodeItem);

         if (  id == iFormID
            && !(  id == DeviceForm::FORM_UNLISTED
                || id == DeviceForm::FORM_NONE
                )
            )
         {
            return true;
         }

         nodeItem = nextNode (nodeItem);
      }
   }

   return false;
}

Enumeration * UPDF_DeviceForm::
getEnumeration ()
{
   class UPDF_Form_Enumerator : public Enumeration
   {
   public:
      UPDF_Form_Enumerator (Device *pDevice, xmlNodePtr nodeItem)
      {
         pDevice_d  = pDevice;
         nodeItem_d = skipInvalidForms (nodeItem);
      }

      virtual bool hasMoreElements ()
      {
         return nodeItem_d ? true : false;
      }

      virtual void *nextElement ()
      {
         void *pvRet = 0;

         if (nodeItem_d)
         {
            int id = getOmniFormID (nodeItem_d);

            pvRet = createFromXMLNode (pDevice_d,
                                       nodeItem_d,
                                       id);

            nodeItem_d = nextNode (nodeItem_d);
            nodeItem_d = skipInvalidForms (nodeItem_d);
         }

         return pvRet;
      }

   private:
      Device     *pDevice_d;
      xmlNodePtr  nodeItem_d;
   };

   UPDF_Device *pUPDFDevice = UPDF_Device::isAUPDF_Device (pDevice_d);

   if (!pUPDFDevice)
   {
      return new UPDF_Form_Enumerator (pDevice_d, 0);
   }

   xmlNodePtr nodeForms = 0;

   nodeForms = findForms (pUPDFDevice);

   if (nodeForms)
   {
      xmlNodePtr nodeItem = firstNode (nodeForms->xmlChildrenNode);

      return new UPDF_Form_Enumerator (pDevice_d, nodeItem);
   }
   else
   {
      return new UPDF_Form_Enumerator (pDevice_d, 0);
   }
}

xmlNodePtr UPDF_DeviceForm::
findForms (UPDF_Device *pUPDFDevice)
{
   xmlNodePtr nodeForms = 0;

   if (!pUPDFDevice)
   {
      return nodeForms;
   }

   if (  ((nodeForms = FINDUDRENTRY (pUPDFDevice, nodeForms, "Device_Features",    UPDFDeviceForm)) != 0)
      && ((nodeForms = FINDUDRENTRY (pUPDFDevice, nodeForms, "PrintCapabilities",  UPDFDeviceForm)) != 0)
      && ((nodeForms = FINDUDRENTRY (pUPDFDevice, nodeForms, "Features",           UPDFDeviceForm)) != 0)
      && ((nodeForms = FINDUDRENTRY (pUPDFDevice, nodeForms, "PrintMediaHandling", UPDFDeviceForm)) != 0)
      && ((nodeForms = FINDUDRENTRY (pUPDFDevice, nodeForms, "MediaSizeList",      UPDFDeviceForm)) != 0)
      )
      return nodeForms;

   return nodeForms;
}

string UPDF_DeviceForm::
toString (ostringstream& oss)
{
   ostringstream oss2;

   oss << "{"
       << DeviceForm::toString (oss2)
       << "}";

   return oss.str ();
}

/* Provide a way to print out class data
*/
ostream&
operator<< (ostream& os, const UPDF_DeviceForm& const_self)
{
   UPDF_DeviceForm& self = const_cast<UPDF_DeviceForm&>(const_self);
   ostringstream    oss;

   os << self.toString (oss);

   return os;
}
