/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include "UPDF_DeviceOrientation.hpp"

/*
*/
static int     getOmniOrientationID      (xmlNodePtr      node);

UPDF_DeviceOrientation::
UPDF_DeviceOrientation (Device     *pDevice,
                        int         id,
                        xmlNodePtr  node)
   : DeviceOrientation (pDevice,
                        id)
{
   node_d = node;
}

UPDF_DeviceOrientation::
~UPDF_DeviceOrientation ()
{
}

DeviceOrientation * UPDF_DeviceOrientation::
createV (int id)
{
   return UPDF_DeviceOrientation::create (pDevice_d, id);
}

DeviceOrientation * UPDF_DeviceOrientation::
create (Device *pDevice, char *pszId)
{
   int id = DeviceOrientation::nameToID (pszId);

   if (DeviceOrientation::ORIENTATION_UNLISTED != id)
   {
      return create (pDevice, id);
   }
   else
   {
      UPDF_Device *pUPDFDevice      = UPDF_Device::isAUPDF_Device (pDevice);
      xmlNodePtr   nodeOrientations = 0;

      nodeOrientations = findOrientations (pUPDFDevice);

      if (nodeOrientations)
      {
         xmlNodePtr nodeItem  = firstNode (nodeOrientations->xmlChildrenNode);
         xmlNodePtr nodeFound = 0;

         while (nodeItem != 0)
         {
            char *pszNodeId = (char *)xmlGetProp (nodeItem, (xmlChar *)"ID");

            if (pszNodeId)
            {
               if (0 == strcmp (pszNodeId, pszId))
               {
                  nodeFound = nodeItem;
               }

               free (pszNodeId);
            }

            if (nodeFound)
            {
               id = getOmniOrientationID (nodeFound);

               if (DeviceOrientation::ORIENTATION_UNLISTED != id)
               {
                  return create (pDevice, id);
               }

               nodeFound = 0;
            }

            nodeItem = nextNode (nodeItem);
         }
      }

      return 0;
   }
}

static int
getOmniOrientationID (xmlNodePtr node)
{
   int id = DeviceOrientation::ORIENTATION_UNLISTED;

   if (!node)
   {
      return id;
   }

   char *pszID = (char *)xmlGetProp (node, (xmlChar *)"Predefined_ID");

   if (pszID)
   {
      id = DeviceOrientation::nameToID (pszID);

      free (pszID);
   }

   return id;
}

static xmlNodePtr
skipInvalidOrientations (xmlNodePtr node)
{
   if (!node)
   {
      return 0;
   }

   int id = getOmniOrientationID (node);

   while (  node
           && (  id == DeviceOrientation::ORIENTATION_UNLISTED
              || id == DeviceOrientation::ORIENTATION_NONE
              )
           )
   {
      node = nextNode (node);
      id   = getOmniOrientationID (node);
   }

   return node;
}

DeviceOrientation * UPDF_DeviceOrientation::
create (Device *pDevice, int id)
{
   UPDF_Device *pUPDFDevice = UPDF_Device::isAUPDF_Device (pDevice);

   if (!pUPDFDevice)
   {
      return 0;
   }

   xmlNodePtr nodeOrientations = 0;

   nodeOrientations = findOrientations (pUPDFDevice);

   if (nodeOrientations)
   {
      xmlNodePtr nodeItem = firstNode (nodeOrientations->xmlChildrenNode);

      while (nodeItem != 0)
      {
         int iOrientationID = getOmniOrientationID (nodeItem);

         if (id == iOrientationID)
         {
            return new UPDF_DeviceOrientation (pDevice, id, nodeItem);
         }

         nodeItem = nextNode (nodeItem);
      }

      return 0;
   }
   else
   {
      return 0;
   }
}

char * UPDF_DeviceOrientation::
getDeviceID ()
{
   return 0;
}

bool UPDF_DeviceOrientation::
isOrientationSupported (int id)
{
   UPDF_Device *pUPDFDevice = UPDF_Device::isAUPDF_Device (pDevice_d);

   if (!pUPDFDevice)
   {
      return 0;
   }

   xmlNodePtr nodeOrientations = 0;

   nodeOrientations = findOrientations (pUPDFDevice);

   if (nodeOrientations)
   {
      xmlNodePtr nodeItem = firstNode (nodeOrientations->xmlChildrenNode);

      while (nodeItem != 0)
      {
         int iOrientationID = getOmniOrientationID (nodeItem);

         if (id == iOrientationID)
         {
            return true;
         }

         nodeItem = nextNode (nodeItem);
      }
   }

   return false;
}

Enumeration * UPDF_DeviceOrientation::
getEnumeration ()
{
   class UPDF_Orientation_Enumerator : public Enumeration
   {
   public:
      UPDF_Orientation_Enumerator (Device *pDevice, xmlNodePtr nodeItem)
      {
         pDevice_d  = pDevice;
         nodeItem_d = skipInvalidOrientations (nodeItem);
      }

      virtual bool hasMoreElements ()
      {
         return nodeItem_d ? true : false;
      }

      virtual void *nextElement ()
      {
         void *pvRet = 0;

         if (nodeItem_d)
         {
            pvRet = create (pDevice_d,
                            getOmniOrientationID (nodeItem_d));

            nodeItem_d = nextNode (nodeItem_d);
            nodeItem_d = skipInvalidOrientations (nodeItem_d);
         }

         return pvRet;
      }

   private:
      Device     *pDevice_d;
      xmlNodePtr  nodeItem_d;
   };

   UPDF_Device *pUPDFDevice = UPDF_Device::isAUPDF_Device (pDevice_d);

   if (!pUPDFDevice)
   {
      return new UPDF_Orientation_Enumerator (pDevice_d, 0);
   }

   xmlNodePtr nodeOrientations = 0;

   nodeOrientations = findOrientations (pUPDFDevice);

   if (nodeOrientations)
   {
      xmlNodePtr nodeItem = firstNode (nodeOrientations->xmlChildrenNode);

      return new UPDF_Orientation_Enumerator (pDevice_d, nodeItem);
   }
   else
   {
      return new UPDF_Orientation_Enumerator (pDevice_d, 0);
   }
}

xmlNodePtr UPDF_DeviceOrientation::
findOrientations (UPDF_Device *pUPDFDevice)
{
   xmlNodePtr nodeOrientations = 0;

   if (!pUPDFDevice)
   {
      return nodeOrientations;
   }

   if (  ((nodeOrientations = FINDUDRENTRY (pUPDFDevice, nodeOrientations, "Device_Features",          UPDFDeviceOrientation)) != 0)
      && ((nodeOrientations = FINDUDRENTRY (pUPDFDevice, nodeOrientations, "PrintCapabilities",        UPDFDeviceOrientation)) != 0)
      && ((nodeOrientations = FINDUDRENTRY (pUPDFDevice, nodeOrientations, "Features",                 UPDFDeviceOrientation)) != 0)
      && ((nodeOrientations = FINDUDRENTRY (pUPDFDevice, nodeOrientations, "PrintMediaHandling",       UPDFDeviceOrientation)) != 0)
      && ((nodeOrientations = FINDUDRENTRY (pUPDFDevice, nodeOrientations, "MediaPageOrientationList", UPDFDeviceOrientation)) != 0)
      )
      return nodeOrientations;

   return nodeOrientations;
}

string UPDF_DeviceOrientation::
toString (ostringstream& oss)
{
   ostringstream oss2;

   oss << "{";

   // @TBD
   printXMLNode (node_d, DebugOutput::shouldOutputUPDFDeviceOrientation ());

   oss << DeviceOrientation::toString (oss2)
       << "}";

   return oss.str ();
}

/* Provide a way to print out class data
*/
ostream&
operator<< (ostream& os, const UPDF_DeviceOrientation& const_self)
{
   UPDF_DeviceOrientation& self = const_cast<UPDF_DeviceOrientation&>(const_self);
   ostringstream           oss;

   os << self.toString (oss);

   return os;
}
