/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <UPDF_DevicePrintMode.hpp>

UPDF_DevicePrintMode::
UPDF_DevicePrintMode (Device      *pDevice,
                      int          id,
                      int          iPhysicalCount,
                      int          iLogicalCount,
                      int          iPlanes,
                      xmlNodePtr   node)
   : DevicePrintMode (pDevice,
                      id,
                      iPhysicalCount,
                      iLogicalCount,
                      iPlanes)
{
   node_d = node;
}

UPDF_DevicePrintMode::
~UPDF_DevicePrintMode ()
{
}

DevicePrintMode * UPDF_DevicePrintMode::
createV (int id)
{
   return UPDF_DevicePrintMode::create (pDevice_d, id);
}

DevicePrintMode * UPDF_DevicePrintMode::
create (Device *pDevice, char *pszId)
{
   return create (pDevice, DevicePrintMode::nameToID (pszId));
}

static int
getOmniPrintModeID (xmlNodePtr node)
{
   int id = DevicePrintMode::PRINT_MODE_NONE;

   if (!node)
   {
      return id;
   }

   char *pszID = (char *)xmlGetProp (node, (xmlChar *)"Predefined_ID");

   if (pszID)
   {

      free (pszID);
   }

   return id;
}

static DevicePrintMode *
createFromXMLNode (Device      *pDevice,
                   xmlNodePtr   node,
                   int          id)
{
   UPDF_Device *pUPDFDevice = UPDF_Device::isAUPDF_Device (pDevice);

   if (!pUPDFDevice)
   {
      return 0;
   }

   int iPhysicalCount = 1;
   int iLogicalCount  = 1;
   int iPlanes        = 1;

   return new UPDF_DevicePrintMode (pUPDFDevice,
                                    DevicePrintMode::PRINT_MODE_1_ANY,
                                    iPhysicalCount,
                                    iLogicalCount,
                                    iPlanes,
                                    node);
}

DevicePrintMode * UPDF_DevicePrintMode::
create (Device *pDevice, int id)
{
   UPDF_Device *pUPDFDevice = UPDF_Device::isAUPDF_Device (pDevice);

#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDevicePrintMode ()) DebugOutput::getErrorStream () << "UPDF_DevicePrintMode::" << __FUNCTION__ << ": pDevice = " << pDevice << ", id = " << id << endl;
#endif

   if (!pUPDFDevice)
   {
      return 0;
   }

   return createFromXMLNode (pDevice, 0, id);
}

char * UPDF_DevicePrintMode::
getDeviceID ()
{
   return 0;
}

bool UPDF_DevicePrintMode::
isPrintModeSupported (int id)
{
   UPDF_Device *pUPDFDevice = UPDF_Device::isAUPDF_Device (pDevice_d);

   if (!pUPDFDevice)
   {
      return 0;
   }

   xmlNodePtr nodePrintModes = 0;

   nodePrintModes = findPrintModes (pUPDFDevice);

   if (nodePrintModes)
   {
      xmlNodePtr nodeItem = firstNode (nodePrintModes->xmlChildrenNode);

      while (nodeItem != 0)
      {
         int iPrintModeID = getOmniPrintModeID (nodeItem);

         if (  id == iPrintModeID
            && !(  id == DevicePrintMode::PRINT_MODE_UNLISTED
                || id == DevicePrintMode::PRINT_MODE_NONE
                )
            )
         {
            return true;
         }

         nodeItem = nextNode (nodeItem);
      }
   }

   return false;
}

Enumeration * UPDF_DevicePrintMode::
getEnumeration ()
{
   class UPDF_PrintMode_Enumerator : public Enumeration
   {
   public:
      UPDF_PrintMode_Enumerator (Device *pDevice, xmlNodePtr nodeItem)
      {
         pDevice_d             = pDevice;
         fHasColor_d           = false; // @TBD
         iPrintModesReturned_d = 0;
      }

      virtual bool hasMoreElements ()
      {
         if (fHasColor_d)
         {
            return iPrintModesReturned_d < 2 ? true : false;
         }
         else
         {
            return iPrintModesReturned_d < 1 ? true : false;
         }
      }

      virtual void *nextElement ()
      {
         void *pvRet = 0;

         if (hasMoreElements ())
         {
            iPrintModesReturned_d++;

            pvRet = createFromXMLNode (pDevice_d,
                                       0,
                                       0);
         }

         return pvRet;
      }

   private:
      Device     *pDevice_d;
      bool        fHasColor_d;
      int         iPrintModesReturned_d;
   };

   UPDF_Device *pUPDFDevice = UPDF_Device::isAUPDF_Device (pDevice_d);

   if (!pUPDFDevice)
   {
      return new UPDF_PrintMode_Enumerator (pDevice_d, 0);
   }

   xmlNodePtr nodePrintModes = 0;

   nodePrintModes = findPrintModes (pUPDFDevice);

   if (nodePrintModes)
   {
      xmlNodePtr nodeItem = firstNode (nodePrintModes->xmlChildrenNode);

      return new UPDF_PrintMode_Enumerator (pDevice_d, nodeItem);
   }
   else
   {
      return new UPDF_PrintMode_Enumerator (pDevice_d, 0);
   }
}

xmlNodePtr UPDF_DevicePrintMode::
findPrintModes (UPDF_Device *pUPDFDevice)
{
   xmlNodePtr nodePrintModes = 0;

   if (!pUPDFDevice)
   {
      return nodePrintModes;
   }

   if (  ((nodePrintModes = FINDUDRENTRY (pUPDFDevice, nodePrintModes, "Device_Features",   UPDFDevicePrintMode)) != 0)
      && ((nodePrintModes = FINDUDRENTRY (pUPDFDevice, nodePrintModes, "PrintCapabilities", UPDFDevicePrintMode)) != 0)
      && ((nodePrintModes = FINDUDRENTRY (pUPDFDevice, nodePrintModes, "Features",          UPDFDevicePrintMode)) != 0)
      && ((nodePrintModes = FINDUDRENTRY (pUPDFDevice, nodePrintModes, "PrintModeHandling", UPDFDevicePrintMode)) != 0)
      && ((nodePrintModes = FINDUDRENTRY (pUPDFDevice, nodePrintModes, "PrintModeTypeList", UPDFDevicePrintMode)) != 0)
      )
      return nodePrintModes;

   return nodePrintModes;
}

string UPDF_DevicePrintMode::
toString (ostringstream& oss)
{
   ostringstream oss2;

   oss << "{"
       << DevicePrintMode::toString (oss2)
       << "}";

   return oss.str ();
}

/* Provide a way to print out class data
*/
ostream&
operator<< (ostream& os, const UPDF_DevicePrintMode& const_self)
{
   UPDF_DevicePrintMode& self = const_cast<UPDF_DevicePrintMode&>(const_self);
   ostringstream         oss;

   os << self.toString (oss);

   return os;
}
