/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <UPDF_DeviceResolution.hpp>

/*
*/
static int     getOmniResolutionID       (xmlNodePtr   node,
                                          UPDF_Device *pUPDFDevice,
                                          int         *piXRes,
                                          int         *piYRes);

UPDF_DeviceResolution::
UPDF_DeviceResolution (Device      *pDevice,
                       int          id,
                       int          iXRes,
                       int          iYRes,
                       int          iXInternalRes,
                       int          iYInternalRes,
                       BinaryData  *pbdData,
                       int          iCapabilities,
                       int          iDestinationBitsPerPel,
                       int          iScanlineMultiple,
                       xmlNodePtr   node)
   : DeviceResolution (pDevice,
                       id,
                       iXRes,
                       iYRes,
                       iXInternalRes,
                       iYInternalRes,
                       pbdData,
                       iCapabilities,
                       iDestinationBitsPerPel,
                       iScanlineMultiple)
{
   node_d = node;
}

UPDF_DeviceResolution::
~UPDF_DeviceResolution ()
{
}

DeviceResolution * UPDF_DeviceResolution::
createV (int id)
{
   return UPDF_DeviceResolution::create (pDevice_d, id);
}

DeviceResolution * UPDF_DeviceResolution::
create (Device *pDevice, char *pszId)
{
   int id = DeviceResolution::nameToID (pszId);

   if (DeviceResolution::RESOLUTION_UNLISTED != id)
   {
      return create (pDevice, id);
   }
   else
   {
      UPDF_Device *pUPDFDevice     = UPDF_Device::isAUPDF_Device (pDevice);
      xmlNodePtr   nodeResolutions = 0;
      int          iXRes           = 0;
      int          iYRes           = 0;

      nodeResolutions = findResolutions (pUPDFDevice);

      if (nodeResolutions)
      {
         xmlNodePtr nodeItem  = firstNode (nodeResolutions->xmlChildrenNode);
         xmlNodePtr nodeFound = 0;

         while (nodeItem != 0)
         {
            char *pszNodeId = (char *)xmlGetProp (nodeItem, (xmlChar *)"ID");

            if (pszNodeId)
            {
               if (0 == strcmp (pszNodeId, pszId))
               {
                  nodeFound = nodeItem;
               }

               free (pszNodeId);
            }

            if (nodeFound)
            {
               id = getOmniResolutionID (nodeFound,
                                         pUPDFDevice,
                                         &iXRes,
                                         &iYRes);

               if (DeviceResolution::RESOLUTION_UNLISTED != id)
               {
                  return create (pDevice, id);
               }

               nodeFound = 0;
            }

            nodeItem = nextNode (nodeItem);
         }
      }

      return 0;
   }
}

static int
getOmniResolutionID (xmlNodePtr   node,
                     UPDF_Device *pUPDFDevice,
                     int         *piXRes,
                     int         *piYRes)
{
   int id            = DeviceResolution::RESOLUTION_UNLISTED;
   int iVirtualUnits = pUPDFDevice->getVirtualUnits ();

   if (iVirtualUnits == 0)
   {
#ifndef RETAIL
      if (DebugOutput::shouldOutputUPDFDeviceResolution ()) DebugOutput::getErrorStream () << "UPDF_DeviceResolution::" << __FUNCTION__ << ": Error: Could not find virtual units @ " << __LINE__ << endl;
#endif

      return id;
   }

   if (!node)
   {
      return id;
   }

   char *pszHorzID = (char *)xmlGetProp (node, (xmlChar *)"HorizontalResolution");
   char *pszVertID = (char *)xmlGetProp (node, (xmlChar *)"VerticalResolution");

   if (  pszHorzID
      && pszVertID
      )
   {
      int iXRes = atoi (pszHorzID);
      int iYRes = atoi (pszVertID);

#ifndef RETAIL
      if (DebugOutput::shouldOutputUPDFDeviceResolution ()) DebugOutput::getErrorStream () << "UPDF_DeviceResolution::" << __FUNCTION__ << ": iXRes = " << iXRes << endl;
      if (DebugOutput::shouldOutputUPDFDeviceResolution ()) DebugOutput::getErrorStream () << "UPDF_DeviceResolution::" << __FUNCTION__ << ": iYRes = " << iYRes << endl;
#endif

      if (  iXRes
         && iYRes
         )
      {
         static char achName[26]; // @TBD

         iXRes = iVirtualUnits / iXRes;
         iYRes = iVirtualUnits / iYRes;

         sprintf (achName, "%dx%d", iXRes, iYRes);

#ifndef RETAIL
         if (DebugOutput::shouldOutputUPDFDeviceResolution ()) DebugOutput::getErrorStream () << "UPDF_DeviceResolution::" << __FUNCTION__ << ": achName = " << achName << endl;
#endif

         id = DeviceResolution::nameToID (achName);

         *piXRes = iXRes;
         *piYRes = iYRes;
      }
   }

   if (pszHorzID)
   {
      free (pszHorzID);
   }
   if (pszVertID)
   {
      free (pszVertID);
   }

   return id;
}

static xmlNodePtr
skipInvalidResolutions (xmlNodePtr   node,
                        UPDF_Device *pUPDFDevice,
                        int         *piXRes,
                        int         *piYRes)
{
   if (!node)
   {
      return 0;
   }

   int id = getOmniResolutionID (node, pUPDFDevice, piXRes, piYRes);

   while (  node
           && (  id == DeviceResolution::RESOLUTION_UNLISTED
              || id == DeviceResolution::RESOLUTION_NONE
              )
           )
   {
      node = nextNode (node);
      id   = getOmniResolutionID (node, pUPDFDevice, piXRes, piYRes);
   }

   return node;
}

static DeviceResolution *
createFromXMLNode (Device      *pDevice,
                   xmlNodePtr   node,
                   int          id,
                   int          iXRes,
                   int          iYRes)
{
   UPDF_Device *pUPDFDevice = UPDF_Device::isAUPDF_Device (pDevice);

   if (!pUPDFDevice)
   {
      return 0;
   }

   int          iXInternalRes          = 0;
   int          iYInternalRes          = 0;
   BinaryData  *pbdData                = 0;
   int          iCapabilities          = 0;
   int          iDestinationBitsPerPel = 0;
   int          iScanlineMultiple      = 1;

   return new UPDF_DeviceResolution (pUPDFDevice,
                                     id,
                                     iXRes,
                                     iYRes,
                                     iXInternalRes,
                                     iYInternalRes,
                                     pbdData,
                                     iCapabilities,
                                     iDestinationBitsPerPel,
                                     iScanlineMultiple,
                                     node);
}

DeviceResolution * UPDF_DeviceResolution::
create (Device *pDevice, int id)
{
   UPDF_Device *pUPDFDevice = UPDF_Device::isAUPDF_Device (pDevice);

#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDeviceResolution ()) DebugOutput::getErrorStream () << "UPDF_DeviceResolution::" << __FUNCTION__ << ": pDevice = " << pDevice << ", id = " << id << endl;
#endif

   if (!pUPDFDevice)
   {
      return 0;
   }

   xmlNodePtr nodeResolutions = findResolutions (pUPDFDevice);

   if (nodeResolutions)
   {
      xmlNodePtr nodeItem = firstNode (nodeResolutions->xmlChildrenNode);

      while (nodeItem != 0)
      {
         int iXRes         = 0;
         int iYRes         = 0;
         int iResolutionID = getOmniResolutionID (nodeItem,
                                                  pUPDFDevice,
                                                  &iXRes,
                                                  &iYRes);

         if (  id == iResolutionID
            && !(  id == DeviceResolution::RESOLUTION_UNLISTED
                || id == DeviceResolution::RESOLUTION_NONE
                )
            )
         {
            return createFromXMLNode (pDevice,
                                      nodeItem,
                                      id,
                                      iXRes,
                                      iYRes);
         }

         nodeItem = nextNode (nodeItem);
      }

      return 0;
   }
   else
   {
      return 0;
   }
}

char * UPDF_DeviceResolution::
getDeviceID ()
{
   return 0;
}

bool UPDF_DeviceResolution::
isResolutionSupported (int id)
{
   UPDF_Device *pUPDFDevice = UPDF_Device::isAUPDF_Device (pDevice_d);

   if (!pUPDFDevice)
   {
      return 0;
   }

   xmlNodePtr nodeResolutions = 0;

   nodeResolutions = findResolutions (pUPDFDevice);

   if (nodeResolutions)
   {
      xmlNodePtr nodeItem = firstNode (nodeResolutions->xmlChildrenNode);

      while (nodeItem != 0)
      {
         int iXRes         = 0;
         int iYRes         = 0;
         int iResolutionID = getOmniResolutionID (nodeItem,
                                                  pUPDFDevice,
                                                  &iXRes,
                                                  &iYRes);

         if (  id == iResolutionID
            && !(  id == DeviceResolution::RESOLUTION_UNLISTED
                || id == DeviceResolution::RESOLUTION_NONE
                )
            )
         {
            return true;
         }

         nodeItem = nextNode (nodeItem);
      }
   }

   return false;
}

Enumeration * UPDF_DeviceResolution::
getEnumeration ()
{
   class UPDF_Resolution_Enumerator : public Enumeration
   {
   public:
      UPDF_Resolution_Enumerator (Device *pDevice, xmlNodePtr nodeItem)
      {
         int iXRes = 0;
         int iYRes = 0;

         pDevice_d     = pDevice;
         pUPDFDevice_d = UPDF_Device::isAUPDF_Device (pDevice);
         nodeItem_d    = skipInvalidResolutions (nodeItem,
                                                 pUPDFDevice_d,
                                                 &iXRes,
                                                 &iYRes);
      }

      virtual bool hasMoreElements ()
      {
         return nodeItem_d ? true : false;
      }

      virtual void *nextElement ()
      {
         void *pvRet = 0;

         if (nodeItem_d)
         {
            int iXRes         = 0;
            int iYRes         = 0;
            int iResolutionID = getOmniResolutionID (nodeItem_d,
                                                     pUPDFDevice_d,
                                                     &iXRes,
                                                     &iYRes);

            pvRet = createFromXMLNode (pDevice_d,
                                       nodeItem_d,
                                       iResolutionID,
                                       iXRes,
                                       iYRes);

            nodeItem_d = nextNode (nodeItem_d);
            nodeItem_d = skipInvalidResolutions (nodeItem_d,
                                                 pUPDFDevice_d,
                                                 &iXRes,
                                                 &iYRes);
         }

         return pvRet;
      }

   private:
      Device      *pDevice_d;
      UPDF_Device *pUPDFDevice_d;
      xmlNodePtr   nodeItem_d;
   };

   UPDF_Device *pUPDFDevice = UPDF_Device::isAUPDF_Device (pDevice_d);

   if (!pUPDFDevice)
   {
      return new UPDF_Resolution_Enumerator (pDevice_d, 0);
   }

   xmlNodePtr nodeResolutions = 0;

   nodeResolutions = findResolutions (pUPDFDevice);

   if (nodeResolutions)
   {
      xmlNodePtr nodeItem = firstNode (nodeResolutions->xmlChildrenNode);

      return new UPDF_Resolution_Enumerator (pDevice_d, nodeItem);
   }
   else
   {
      return new UPDF_Resolution_Enumerator (pDevice_d, 0);
   }
}

xmlNodePtr UPDF_DeviceResolution::
findResolutions (UPDF_Device *pUPDFDevice)
{
   xmlNodePtr nodeResolutions = 0;

   if (!pUPDFDevice)
   {
      return nodeResolutions;
   }

   if (  ((nodeResolutions = FINDUDRENTRY (pUPDFDevice, nodeResolutions, "Device_Features",      UPDFDeviceResolution)) != 0)
      && ((nodeResolutions = FINDUDRENTRY (pUPDFDevice, nodeResolutions, "PrintCapabilities",    UPDFDeviceResolution)) != 0)
      && ((nodeResolutions = FINDUDRENTRY (pUPDFDevice, nodeResolutions, "Features",             UPDFDeviceResolution)) != 0)
      && ((nodeResolutions = FINDUDRENTRY (pUPDFDevice, nodeResolutions, "DeviceResolutionList", UPDFDeviceResolution)) != 0)
      )
      return nodeResolutions;

   return nodeResolutions;
}

string UPDF_DeviceResolution::
toString (ostringstream& oss)
{
   ostringstream oss2;

   oss << "{"
       << DeviceResolution::toString (oss2)
       << "}";

   return oss.str ();
}

/* Provide a way to print out class data
*/
ostream&
operator<< (ostream& os, const UPDF_DeviceResolution& const_self)
{
   UPDF_DeviceResolution& self = const_cast<UPDF_DeviceResolution&>(const_self);
   ostringstream          oss;

   os << self.toString (oss);

   return os;
}
