/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include "UPDF_Instance.hpp"
#include "defines.hpp"

#include <stdio.h>
#include <string.h>

#include <iostream>
#include <strstream>

const static bool fTestNoCompression = true;

UPDF_Instance::
UPDF_Instance (PrintDevice *pDevice)
   : DeviceInstance (pDevice)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) cerr << "UPDF_Instance::" << __FUNCTION__ << " ()" << endl;
#endif

   fHaveInitialized_d  = false;
   fHaveSetupPrinter_d = false;
   ptlPrintHead_d.x    = 0;
   ptlPrintHead_d.y    = 0;

   UPDF_Device *pUPDFDevice = UPDF_Device::isAUPDF_Device (pDevice_d);

   if (pUPDFDevice)
   {
      loadNonDominantDefaults (pUPDFDevice, pUPDFDevice->getRootUDR ());
   }
   loadLocaleDefaults ();
}

UPDF_Instance::
~UPDF_Instance ()
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) cerr << "UPDF_Instance::~" << __FUNCTION__ << " ()" << endl;
#endif

#ifdef HAVE_SYMBOL_TABLE
   for ( SYMBOLTABLE::iterator next = symbolTable_d.begin () ;
         next != symbolTable_d.end () ;
         next++ )
   {
#ifndef RETAIL
      if (DebugOutput::shouldOutputUPDFDeviceInstance ()) cerr << "UPDF_Instance::~" << __FUNCTION__ << ": deleting symbol table entry" << endl;
#endif
      delete (*next).second;
   }
#endif
}

void UPDF_Instance::
initializeInstance ()
{
   if (fHaveInitialized_d)
      return;

   fHaveInitialized_d = true;

#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) cerr << "UPDF_Instance::initializeInstance ()" << endl;
#endif
}

void UPDF_Instance::
loadNonDominantDefaults (UPDF_Device *pUPDFDevice, xmlNodePtr node)
{
   if (!node)
      return;

   node = firstNode (node);

   while (node)
   {
      if (node->xmlChildrenNode)
      {
         loadNonDominantDefaults (pUPDFDevice, node->xmlChildrenNode);
      }

      char *pszNDR = (char *)xmlGetProp (node, (xmlChar *)"Non_Dominant_Representative");

      if (pszNDR)
      {
         char *pszID = (char *)xmlGetProp (node, (xmlChar *)"ID");

#ifndef RETAIL
         if (DebugOutput::shouldOutputUPDFDeviceInstance ()) cerr << "UPDF_Instance::" << __FUNCTION__ << ": pszNDR = " << pszNDR << ", pszID = " << pszID << endl;
#endif

         if (pszID)
         {
            xmlNodePtr elm = pUPDFDevice->findEntryKeyValue (node,
                                                             "ID",
                                                             pszNDR,
                                                             DebugOutput::shouldOutputUPDFDeviceInstance ());

            if (elm)
            {
               string stringID = pszID;

               xmlObjectMap_d[stringID] = elm;
            }

            free (pszID);
         }

         free (pszNDR);
      }

      node = nextNode (node);
   }
}

void UPDF_Instance::
loadLocaleDefaults ()
{
   UPDF_Device *pUPDFDevice       = UPDF_Device::isAUPDF_Device (pDevice_d);
   xmlNodePtr   nodeLocalDefaults = 0;

   if (!pUPDFDevice)
      return;

   if ((nodeLocalDefaults = FINDLOCALEENTRY (pUPDFDevice, nodeLocalDefaults, "LocaleDefaults", UPDFDeviceInstance)) != 0)
   {
      xmlNodePtr nodeItem = firstNode (nodeLocalDefaults->xmlChildrenNode);

      while (nodeItem != 0)
      {
         char *pszFeature = (char *)xmlGetProp (nodeItem, (xmlChar *)"Default_Feature");
         char *pszSetting = (char *)xmlGetProp (nodeItem, (xmlChar *)"Default_Setting");

         if (  pszFeature
            && pszSetting
            )
         {
            string     stringFeature = pszFeature;
            xmlNodePtr elm           = 0;

#ifndef RETAIL
            if (DebugOutput::shouldOutputUPDFDeviceInstance ()) cerr << "UPDF_Instance::" << __FUNCTION__ << ": pszFeature = " << pszFeature << ", pszSetting = " << pszSetting << endl;
#endif

            elm = FINDUDRENTRYKEYVALUE (pUPDFDevice, "ID", pszSetting, UPDFDeviceInstance);

            if (elm)
            {
               xmlObjectMap_d[stringFeature] = elm;
            }
         }

         if (pszFeature)
         {
            free (pszFeature);
         }
         if (pszSetting)
         {
            free (pszSetting);
         }

         nodeItem = nextNode (nodeItem);
      }
   }

}

#ifdef HAVE_SYMBOL_TABLE
static char *apszDeviceJobPropertyKeys[] = {
   "RandomAccessMemory.RAM_KB",
   "DeviceResolution.HorizontalResolution",
   "DeviceResolution.VerticalResolution",
   "RasterGraphic.HorizontalResolution",
   "RasterGraphic.VerticalResolution",
   "Positioning.HorizontalResolution",
   "Positioning.VerticalResolution",
   "MediaCopy.Media_Copy_ID",
   "MediaCollation.MediaCollation_Global_ID",
   "MediaCollation.Media_Proprietary_ID",
   "MediaPageOrientation.MediaPageOrientation_Global_ID",
   "MediaSize.MediaSize_Global_ID",
   "MediaSize.Media_Proprietary_ID",
   "MediaType.MediaType_Global_ID",
   "MediaType.Media_Proprietary_ID",
   "MediaSource.MediaSource_Global_ID",
   "MediaSource.Media_Proprietary_ID",
   "MediaDestination.MediaDestination_Global_ID",
   "MediaDestination.Media_Proprietary_ID",
   "MediaCoating.MediaCoating_Global_ID",
   "MediaCoating.Media_Proprietary_ID",
   "MediaColor.MediaColor_Global_ID",
   "MediaColor.Media_Proprietary_ID",
   "MediaWeight.MediaWeight_Global_ID",
   "MediaWeight.Media_Proprietary_ID",
   "MediaStaplingSet.Media_Proprietary_ID",
   "MediaPunchingSet.Media_Proprietary_ID",
   "MediaJogging.MediaJogging_Global_ID",
   "MediaJogging.Media_Proprietary_ID",
   "MediaDuplex.MediaDuplex_Global_ID",
   "MediaDuplex.Media_Proprietary_ID",
   "MediaBooklet.MediaBooklet_Global_ID",
   "MediaBooklet.Media_Proprietary_ID",
   "MediaPageOrder.MediaPageOrder_Global_ID",
   "MediaPageOrder.Media_Proprietary_ID",
   "MediaNUp.MediaNUp_HorizontalPages",
   "MediaNUp.MediaNUp_VerticalPages",
   "MediaNUpBorder.MediaNUpBorder_Global_ID",
   "MediaNUpBorder.Media_Proprietary_ID",
   "MediaNUpPageOrder.MediaNUpPageOrder_Global_ID",
   "MediaNUpPageOrder.Media_Proprietary_ID",
   "MediaTargetSize.MediaTargetSize_Global_ID",
   "MediaTargetSize.Media_Proprietary_ID",
   "MediaScaling.MediaScaling_Global_ID",
   "MediaScaling.Media_Proprietary_ID",
   "MediaPercentage.Media_Percentage_ID"
};

bool UPDF_Instance::
handleDeviceJobProperty (char *pszKey,
                         char *pszValue)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) cerr << "UPDF_Instance::" << __FUNCTION__ << " ("
                          << pszKey << ", "
                          << pszValue << ")"
                          << endl;
#endif

   for (int i = 0; i < (int)dimof (apszDeviceJobPropertyKeys); i++)
   {
      if (0 == strcmp (pszKey, apszDeviceJobPropertyKeys[i]))
      {
         symbolTable_d[string (pszKey)] = new Expression (pszValue, true);

         return true;
      }
   }

   return false;
}

Enumeration * UPDF_Instance::
listDeviceJobPropertyKeys ()
{
   class DeviceJobPropertyKeyEnumerator : public Enumeration
   {
   public:
      DeviceJobPropertyKeyEnumerator (char **apszKeys, int iDimOfKeys)
      {
         iWhichKey_d = iDimOfKeys;
         apszKeys_d  = apszKeys;
      }

      virtual bool hasMoreElements ()
      {
         return iWhichKey_d > 0;
      }

      virtual void * nextElement ()
      {
         if (0 < iWhichKey_d)
         {
            iWhichKey_d--;
            return apszKeys_d[iWhichKey_d];
         }

         return 0;
      }

   private:
      int    iWhichKey_d;
      char **apszKeys_d;
   };

   return new DeviceJobPropertyKeyEnumerator (apszDeviceJobPropertyKeys, dimof (apszDeviceJobPropertyKeys));
}

string * UPDF_Instance::
getJobPropertyType (char *pszKey)
{
   // @TBD
   return new string ("string bob");
}

string * UPDF_Instance::
getJobProperty (char *pszKey)
{
   Expression *pExpressionValue = 0;
   string     *pStringReturn    = 0;

   pExpressionValue = symbolTable_d[string (pszKey)];

   if (pExpressionValue)
   {
      switch (pExpressionValue->getType ())
      {
      case TYPE_STRING:
      {
         pStringReturn = new string (pExpressionValue->getString ());
         break;
      }

      case TYPE_HEX:
      {
         // @TBD
/////////pExpressionValue->getBinaryData ()
         break;
      }
      }
   }

   return pStringReturn;
}

#endif

bool UPDF_Instance::
handleDeviceJobProperty (char *pszKey,
                         char *pszValue)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) cerr << "UPDF_Instance::" << __FUNCTION__ << " ("
                          << pszKey << ", "
                          << pszValue << ")"
                          << endl;
#endif

   if (  !pszKey
      || !*pszKey
      || !pszValue
      || !*pszValue
      )
      return false;

   UPDF_Device *pUPDFDevice = UPDF_Device::isAUPDF_Device (pDevice_d);

   if (!pUPDFDevice)
      return false;

   xmlNodePtr  node      = 0;
   string      stringKey = pszKey;

   node = xmlObjectMap_d[stringKey];

   if (node)
   {
      xmlNodePtr root = pUPDFDevice->findEntryKeyValue (node,
                                                        "ID",
                                                        pszKey,
                                                        DebugOutput::shouldOutputUPDFDeviceInstance ());

      if (root)
      {
         xmlNodePtr newNode = pUPDFDevice->findEntryKeyValue (root,
                                                              "ID",
                                                              pszValue,
                                                              DebugOutput::shouldOutputUPDFDeviceInstance ());

         if (newNode)
         {
            xmlObjectMap_d[stringKey] = newNode;

            return true;
         }
      }
   }

///for (XMLObjectMap::iterator next = xmlObjectMap_d.begin ();
///     next != xmlObjectMap_d.end ();
///     next++)
///{
///}

   return false;
}

Enumeration * UPDF_Instance::
listDeviceJobPropertyKeys ()
{
   class DeviceJobPropertyKeyEnumerator : public Enumeration
   {
   public:
      DeviceJobPropertyKeyEnumerator (XMLObjectMap::iterator first,
                                      XMLObjectMap::iterator last)
      {
         next_d = first;
         last_d = last;
      }

      virtual ~
      DeviceJobPropertyKeyEnumerator ()
      {
      }

      virtual bool
      hasMoreElements ()
      {
         return next_d != last_d;
      }

      virtual void *
      nextElement ()
      {
         void *pvRet = 0;

         if (next_d != last_d)
         {
            pvRet = (void *)next_d->first.c_str ();

            next_d++;
         }

         return pvRet;
      }

   private:
      XMLObjectMap::iterator next_d;
      XMLObjectMap::iterator last_d;
   };

#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) cerr << "UPDF_Instance::" << __FUNCTION__ << endl;
#endif

   return new DeviceJobPropertyKeyEnumerator (xmlObjectMap_d.begin (),
                                              xmlObjectMap_d.end ());
}

Enumeration * UPDF_Instance::
listKeyValues (char *pszKey)
{
   class DeviceJobPropertyKeyValueEnumerator : public Enumeration
   {
   public:
      DeviceJobPropertyKeyValueEnumerator (xmlNodePtr elm)
      {
         elm_d     = 0;
         pszNext_d = 0;
         pszLast_d = 0;

         if (  elm
            && elm->xmlChildrenNode
            )
         {
            elm_d = firstNode (elm->xmlChildrenNode);

            pszNext_d = getNextID ();
         }
      }

      virtual ~
      DeviceJobPropertyKeyValueEnumerator ()
      {
         if (pszNext_d)
         {
            free (pszNext_d);
            pszNext_d = 0;
         }

         if (pszLast_d)
         {
            free (pszLast_d);
            pszLast_d = 0;
         }
      }

      virtual bool
      hasMoreElements ()
      {
         return pszNext_d != 0;
      }

      virtual void *
      nextElement ()
      {
         if (pszLast_d)
         {
            free (pszLast_d);
            pszLast_d = 0;
         }

         pszLast_d = pszNext_d;
         pszNext_d = getNextID ();

         return (void *)pszLast_d;
      }

   private:
      char *
      getNextID ()
      {
         while (elm_d != 0)
         {
            char *pszID = (char *)xmlGetProp (elm_d, (xmlChar *)"ID");

            elm_d = nextNode (elm_d);

            if (pszID)
            {
               return pszID;
            }
         }

         return 0;
      }

      xmlNodePtr  elm_d;
      char       *pszNext_d;
      char       *pszLast_d;
   };

#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) cerr << "UPDF_Instance::" << __FUNCTION__ << " ("
                          << pszKey << ")"
                          << endl;
#endif

   Enumeration *pRet        = 0;
   UPDF_Device *pUPDFDevice = UPDF_Device::isAUPDF_Device (pDevice_d);

   if (  pszKey
      && *pszKey
      && pUPDFDevice
      )
   {
      string stringKey = pszKey;

      if (xmlObjectMap_d[stringKey])
      {
         xmlNodePtr root = pUPDFDevice->findEntryKeyValue (pUPDFDevice->getRootUDR (),
                                                           "ID",
                                                           pszKey,
                                                           DebugOutput::shouldOutputUPDFDeviceInstance ());

         pRet = new DeviceJobPropertyKeyValueEnumerator (root);
      }
   }

   if (!pRet)
   {
      pRet = new DeviceJobPropertyKeyValueEnumerator (0);
   }

   return pRet;
}

string * UPDF_Instance::
getJobPropertyType (char *pszKey)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) cerr << "UPDF_Instance::" << __FUNCTION__ << " ("
                          << pszKey << ")"
                          << endl;
#endif

   xmlNodePtr  node       = 0;
   string      stringKey  = pszKey;
   string     *pstringRet = 0;

   if (  pszKey
      && *pszKey
      )
   {
      node = xmlObjectMap_d[stringKey];

      if (node)
      {
         char *pszID = (char *)xmlGetProp (node, (xmlChar *)"ID");

         if (pszID)
         {
            pstringRet = new string ("string");

            *pstringRet += pszID;

            free (pszID);
         }
      }
   }

#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) cerr << "UPDF_Instance::" << __FUNCTION__ << " returns "
                          << (pstringRet ? *pstringRet : "NULL")
                          << endl;
#endif

   return pstringRet;
}

string * UPDF_Instance::
getJobProperty (char *pszKey)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) cerr << "UPDF_Instance::" << __FUNCTION__ << " ("
                          << pszKey << ")"
                          << endl;
#endif

   xmlNodePtr  node       = 0;
   string      stringKey  = pszKey;
   string     *pstringRet = 0;

   if (  pszKey
      && *pszKey
      )
   {
      node = xmlObjectMap_d[stringKey];

      if (node)
      {
         char *pszID = (char *)xmlGetProp (node, (xmlChar *)"ID");

         if (pszID)
         {
            pstringRet = new string (pszID);

            free (pszID);
         }
      }
   }

#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) cerr << "UPDF_Instance::" << __FUNCTION__ << " returns "
                          << (pstringRet ? *pstringRet : "NULL")
                          << endl;
#endif

   return pstringRet;
}

string * UPDF_Instance::
translateKeyValue (char *pszKey,
                   char *pszValue)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) cerr << "UPDF_Instance::" << __FUNCTION__ << " ("
                          << pszKey << ", "
                          << pszValue << ")"
                          << endl;
#endif

   UPDF_Device *pUPDFDevice = UPDF_Device::isAUPDF_Device (pDevice_d);
   xmlNodePtr  node         = 0;
   string      stringKey    = pszKey;
   string     *pstringRet   = 0;

   if (  pszKey
      && *pszKey
      && pUPDFDevice
      )
   {
      node = xmlObjectMap_d[stringKey];

      if (node)
      {
         node = FINDUDRENTRYKEYVALUE (pUPDFDevice,
                                      "ID",
                                      pszValue,
                                      UPDFDeviceInstance);

         char *pszNameID = (char *)xmlGetProp (node, (xmlChar *)"Name_ID");

         if (pszNameID)
         {
            xmlNodePtr nodeLocale = FINDLOCALEENTRYKEYVALUE (pUPDFDevice,
                                                             "Name_ID",
                                                             pszNameID,
                                                             UPDFDeviceInstance);

            if (nodeLocale)
            {
               char *pszLocalizedString = (char *)xmlGetProp (nodeLocale,
                                                              (xmlChar *)"Localized_String");

               if (pszLocalizedString)
               {
                  pstringRet = new string (pszKey);

                  *pstringRet += "=";
                  *pstringRet += pszLocalizedString;

                  free (pszLocalizedString);
               }
            }

            free (pszNameID);
         }
      }
   }

#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) cerr << "UPDF_Instance::" << __FUNCTION__ << " returns "
                          << (pstringRet ? *pstringRet : "NULL")
                          << endl;
#endif

   return pstringRet;
}

void UPDF_Instance::
setupPrinter ()
{
   if (fHaveSetupPrinter_d)
      return;

#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) cerr << "UPDF_Instance::setupPrinter ()" << endl;
#endif

   fHaveSetupPrinter_d = true;
}

bool UPDF_Instance::
beginJob ()
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) cerr << "UPDF_Instance::" << __FUNCTION__ << " ()" << endl;
#endif

   return executeEvent ("Event_PDL_Start", true);
}

bool UPDF_Instance::
beginJob (bool fJobPropertiesChanged)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) cerr << "UPDF_Instance::" << __FUNCTION__ << " (with props)" << endl;
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) cerr << "The job properties are as follows:" << endl;
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) cerr << "The orientation is = " << getCurrentOrientation() << endl;
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) cerr << "iDitherID = " << getCurrentDitherID () << endl;
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) cerr << "The form is = " << getCurrentForm () << endl;
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) cerr << "The tray is = " << getCurrentTray () << endl;
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) cerr << "The media is = " << getCurrentMedia () << endl;
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) cerr << "The resolution is = " << getCurrentResolution () << endl;
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) cerr << "The print mode is = " << getCurrentPrintMode () << endl;
#endif

   // @TBD - reinitialize with new job properties

   // Call common code
   return beginJob ();
}

bool UPDF_Instance::
newFrame ()
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) cerr << "UPDF_Instance::" << __FUNCTION__ << " ()" << endl;
#endif

   ditherNewFrame ();

   ptlPrintHead_d.x = 0;
   ptlPrintHead_d.y = 0;

   return true;
}

bool UPDF_Instance::
newFrame (bool fJobPropertiesChanged)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) cerr << "UPDF_Instance::" << __FUNCTION__ << " (with props)" << endl;
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) cerr << "The job properties are as follows:" << endl;
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) cerr << "The orientation is = " << getCurrentOrientation() << endl;
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) cerr << "iDitherID = " << getCurrentDitherID () << endl;
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) cerr << "The form is = " << getCurrentForm () << endl;
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) cerr << "The tray is = " << getCurrentTray () << endl;
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) cerr << "The media is = " << getCurrentMedia () << endl;
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) cerr << "The resolution is = " << getCurrentResolution () << endl;
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) cerr << "The print mode is = " << getCurrentPrintMode () << endl;
#endif

   // @TBD - reinitialize with new job properties

   // Call common code
   return newFrame ();
}

bool UPDF_Instance::
endJob ()
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) cerr << "UPDF_Instance::" << __FUNCTION__ << " ()" << endl;
#endif

   return true;
}

bool UPDF_Instance::
abortJob ()
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) cerr << "UPDF_Instance::" << __FUNCTION__ << " ()" << endl;
#endif

   return true;
}

char * UPDF_Instance::
getXMLObjectValue (char *pszObjectName,
                   char *pszVariableName)
{
   if (  !pszObjectName
      || !pszVariableName
      )
   {
      return 0;
   }

   string     stringObjectName = pszObjectName;
   xmlNodePtr elm              = 0;

   elm = xmlObjectMap_d[stringObjectName];

   if (elm)
   {
      return (char *)xmlGetProp (elm, (xmlChar *)pszVariableName);
   }

   return 0;
}

bool UPDF_Instance::
executeEvent (char *pszPrefix, bool fPreEvent)
{
   char        achParameterName[128]; // @TBD
   DeviceData *pData                 = getDeviceData ();
   bool        fExecute              = fPreEvent;

   if (!pData)
      return true;

   int iCount = 0;

   sprintf (achParameterName, "%s_count", pszPrefix);

   if (  !pData->getIntData (achParameterName, &iCount)
      || 0 == iCount
      )
      return true;

   for (int i = 0; i < iCount; i++)
   {
      BinaryData *pbdExpression = 0;
      char       *pszEventName  = 0;

      sprintf (achParameterName, "%s_binary_%d", pszPrefix, i);

      if (pData->getBinaryData (achParameterName, &pbdExpression))
      {
      }
      else if (pData->getStringData (achParameterName, &pszEventName))
      {
         fExecute = !fExecute;
      }
      else
      {
      }
   }

   return true;
}

string UPDF_Instance::
toString (ostringstream& oss)
{
   ostringstream oss2;

   oss << "{ "
       << DeviceInstance::toString (oss2) // Add parent's output as well
       << " }";

   return oss.str ();
}

/* Provide a way to print out class data
*/
ostream&
operator<< (ostream& os, const UPDF_Instance& const_self)
{
   UPDF_Instance& self = const_cast<UPDF_Instance&>(const_self);
   ostringstream  oss;

   os << self.toString (oss);

   return os;
}
