/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include <ctype.h>
#include <stdio.h>
#include <sys/stat.h>
#include <iostream>
#include <fstream>

#include "UPDFDomParser.hpp"
#include "MyErrorHandler.hpp"
#include "ParameterConverter.hpp"

const static bool fDebugOutput = true;

extern char *vpszExeName;

UPDFDomParser::
UPDFDomParser (const char *pszXMLFile,
               xmlNodePtr  nodeRootCommands,
               xmlNodePtr  nodeLocale)
{
   pstrXMLFile_d      = new string (pszXMLFile);
   fErrorCondition_d  = false;
   doc_d              = 0;
   nodeRootCommands_d = nodeRootCommands;
   nodeLocale_d       = nodeLocale;
}

UPDFDomParser::
~UPDFDomParser ()
{
   delete pstrXMLFile_d;

   if (doc_d)
   {
      xmlFreeDoc (doc_d);
   }
}

void UPDFDomParser::
setErrorCondition ()
{
   fErrorCondition_d = true;
}

void
outputHeader (ofstream *pofstream, char *pszType)
{
   *pofstream << "<?xml version=\"1.0\" encoding=\"UTF-8\"?>" << endl;
   *pofstream << "<!DOCTYPE " << pszType << " SYSTEM \"../OmniDevice.dtd\">" << endl;
   *pofstream << "<!--" << endl;
   *pofstream << "     IBM Omni driver" << endl;
   *pofstream << "     Copyright (c) International Business Machines Corp., 2000" << endl;
   *pofstream << "" << endl;
   *pofstream << "     This library is free software; you can redistribute it and/or modify" << endl;
   *pofstream << "     it under the terms of the GNU Lesser General Public License as published" << endl;
   *pofstream << "     by the Free Software Foundation; either version 2.1 of the License, or" << endl;
   *pofstream << "     (at your option) any later version." << endl;
   *pofstream << "" << endl;
   *pofstream << "     This library is distributed in the hope that it will be useful," << endl;
   *pofstream << "     but WITHOUT ANY WARRANTY; without even the implied warranty of" << endl;
   *pofstream << "     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See" << endl;
   *pofstream << "     the GNU Lesser General Public License for more details." << endl;
   *pofstream << "" << endl;
   *pofstream << "     You should have received a copy of the GNU Lesser General Public License" << endl;
   *pofstream << "     along with this library; if not, write to the Free Software" << endl;
   *pofstream << "     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA" << endl;
   *pofstream << "-->" << endl;
}

xmlNodePtr
firstNode (xmlNodePtr elm)
{
   while (  0 != elm
         && (  XML_COMMENT_NODE == elm->type
            || XML_TEXT_NODE    == elm->type
            )
         )
   {
      // Move to the next node
      elm = elm->next;
   }

   return elm;
}

xmlNodePtr
nextNode (xmlNodePtr elm)
{
   do
   {
      // Move to the next node
      elm = elm->next;

   } while (  0 != elm
           && (  XML_COMMENT_NODE == elm->type
              || XML_TEXT_NODE    == elm->type
              )
           );

   return elm;
}

xmlNodePtr
findEntry (xmlNodePtr root, char *pszName)
{
   if (root == 0)
      return 0;

   xmlNodePtr elm = firstNode (root->xmlChildrenNode);

   while (elm != 0)
   {
      if (fDebugOutput) cerr << elm->name << " " << pszName << endl;

      if (0 == strcmp ((char *)elm->name, pszName))
      {
         break;
      }

      elm = elm->next;
   }

   return elm;
}

void
printXMLDocument (xmlDocPtr doc, int iLevel)
{
   xmlDocDump (stdout, doc);
}

void
printXMLNode (xmlNodePtr elm)
{
   xmlElemDump (stdout, elm->doc, elm);
   fprintf (stdout, "\n");
}

char *
bundleStringData (xmlNodePtr nodeItem)
{
   return (char *)xmlNodeListGetString (nodeItem->doc, nodeItem->xmlChildrenNode, 1);
}

int
bundleIntegerData (xmlNodePtr nodeItem)
{
   char *pszData = (char *)xmlNodeListGetString (nodeItem->doc, nodeItem->xmlChildrenNode, 1);
   int   iRet    = 0;

   if (1 != sscanf (pszData, "%d", &iRet))
   {
      iRet = 0;
   }

   free (pszData);

   return iRet;
}

void
printNodeName (xmlNodePtr cur)
{
   const char *pszNodeName = 0;

   if (!fDebugOutput)
      return;

   pszNodeName = bundleStringData (cur);

   cerr << "Node <"
        << cur->name
        << " "
        << pszNodeName
        << ">"
        << endl;

   if (pszNodeName)
   {
      free ((char *)pszNodeName);
   }
}

char *
encodeToXML (byte *pbFrom, int cbData)
{
   char *pszRet = 0;

   pszRet = (char *)malloc (cbData * 7 + (cbData - 1) + 1);

   if (pszRet)
   {
      char *pszTo = pszRet;

      for (int i = 0; i < cbData; i++)
      {
         pszTo += sprintf (pszTo, "HEX(%02X)", pbFrom[i]);

         if (i + 1 < cbData)
         {
            *pszTo++ = ' ';
         }
      }

      *pszTo++ = 0;
   }

   return pszRet;
}

char *
encodeToXMLExpression (byte *pbFrom, int cbDataFrom)

{
   class GrowableData
   {
   public:
   GrowableData ()
   {
      cbDataSize_d = BLOCK_SIZE;
      pszData_d    = (char *)malloc (cbDataSize_d);
      cbDataLeft_d = cbDataSize_d;
      iDataTo_d    = 0;

      if (!pszData_d)
      {
         throw "GrowableData::GrowableData: malloc failed!";
      }
   }

   void
   append (char ch)
   {
      if (8 > cbDataLeft_d)
      {
         cbDataSize_d += BLOCK_SIZE;

         pszData_d = (char *)realloc (pszData_d, cbDataSize_d);

         if (!pszData_d)
         {
            throw "GrowableData::append: realloc failed!";
         }

         cbDataLeft_d += BLOCK_SIZE;
      }

      int iDigit;
      int iFirstDigit;
      int iSecondDigit;

      iDigit        = (int)ch;
      iFirstDigit   = iDigit % 16;
      iDigit       /= 16;
      iSecondDigit  = iDigit % 16;

      if (0 < iDataTo_d)
      {
         pszData_d[iDataTo_d++] = ' ';
         cbDataLeft_d--;
      }

      pszData_d[iDataTo_d++] = 'H';
      pszData_d[iDataTo_d++] = 'E';
      pszData_d[iDataTo_d++] = 'X';
      pszData_d[iDataTo_d++] = '(';
      if (10 > iSecondDigit)
      {
         pszData_d[iDataTo_d++] = '0' + iSecondDigit;
      }
      else
      {
         pszData_d[iDataTo_d++] = 'A' + iSecondDigit - 10;
      }
      if (10 > iFirstDigit)
      {
         pszData_d[iDataTo_d++] = '0' + iFirstDigit;
      }
      else
      {
         pszData_d[iDataTo_d++] = 'A' + iFirstDigit - 10;
      }
      pszData_d[iDataTo_d++] = ')';

      cbDataLeft_d -= 7;

      pszData_d[iDataTo_d] = '\0';
   }

   void
   appendHex (char ch)
   {
      int iDigit;
      int iFirstDigit;
      int iSecondDigit;
      int iThirdDigit;

      iDigit        = (int)ch;
      iFirstDigit   = iDigit % 10;
      iDigit       /= 10;
      iSecondDigit  = iDigit % 10;
      iDigit       /= 10;
      iThirdDigit   = iDigit % 10;

      append ('%');
      append ('H');
      append ('1');
      append ('(');
      if (iThirdDigit)  append ('0' + iThirdDigit);
      if (iSecondDigit) append ('0' + iSecondDigit);
      if (iFirstDigit)  append ('0' + iFirstDigit);
      append (')');
   }

   char *
   getData ()
   {
      return pszData_d;
   }

   private:
      enum {
         BLOCK_SIZE = 16 * 1024
      };

      int   cbDataSize_d;
      char *pszData_d;
      int   cbDataLeft_d;
      int   iDataTo_d;
   };

   try
   {
      GrowableData data;

      for (int i = 0; i < cbDataFrom; i++)
      {
         switch (pbFrom[i])
         {
         case ' ':
         case '(':
         case ')':
         case '+':
         case '-':
         case '*':
         case '/':
         case '=':
         case '!':
         case '<':
         case '>':
         {
            data.appendHex (pbFrom[i]);
            break;
         }

         case '%':
         {
            switch (pbFrom[i + 1])
            {
            case 'A':
            case 'F':
            case 'H':
            case 'L':
            case 'K':
            case 'I':
            case 'E':
            {
               data.append (pbFrom[i]);
               break;
            }

            default:
            {
               data.appendHex (pbFrom[i]);
               break;
            }
            }
            break;
         }

         default:
         {
            data.append (pbFrom[i]);
            break;
         }
         }
      }

      char *pszRet = data.getData ();

      return pszRet;
   }
   catch (const char *pszException)
   {
      cerr << "Caught exception: " << pszException << endl;
      return 0;
   }
   catch (...)
   {
      cerr << "Caught unknown exception @" << __FILE__ << ":" << __LINE__ << endl;
      return 0;
   }
}

char *
convertToBinaryData (xmlNodePtr node)
{
   char        *pszRet              = 0;
   char        *pszCommand          = 0;
   char        *pszComSeqParameter1 = 0;
   char        *pszComSeqParameter2 = 0;
   char        *pszComSeqParameter3 = 0;
   char        *pszComSeqParameter4 = 0;
   char        *pszComSeqParameter5 = 0;
   char        *pszComSeqParameter6 = 0;
   char        *pszExpression       = 0;
   Expression  *pResult             = 0;
   SYMBOLTABLE  symbolTable;

   pszCommand          = (char *)xmlGetProp (node, (xmlChar *)"CommandSequence");
   pszComSeqParameter1 = (char *)xmlGetProp (node, (xmlChar *)"ComSeq_Parameter_1");
   pszComSeqParameter2 = (char *)xmlGetProp (node, (xmlChar *)"ComSeq_Parameter_2");
   pszComSeqParameter3 = (char *)xmlGetProp (node, (xmlChar *)"ComSeq_Parameter_3");
   pszComSeqParameter4 = (char *)xmlGetProp (node, (xmlChar *)"ComSeq_Parameter_4");
   pszComSeqParameter5 = (char *)xmlGetProp (node, (xmlChar *)"ComSeq_Parameter_5");
   pszComSeqParameter6 = (char *)xmlGetProp (node, (xmlChar *)"ComSeq_Parameter_6");

///printXMLNode (node);

///if (fDebugOutput) cerr << "pszCommand          = " << pszCommand << endl;
///if (fDebugOutput) cerr << "pszComSeqParameter1 = " << pszComSeqParameter1 << endl;
///if (fDebugOutput) cerr << "pszComSeqParameter2 = " << pszComSeqParameter2 << endl;
///if (fDebugOutput) cerr << "pszComSeqParameter3 = " << pszComSeqParameter3 << endl;
///if (fDebugOutput) cerr << "pszComSeqParameter4 = " << pszComSeqParameter4 << endl;
///if (fDebugOutput) cerr << "pszComSeqParameter5 = " << pszComSeqParameter5 << endl;
///if (fDebugOutput) cerr << "pszComSeqParameter6 = " << pszComSeqParameter6 << endl;

   if (pszComSeqParameter1)
      symbolTable[string ("ComSeq_Parameter_1")] = new Expression (pszComSeqParameter1, true);
   if (pszComSeqParameter2)
      symbolTable[string ("ComSeq_Parameter_2")] = new Expression (pszComSeqParameter2, true);
   if (pszComSeqParameter3)
      symbolTable[string ("ComSeq_Parameter_3")] = new Expression (pszComSeqParameter3, true);
   if (pszComSeqParameter4)
      symbolTable[string ("ComSeq_Parameter_4")] = new Expression (pszComSeqParameter4, true);
   if (pszComSeqParameter5)
      symbolTable[string ("ComSeq_Parameter_5")] = new Expression (pszComSeqParameter5, true);
   if (pszComSeqParameter6)
      symbolTable[string ("ComSeq_Parameter_6")] = new Expression (pszComSeqParameter6, true);

   pszExpression = pszCommand;

   try
   {
      if (fDebugOutput) cerr << "pszExpression = " << pszExpression << endl;

      pResult = convert (&pszExpression,
                         true,
                         false,
                         &symbolTable);

      if (fDebugOutput) cerr << "pResult = " << *pResult << endl;

      if (pResult)
      {
         if (pResult->containsExpression ())
         {
            pszRet = encodeToXMLExpression (pResult->getBinaryData (), pResult->getLength ());
         }
         else
         {
            pszRet = encodeToXML (pResult->getBinaryData (), pResult->getLength ());
         }
      }
   }
   catch (ExpressionException *pException)
   {
      cerr << "Caught " << *pException << " for " << pszCommand << endl;
   }
   catch (...)
   {
      cerr << "Caught unknown exception @" << __FILE__ << ":" << __LINE__ << endl;
      throw;
   }

   if (pszCommand)
   {
      free (pszCommand);
   }
   if (pszComSeqParameter1)
   {
      free (pszComSeqParameter1);
   }
   if (pszComSeqParameter2)
   {
      free (pszComSeqParameter2);
   }
   if (pszComSeqParameter3)
   {
      free (pszComSeqParameter3);
   }
   if (pszComSeqParameter4)
   {
      free (pszComSeqParameter4);
   }
   if (pszComSeqParameter5)
   {
      free (pszComSeqParameter5);
   }
   if (pszComSeqParameter6)
   {
      free (pszComSeqParameter6);
   }

   delete pResult;

   for ( SYMBOLTABLE::iterator next = symbolTable.begin () ;
         next != symbolTable.end () ;
         next++ )
   {
      delete (*next).second;
   }

   return pszRet;
}

char *
convertUPDFFormToOmniForm (char *pszUPDFName)
{
   if (0 == strcasecmp (pszUPDFName, "Size_Letter"))
      return "FORM_LETTER";
   if (0 == strcasecmp (pszUPDFName, "Size_Legal"))
      return "FORM_LEGAL";
   if (0 == strcasecmp (pszUPDFName, "Size_Executive"))
      return "FORM_EXECUTIVE";
   if (0 == strcasecmp (pszUPDFName, "Size_A5"))
      return "FORM_A5";
   if (0 == strcasecmp (pszUPDFName, "Size_A4"))
      return "FORM_A4";
   if (0 == strcasecmp (pszUPDFName, "Size_A3"))
      return "FORM_A3";
///if (0 == strcasecmp (pszUPDFName, "Size_Com10"))
///   return "FORM_";
   if (0 == strcasecmp (pszUPDFName, "Size_DL"))
      return "FORM_DL_ENVELOPE";
   if (0 == strcasecmp (pszUPDFName, "Size_C5"))
      return "FORM_C5";
   if (0 == strcasecmp (pszUPDFName, "Size_B5"))
      return "FORM_B5";
   if (0 == strcasecmp (pszUPDFName, "Size_Monarch"))
      return "FORM_MONARCH_ENVELOPE";
   if (0 == strcasecmp (pszUPDFName, "Size_Ledger"))
      return "FORM_LEDGER";
   if (0 == strcasecmp (pszUPDFName, "Size_JISB4"))
      return "FORM_JIS_B4";
   if (0 == strcasecmp (pszUPDFName, "Size_JISB5"))
      return "FORM_JIS_B5";
///if (0 == strcasecmp (pszUPDFName, "Size_DoublePostcard"))
///   return "FORM_";
   if (0 == strcasecmp (pszUPDFName, "Size_JISEXEC"))
      return "FORM_JIS_EXEC";
   if (0 == strcasecmp (pszUPDFName, "Size_ROC16K"))
      return "FORM_ROC_16K";
   if (0 == strcasecmp (pszUPDFName, "Size_ROC8K"))
      return "FORM_ROC_8K";

   return 0;
}

int
converUPDFMarginToOmniMargin (char *pszMargin)
{
   char *pszDot;
   int   iReturn = 0;

   pszDot = strchr (pszMargin, '.');

   if (pszDot)
   {
      sscanf (pszDot + 1, "%d", &iReturn);
   }

   return iReturn; // @TBD - convert to 1/1000 mm
}

char *
convertUPDFMediaToOmniMedia (char *pszUPDFName)
{
   if (0 == strcasecmp (pszUPDFName, "Type_Auto"))
      return "MEDIA_AUTO";
   if (0 == strcasecmp (pszUPDFName, "Type_Plain"))
      return "MEDIA_PLAIN";
   if (0 == strcasecmp (pszUPDFName, "Type_Preprinted"))
      return "MEDIA_PREPRINTED";
   if (0 == strcasecmp (pszUPDFName, "Type_Letterhead"))
      return "MEDIA_LETTERHEAD";
   if (0 == strcasecmp (pszUPDFName, "Type_Transparency"))
      return "MEDIA_TRANSPARENCY";
   if (0 == strcasecmp (pszUPDFName, "Type_Prepunched"))
      return "MEDIA_PREPUNCHED";
   if (0 == strcasecmp (pszUPDFName, "Type_Labels"))
      return "MEDIA_LABELS";
   if (0 == strcasecmp (pszUPDFName, "Type_Bond"))
      return "MEDIA_BOND";
   if (0 == strcasecmp (pszUPDFName, "Type_Recycled"))
      return "MEDIA_RECYCLED";
   if (0 == strcasecmp (pszUPDFName, "Type_Color"))
      return "MEDIA_COLOR";
   if (0 == strcasecmp (pszUPDFName, "Type_CardStock"))
      return "MEDIA_CARDSTOCK";
   if (0 == strcasecmp (pszUPDFName, "Type_Rough"))
      return "MEDIA_ROUGH";

   return 0;
}

char *
convertUPDFTrayToOmniTray (char *pszUPDFName)
{
   if (0 == strcasecmp (pszUPDFName, "Source_Auto"))
      return "TRAY_AUTO";
   if (0 == strcasecmp (pszUPDFName, "Source_Tray2"))
      return "TRAY_TRAY_2";
   if (0 == strcasecmp (pszUPDFName, "Source_Tray3"))
      return "TRAY_TRAY_3";

   return 0;
}

char *
convertUPDFResolutionToOmniResolution (int iXRes, int iYRes)
{
   if (60 == iXRes && 60 == iYRes)
      return "RESOLUTION_60_X_60";
   if (60 == iXRes && 72 == iYRes)
      return "RESOLUTION_60_X_72";
   if (60 == iXRes && 180 == iYRes)
      return "RESOLUTION_60_X_180";
   if (60 == iXRes && 360 == iYRes)
      return "RESOLUTION_60_X_360";
   if (72 == iXRes && 72 == iYRes)
      return "RESOLUTION_72_X_72";
   if (75 == iXRes && 75 == iYRes)
      return "RESOLUTION_75_X_75";
   if (80 == iXRes && 60 == iYRes)
      return "RESOLUTION_80_X_60";
   if (80 == iXRes && 72 == iYRes)
      return "RESOLUTION_80_X_72";
   if (90 == iXRes && 180 == iYRes)
      return "RESOLUTION_90_X_180";
   if (90 == iXRes && 360 == iYRes)
      return "RESOLUTION_90_X_360";
   if (90 == iXRes && 60 == iYRes)
      return "RESOLUTION_90_X_60";
   if (90 == iXRes && 72 == iYRes)
      return "RESOLUTION_90_X_72";
   if (90 == iXRes && 90 == iYRes)
      return "RESOLUTION_90_X_90";
   if (100 == iXRes && 100 == iYRes)
      return "RESOLUTION_100_X_100";
   if (120 == iXRes && 60 == iYRes)
      return "RESOLUTION_120_X_60";
   if (120 == iXRes && 72 == iYRes)
      return "RESOLUTION_120_X_72";
   if (120 == iXRes && 144 == iYRes)
      return "RESOLUTION_120_X_144";
   if (120 == iXRes && 180 == iYRes)
      return "RESOLUTION_120_X_180";
   if (120 == iXRes && 360 == iYRes)
      return "RESOLUTION_120_X_360";
   if (144 == iXRes && 72 == iYRes)
      return "RESOLUTION_144_X_72";
   if (150 == iXRes && 150 == iYRes)
      return "RESOLUTION_150_X_150";
   if (180 == iXRes && 180 == iYRes)
      return "RESOLUTION_180_X_180";
   if (180 == iXRes && 360 == iYRes)
      return "RESOLUTION_180_X_360";
   if (200 == iXRes && 200 == iYRes)
      return "RESOLUTION_200_X_200";
   if (240 == iXRes && 60 == iYRes)
      return "RESOLUTION_240_X_60";
   if (240 == iXRes && 72 == iYRes)
      return "RESOLUTION_240_X_72";
   if (240 == iXRes && 144 == iYRes)
      return "RESOLUTION_240_X_144";
   if (240 == iXRes && 240 == iYRes)
      return "RESOLUTION_240_X_240";
   if (300 == iXRes && 300 == iYRes)
      return "RESOLUTION_300_X_300";
   if (360 == iXRes && 180 == iYRes)
      return "RESOLUTION_360_X_180";
   if (360 == iXRes && 360 == iYRes)
      return "RESOLUTION_360_X_360";
   if (360 == iXRes && 720 == iYRes)
      return "RESOLUTION_360_X_720";
   if (400 == iXRes && 400 == iYRes)
      return "RESOLUTION_400_X_400";
   if (600 == iXRes && 300 == iYRes)
      return "RESOLUTION_600_X_300";
   if (600 == iXRes && 600 == iYRes)
      return "RESOLUTION_600_X_600";
   if (720 == iXRes && 360 == iYRes)
      return "RESOLUTION_720_X_360";
   if (720 == iXRes && 720 == iYRes)
      return "RESOLUTION_720_X_720";
   if (1200 == iXRes && 600 == iYRes)
      return "RESOLUTION_1200_X_600";
   if (1200 == iXRes && 1200 == iYRes)
      return "RESOLUTION_1200_X_1200";
   if (1440 == iXRes && 720 == iYRes)
      return "RESOLUTION_1440_X_720";

   return "RESOLUTION_UNLISTED";
}

char *
convertUPDFOrientationToOmniOrientation (char *pszUPDFName)
{
   if (0 == strcasecmp (pszUPDFName, "Orientation_Portrait"))
      return "ORIENTATION_PORTRAIT";
   if (0 == strcasecmp (pszUPDFName, "Orientation_Landscape"))
      return "ORIENTATION_LANDSCAPE";
   if (0 == strcasecmp (pszUPDFName, "Orientation_PortraitRotated"))
      return "ORIENTATION_REVERSE_PORTRAIT";
   if (0 == strcasecmp (pszUPDFName, "Orientation_LandscapeRotated"))
      return "ORIENTATION_REVERSE_LANDSCAPE";

   return "ORIENTATION_UNLISTED";
}

char * UPDFDomParser::
findCommand (char *pszCmd, bool fEvaluate)
{
   xmlNodePtr  nodeCurrent    = nodeRootCommands_d;
   char       *pszOmniCommand = 0;

   if (fDebugOutput) cerr << "findCommand (" << pszCmd << ", " << fEvaluate << ")" << endl;

   if ((nodeCurrent = findEntry (nodeCurrent, "CommandSequences")) != 0)
   {
      xmlNodePtr nodeItem = firstNode (nodeCurrent->xmlChildrenNode); // first CommandSequences

      while (  nodeItem != 0
            && !pszOmniCommand
            )
      {
         char *pszCommandSequenceID = (char *)xmlGetProp (nodeItem, (xmlChar *)"CommandSequence_ID");

         if (fDebugOutput) cerr << "pszCommandSequenceID = " << pszCommandSequenceID << endl;

         if (  pszCommandSequenceID
            && 0 == strcmp (pszCmd, pszCommandSequenceID)
            )
         {
            if (fEvaluate)
            {
               pszOmniCommand = convertToBinaryData (nodeCurrent);
            }
            else
            {
               pszOmniCommand = (char *)xmlGetProp (nodeItem, (xmlChar *)"CommandSequence");
            }
         }

         nodeItem = nextNode (nodeItem);

         if (pszCommandSequenceID)
         {
            free (pszCommandSequenceID);
         }
      }
   }

   return pszOmniCommand;
}

xmlNodePtr UPDFDomParser::
findCommandNode (char *pszCmd)
{
   xmlNodePtr  nodeCurrent    = firstNode (nodeRootCommands_d->xmlChildrenNode);
   xmlNodePtr  nodeReturn     = 0;

   while (nodeCurrent != 0)
   {
      char *pszCommandSequenceID = (char *)xmlGetProp (nodeCurrent, (xmlChar *)"CommandSequence_ID");

      if (  pszCommandSequenceID
         && 0 == strcmp (pszCmd, pszCommandSequenceID)
         )
      {
         nodeReturn = nodeCurrent;
      }

      if (pszCommandSequenceID)
      {
         free (pszCommandSequenceID);
      }

      if (nodeReturn)
      {
         return nodeCurrent;
      }

      nodeCurrent = nextNode (nodeCurrent);
   }

   return 0;
}

char * UPDFDomParser::
findEventCommand (char       *pszEventCommand,
                  xmlNodePtr  nodeRasterGraphicList)
{
   char       *pszCommandName = 0;
   xmlNodePtr  nodeItem       = firstNode (nodeRasterGraphicList->xmlChildrenNode);

   while (nodeItem != 0)
   {
      pszCommandName = (char *)xmlGetProp (nodeItem, (xmlChar *)pszEventCommand);

      if (pszCommandName)
      {
         char *pszCommandSequence = 0;

         nodeItem = firstNode (nodeRootCommands_d->xmlChildrenNode);

         while (nodeItem != 0)
         {
            char *pszCommandSequenceID = (char *)xmlGetProp (nodeItem, (xmlChar *)"CommandSequence_ID");

            if (  pszCommandSequenceID
               && 0 == strcmp (pszCommandName, pszCommandSequenceID)
               )
            {
#if 0
               pszCommandSequence = (char *)xmlGetProp (nodeItem, (xmlChar *)"CommandSequence");
#else
               return pszCommandSequenceID;
#endif
            }

            nodeItem = nextNode (nodeItem);

            if (pszCommandSequenceID)
            {
               free (pszCommandSequenceID);
            }
         }

         free (pszCommandName);

         return pszCommandSequence;
      }

      nodeItem = nextNode (nodeItem);
   }

   return 0;
}

void UPDFDomParser::
processForms (xmlNodePtr rootNode, ofstream *pofstream)
{
   xmlNodePtr nodeCurrent;

   if (  ((nodeCurrent = findEntry (rootNode,    "Device_Features")) != 0)
      && ((nodeCurrent = findEntry (nodeCurrent, "PrintCapabilities")) != 0)
      && ((nodeCurrent = findEntry (nodeCurrent, "Features")) != 0)
      && ((nodeCurrent = findEntry (nodeCurrent, "PrintMediaHandling")) != 0)
      && ((nodeCurrent = findEntry (nodeCurrent, "MediaSizeList")) != 0)
      )
   {
      xmlNodePtr nodeItem = firstNode (nodeCurrent->xmlChildrenNode); // first MediaSize

      outputHeader (pofstream, "deviceForms");

      *pofstream << "<deviceForms>" << endl;

      while (nodeItem != 0)
      {
         xmlNodePtr nodeHardwareMargins = 0;
         xmlNodePtr nodeElement         = 0;

         nodeHardwareMargins = findEntry (nodeItem, "MediaSizeHardwareMargins");

         if (!nodeHardwareMargins)
         {
            cerr << "Error: Could not findEntry @ " << __LINE__ << endl;

            return;
         }

         char *pszID                = (char *)xmlGetProp (nodeItem, (xmlChar *)"ID");
         char *pszCommandSequenceID = 0;
         char *pszTopMargin         = (char *)xmlGetProp (nodeHardwareMargins, (xmlChar *)"Top");
         char *pszBottomMargin      = (char *)xmlGetProp (nodeHardwareMargins, (xmlChar *)"Bottom");
         char *pszLeftMargin        = (char *)xmlGetProp (nodeHardwareMargins, (xmlChar *)"Left");
         char *pszRightMargin       = (char *)xmlGetProp (nodeHardwareMargins, (xmlChar *)"Right");
         char *pszOmniCommand       = 0;
         char *pszFormName          = 0;
         bool  fFreeOmniCommand     = true;

         if ((nodeElement = findEntry (nodeItem, "Command")) != 0)
         {
            pszCommandSequenceID = (char *)xmlGetProp (nodeElement, (xmlChar *)"CommandSequence_ID");

            if (pszCommandSequenceID)
               pszOmniCommand = findCommand (pszCommandSequenceID, true);
         }

         if (!pszOmniCommand)
         {
            cerr << "Warning: Missing command for " << pszID << endl;
            fFreeOmniCommand = false;
            pszOmniCommand   = "_NUL_";
         }

         pszFormName = convertUPDFFormToOmniForm (pszID);

         if (fDebugOutput) cerr << "pszID                = " << pszID << endl
                                << "pszCommandSequenceID = " << pszCommandSequenceID << endl
                                << "pszOmniCommand       = " << pszOmniCommand << endl
                                << "pszFormName          = " << pszFormName << endl;

         if (pszFormName)
         {
            *pofstream << "  <deviceForm>" << endl
                       << "    <name>" << pszFormName << "</name>" << endl
                       << "    <formCapabilities>NO_CAPABILITIES</formCapabilities>" << endl // @TBD
                       << "    <command>" << pszOmniCommand << "</command>" << endl
                       << "    <hardCopyCap>" << endl
                       << "      <hardCopyCapLeft>" << converUPDFMarginToOmniMargin (pszLeftMargin) << "</hardCopyCapLeft>" << endl
                       << "      <hardCopyCapTop>" << converUPDFMarginToOmniMargin (pszTopMargin) << "</hardCopyCapTop>" << endl
                       << "      <hardCopyCapRight>" << converUPDFMarginToOmniMargin (pszRightMargin) << "</hardCopyCapRight>" << endl
                       << "      <hardCopyCapBottom>" << converUPDFMarginToOmniMargin (pszBottomMargin) << "</hardCopyCapBottom>" << endl
                       << "    </hardCopyCap>" << endl
                       << "   </deviceForm>" << endl;

            listForms_d.push_back (string (pszFormName));
         }
         else
         {
            cerr << "Warning: Missing form: " << pszID << " " << pszCommandSequenceID << endl;
         }

         nodeItem = nextNode (nodeItem);

         if (pszID)
         {
            free (pszID);
         }
         if (pszCommandSequenceID)
         {
            free (pszCommandSequenceID);
         }
         if (  fFreeOmniCommand
            && pszOmniCommand
            )
         {
            free (pszOmniCommand);
         }
         if (pszTopMargin)
         {
            free (pszTopMargin);
         }
         if (pszBottomMargin)
         {
            free (pszBottomMargin);
         }
         if (pszLeftMargin)
         {
            free (pszLeftMargin);
         }
         if (pszRightMargin)
         {
            free (pszRightMargin);
         }
      }

      *pofstream << "</deviceForms>" << endl;
   }
   else
   {
      cerr << "Error: Could not findEntry @ " << __LINE__ << endl;
   }
}

void UPDFDomParser::
processMedias (xmlNodePtr rootNode, ofstream *pofstream)
{
   xmlNodePtr nodeCurrent;

   if (  ((nodeCurrent = findEntry (rootNode,    "Device_Features")) != 0)
      && ((nodeCurrent = findEntry (nodeCurrent, "PrintCapabilities")) != 0)
      && ((nodeCurrent = findEntry (nodeCurrent, "Features")) != 0)
      && ((nodeCurrent = findEntry (nodeCurrent, "PrintMediaHandling")) != 0)
      && ((nodeCurrent = findEntry (nodeCurrent, "MediaTypeList")) != 0)
      )
   {
      xmlNodePtr nodeItem = firstNode (nodeCurrent->xmlChildrenNode); // first MediaType

      outputHeader (pofstream, "deviceMedias");

      *pofstream << "<deviceMedias>" << endl;

      while (nodeItem != 0)
      {
         char *pszID                = (char *)xmlGetProp (nodeItem, (xmlChar *)"ID");
         char *pszCommandSequenceID = (char *)xmlGetProp (nodeItem, (xmlChar *)"CommandSequence_ID");
         char *pszOmniCommand       = 0;
         char *pszMediaName         = 0;
         bool  fFreeOmniCommand     = true;

         if (pszCommandSequenceID)
            pszOmniCommand = findCommand (pszCommandSequenceID, true);

         if (!pszOmniCommand)
         {
            cerr << "Warning: Missing command for " << pszID << endl;
            fFreeOmniCommand = false;
            pszOmniCommand   = "_NUL_";
         }

         pszMediaName = convertUPDFMediaToOmniMedia (pszID);

         if (fDebugOutput) cerr << "pszID                = " << pszID << endl
                                << "pszCommandSequenceID = " << pszCommandSequenceID << endl
                                << "pszOmniCommand       = " << pszOmniCommand << endl
                                << "pszMediaName         = " << pszMediaName << endl;

         if (pszMediaName)
         {
            *pofstream << "  <deviceMedia>" << endl
                       << "    <name>" << pszMediaName << "</name>" << endl
                       << "    <command>" << pszOmniCommand << "</command>" << endl
                       << "    <mediaColorAdjustRequired>0</mediaColorAdjustRequired>" << endl // @TBD
                       << "    <mediaAbsorption>MEDIA_HEAVY_ABSORPTION</mediaAbsorption>" << endl // @TBD
                       << "  </deviceMedia>" << endl;

            listMedias_d.push_back (string (pszMediaName));
         }
         else
         {
            cerr << "Warning: Missing media: " << pszID << " " << pszCommandSequenceID << endl;
         }

         nodeItem = nextNode (nodeItem);

         if (pszID)
         {
            free (pszID);
         }
         if (pszCommandSequenceID)
         {
            free (pszCommandSequenceID);
         }
         if (  fFreeOmniCommand
            && pszOmniCommand
            )
         {
            free (pszOmniCommand);
         }
      }

      *pofstream << "</deviceMedias>" << endl;
   }
   else
   {
      cerr << "Error: Could not findEntry @ " << __LINE__ << endl;
   }
}

void UPDFDomParser::
processTrays (xmlNodePtr rootNode, ofstream *pofstream)
{
   xmlNodePtr nodeCurrent;

   if (  ((nodeCurrent = findEntry (rootNode,    "Device_Features")) != 0)
      && ((nodeCurrent = findEntry (nodeCurrent, "PrintCapabilities")) != 0)
      && ((nodeCurrent = findEntry (nodeCurrent, "Features")) != 0)
      && ((nodeCurrent = findEntry (nodeCurrent, "PrintMediaHandling")) != 0)
      && ((nodeCurrent = findEntry (nodeCurrent, "MediaSourceList")) != 0)
      )
   {
      xmlNodePtr nodeItem = firstNode (nodeCurrent->xmlChildrenNode); // first MediaSize

      outputHeader (pofstream, "deviceTrays");

      *pofstream << "<deviceTrays>" << endl;

      while (nodeItem != 0)
      {
         char *pszID                = (char *)xmlGetProp (nodeItem, (xmlChar *)"ID");
         char *pszCommandSequenceID = (char *)xmlGetProp (nodeItem, (xmlChar *)"CommandSequence_ID");
         char *pszOmniCommand       = 0;
         char *pszTrayName          = 0;
         bool  fFreeOmniCommand     = true;

         if (pszCommandSequenceID)
            pszOmniCommand = findCommand (pszCommandSequenceID, true);

         if (!pszOmniCommand)
         {
            cerr << "Warning: Missing command for " << pszID << endl;
            fFreeOmniCommand = false;
            pszOmniCommand   = "_NUL_";
         }

         pszTrayName = convertUPDFTrayToOmniTray (pszID);

         if (fDebugOutput) cerr << "pszID                = " << pszID << endl
                                << "pszCommandSequenceID = " << pszCommandSequenceID << endl
                                << "pszOmniCommand       = " << pszOmniCommand << endl
                                << "pszTrayName          = " << pszTrayName << endl;

         if (pszTrayName)
         {
            *pofstream << "  <deviceTray>" << endl
                       << "    <name>" << pszTrayName << "</name>" << endl
                       << "    <trayType>TRAY_TYPE_AUTO</trayType>" << endl // @TBD
                       << "    <command>" << pszOmniCommand << "</command>" << endl
                       << "  </deviceTray>" << endl;

            listTrays_d.push_back (string (pszTrayName));
         }
         else
         {
            cerr << "Warning: Missing tray: " << pszID << " " << pszCommandSequenceID << endl;
         }

         nodeItem = nextNode (nodeItem);

         if (pszID)
         {
            free (pszID);
         }
         if (pszCommandSequenceID)
         {
            free (pszCommandSequenceID);
         }
         if (  fFreeOmniCommand
            && pszOmniCommand
            )
         {
            free (pszOmniCommand);
         }
      }

      *pofstream << "</deviceTrays>" << endl;
   }
   else
   {
      cerr << "Error: Could not findEntry @ " << __LINE__ << endl;
   }
}

void UPDFDomParser::
processResolutions (xmlNodePtr rootNode, ofstream *pofstream)
{
   xmlNodePtr  nodeCurrent;
   char       *pszVirtualUnits;
   int         iVirtualUnits;

   if (  ((nodeCurrent = findEntry (rootNode,    "Device_Features")) != 0)
      && ((nodeCurrent = findEntry (nodeCurrent, "PrintCapabilities")) != 0)
      && ((nodeCurrent = findEntry (nodeCurrent, "Header")) != 0)
      )
   {
      pszVirtualUnits = (char *)xmlGetProp (nodeCurrent, (xmlChar *)"VirtualUnits");
      iVirtualUnits   = atoi (pszVirtualUnits);
   }
   else
   {
      cerr << "Error: Could not findEntry @ " << __LINE__ << endl;
      return;
   }

   if (  ((nodeCurrent = findEntry (rootNode,    "Device_Features")) != 0)
      && ((nodeCurrent = findEntry (nodeCurrent, "PrintCapabilities")) != 0)
      && ((nodeCurrent = findEntry (nodeCurrent, "Features")) != 0)
      && ((nodeCurrent = findEntry (nodeCurrent, "DeviceResolutionList")) != 0)
      )
   {
      xmlNodePtr nodeItem = firstNode (nodeCurrent->xmlChildrenNode); // first DeviceResolution

      outputHeader (pofstream, "deviceResolutions");

      *pofstream << "<deviceResolutions>" << endl;

      while (nodeItem != 0)
      {
         char *pszID                = (char *)xmlGetProp (nodeItem, (xmlChar *)"ID");
         char *pszCommandSequenceID = (char *)xmlGetProp (nodeItem, (xmlChar *)"Master_CommandSequence"); // @TBD
         char *pszHorizontalRes     = (char *)xmlGetProp (nodeItem, (xmlChar *)"HorizontalResolution");
         char *pszVerticalRes       = (char *)xmlGetProp (nodeItem, (xmlChar *)"VerticalResolution");
         int   iVerticalRes         = atoi (pszVerticalRes);
         int   iHorizontalRes       = atoi (pszHorizontalRes);
         char *pszResolutionName    = 0;
         char *pszOmniCommand       = 0;
         bool  fFreeOmniCommand     = true;

         if (pszCommandSequenceID)
            pszOmniCommand = findCommand (pszCommandSequenceID, true);

         if (!pszOmniCommand)
         {
            cerr << "Warning: Missing command for " << pszID << endl;
            fFreeOmniCommand = false;
            pszOmniCommand   = "_NUL_";
         }

         iVerticalRes   = iVirtualUnits / iVerticalRes;
         iHorizontalRes = iVirtualUnits / iHorizontalRes;

         pszResolutionName = convertUPDFResolutionToOmniResolution (iHorizontalRes, iVerticalRes);

         if (fDebugOutput) cerr << "pszID                = " << pszID << endl
                                << "pszCommandSequenceID = " << pszCommandSequenceID << endl
                                << "pszOmniCommand       = " << pszOmniCommand << endl
                                << "iVerticalRes         = " << iVerticalRes << endl
                                << "iHorizontalRes       = " << iHorizontalRes << endl;

         if (pszResolutionName)
         {
            *pofstream << "  <deviceResolution>" << endl
                       << "    <name>" << pszResolutionName << "</name>" << endl
                       << "    <xRes>" << iVerticalRes << "</xRes>" << endl
                       << "    <yRes>" << iHorizontalRes << "</yRes>" << endl
                       << "    <command>" << pszOmniCommand << "</command>" << endl
                       << "    <resolutionCapability>0</resolutionCapability>" << endl // @TBD
                       << "    <resolutionDestinationBitsPerPel>1</resolutionDestinationBitsPerPel>" << endl // @TBD
                       << "    <resolutionScanlineMultiple>1</resolutionScanlineMultiple>" << endl // @TBD
                       << "  </deviceResolution>" << endl;
         }
         else
         {
            cerr << "Warning: Missing resolution: " << pszID << " " << pszCommandSequenceID << endl;
         }

         nodeItem = nextNode (nodeItem);

         if (pszID)
         {
            free (pszID);
         }
         if (pszCommandSequenceID)
         {
            free (pszCommandSequenceID);
         }
         if (  fFreeOmniCommand
            && pszOmniCommand
            )
         {
            free (pszOmniCommand);
         }
         if (pszHorizontalRes)
         {
            free (pszHorizontalRes);
         }
         if (pszVerticalRes)
         {
            free (pszVerticalRes);
         }
      }

      *pofstream << "</deviceResolutions>" << endl;

      if (pszVirtualUnits)
      {
         free (pszVirtualUnits);
      }
   }
   else
   {
      cerr << "Error: Could not findEntry @ " << __LINE__ << endl;
   }
}

void UPDFDomParser::
processOrientations (xmlNodePtr rootNode, ofstream *pofstream)
{
   bool fOrientationsFound = false;

   outputHeader (pofstream, "deviceOrientations");

   *pofstream << "<deviceOrientations>" << endl;

   xmlNodePtr nodeCurrent;

   if (  ((nodeCurrent = findEntry (rootNode,    "Device_Features")) != 0)
      && ((nodeCurrent = findEntry (nodeCurrent, "PrintCapabilities")) != 0)
      && ((nodeCurrent = findEntry (nodeCurrent, "Features")) != 0)
      && ((nodeCurrent = findEntry (nodeCurrent, "PrintMediaHandling")) != 0)
      && ((nodeCurrent = findEntry (nodeCurrent, "MediaPageOrientationList")) != 0)
      )
   {
      xmlNodePtr nodeItem = firstNode (nodeCurrent->xmlChildrenNode); // first MediaPageOrientation

      fOrientationsFound = true;

      while (nodeItem != 0)
      {
         char *pszID              = (char *)xmlGetProp (nodeItem, (xmlChar *)"ID");
         char *pszOrientationName = 0;

         pszOrientationName = convertUPDFOrientationToOmniOrientation (pszID);

         *pofstream << "  <deviceOrientation>" << endl
                    << "    <name>" << pszOrientationName << "</name>" << endl
                    << "  </deviceOrientation>" << endl;

         nodeItem = nextNode (nodeItem);

         if (pszID)
         {
            free (pszID);
         }
      }
   }
   else
   {
      cerr << "Error: Could not findEntry @ " << __LINE__ << endl;
   }

   if (!fOrientationsFound)
   {
      cerr << "Warning: Missing orientations.  Is this a printer?" << endl;

      *pofstream << "  <deviceOrientation>" << endl
                 << "    <name>ORIENTATION_PORTRAIT</name>" << endl
                 << "  </deviceOrientation>" << endl;
   }

   *pofstream << "</deviceOrientations>" << endl;
}

void UPDFDomParser::
processConnections (xmlNodePtr rootNode, ofstream *pofstream)
{
   outputHeader (pofstream, "deviceConnections");

   *pofstream << "<deviceConnections>" << endl;

   listForms_d.unique ();
   listTrays_d.unique ();
   listMedias_d.unique ();

   int iConnId = 1;

   for ( FormList::iterator nextForm = listForms_d.begin () ;
         nextForm != listForms_d.end () ;
         nextForm++ )
   {
      for ( TrayList::iterator nextTray = listTrays_d.begin () ;
            nextTray != listTrays_d.end () ;
            nextTray++ )
      {
         for ( MediaList::iterator nextMedia = listMedias_d.begin () ;
               nextMedia != listMedias_d.end () ;
               nextMedia++ )
         {
            *pofstream << "<deviceConnection>" << endl
                       << "  <name>" << "@TBD" << "_CONN_ID_" << iConnId << "</name>" << endl // @TBD
                       << "  <connectionForm>" << nextForm->c_str () << "</connectionForm>" << endl
                       << "  <connectionTray>" << nextTray->c_str () << "</connectionTray>" << endl
                       << "  <connectionMedia>" << nextMedia->c_str () << "</connectionMedia>" << endl
                       << "</deviceConnection>" << endl;

            iConnId++;
         }
      }
   }

   *pofstream << "</deviceConnections>" << endl;
}

void UPDFDomParser::
processData (xmlNodePtr rootNode, ofstream *pofstream)
{
   xmlNodePtr nodeCurrent           = 0;
   xmlNodePtr nodeRasterGraphicList = 0;

   if (  ((nodeCurrent = findEntry (rootNode,    "Device_Features")) != 0)
      && ((nodeCurrent = findEntry (nodeCurrent, "PrintCapabilities")) != 0)
      && ((nodeCurrent = findEntry (nodeCurrent, "Features")) != 0)
      && ((nodeCurrent = findEntry (nodeCurrent, "TBD_RasterGraphicList")) != 0)
      )
   {
      nodeRasterGraphicList = nodeCurrent;
   }
   else
   {
      cerr << "Error: Could not findEntry @ " << __LINE__ << endl;
   }

   if (!nodeRasterGraphicList)
   {
      return;
   }

   outputHeader (pofstream, "deviceDatas");

   *pofstream << "<deviceDatas>" << endl;

   if (  ((nodeCurrent = findEntry (rootNode,    "Device_Features")) != 0)
      && ((nodeCurrent = findEntry (nodeCurrent, "PrintCapabilities")) != 0)
      && ((nodeCurrent = findEntry (nodeCurrent, "EventList")) != 0)
      )
   {
      xmlNodePtr nodeItem = firstNode (nodeCurrent->xmlChildrenNode); // first EventList

      while (nodeItem != 0)
      {
         xmlNodePtr nodeChildren = firstNode (nodeItem->xmlChildrenNode);
         int        iNumChildren = 0;

         if (fDebugOutput) cerr << "name = " << nodeItem->name << endl;

         while (nodeChildren != 0)
         {
            static char *pszEventFeatureHeader = "RasterGraphic.";
            static int   cbEventFeatureHeader  = strlen (pszEventFeatureHeader);
            char        *pszEventFeature       = (char *)xmlGetProp (nodeChildren, (xmlChar *)"EventFeature");

            if (fDebugOutput) cerr << "pszEventFeature = " << pszEventFeature << endl;

            if (0 == strncmp (pszEventFeature, pszEventFeatureHeader, cbEventFeatureHeader))
            {
               char       *pszEventCommand = 0;
               char       *pszComSequence  = 0;
               char       *pszData         = 0;

               pszEventCommand = findEventCommand (pszEventFeature + cbEventFeatureHeader,
                                                   nodeRasterGraphicList);

               if (fDebugOutput) cerr << "pszEventCommand = " << pszEventCommand << endl;

               pszComSequence = findCommand (pszEventCommand, true);

               if (fDebugOutput) cerr << "pszComSequence = " << pszComSequence << endl;

               pszData = encodeToXMLExpression ((byte *)pszComSequence, strlen (pszComSequence));

               *pofstream << "  <deviceData name=\""
                          << nodeItem->name
                          << "_binary_"
                          << iNumChildren
                          << "\" type=\"binary\">"
                          << pszData
                          << "</deviceData>" << endl;

               if (pszData)
               {
                  free (pszData);
               }
               if (pszComSequence)
               {
                  free (pszComSequence);
               }
            }
            else
            {
               *pofstream << "  <deviceData name=\""
                          << nodeItem->name
                          << "_string_"
                          << iNumChildren
                          << "\" type=\"string\">"
                          << pszEventFeature
                          << "</deviceData>" << endl;
            }

            iNumChildren++;

            nodeChildren = nextNode (nodeChildren);

            if (pszEventFeature)
            {
               free (pszEventFeature);
            }
         }

         *pofstream << "  <deviceData name=\""
                    << nodeItem->name
                    << "_count\" type=\"integer\">"
                    << iNumChildren
                    << "</deviceData>" << endl;

         nodeItem = nextNode (nodeItem);
      }
   }
   else
   {
      cerr << "Error: Could not findEntry @ " << __LINE__ << endl;
   }

   *pofstream << "</deviceDatas>" << endl;
}

void UPDFDomParser::
processLocale (xmlNodePtr   rootNode,
               ofstream    *pofstream)
{
   static char *pszOrientation = "ORIENTATION_PORTRAIT";
   static char *pszResolution  = "RESOLUTION_600_X_600";
   static char *pszDither      = "DITHER_STUCKI_DIFFUSION";
   static char *pszPrintMode   = "PRINT_MODE_1_ANY";
   char        *pszForm        = 0;
   char        *pszTray        = 0;
   char        *pszMedia       = 0;

   xmlNodePtr nodeItem = firstNode (rootNode->xmlChildrenNode);

   while (nodeItem != 0)
   {
      if (0 == strcmp ((char *)nodeItem->name, "LocaleDefault"))
      {
         char *pszDefaultFeature = (char *)xmlGetProp (nodeItem, (xmlChar *)"Default_Feature");
         char *pszDefaultSetting = (char *)xmlGetProp (nodeItem, (xmlChar *)"Default_Setting");

         if (fDebugOutput) cerr << "pszDefaultFeature = " << pszDefaultFeature << endl;
         if (fDebugOutput) cerr << "pszDefaultSetting = " << pszDefaultSetting << endl;

         if (0 == strcmp (pszDefaultFeature, "MediaSize"))
         {
            char *pszFormName = convertUPDFFormToOmniForm (pszDefaultSetting);

            if (pszFormName)
            {
               pszForm = (char *)malloc (strlen (pszFormName));
               strcpy (pszForm, pszFormName);
            }
         }
         else if (0 == strcmp (pszDefaultFeature, "MediaType"))
         {
            char *pszMediaName = convertUPDFMediaToOmniMedia (pszDefaultSetting);

            if (pszMediaName)
            {
               pszMedia = (char *)malloc (strlen (pszMediaName));
               strcpy (pszMedia, pszMediaName);
            }
         }
         else if (0 == strcmp (pszDefaultFeature, "MediaSource"))
         {
            char *pszTrayName = convertUPDFTrayToOmniTray (pszDefaultSetting);

            if (pszTrayName)
            {
               pszTray = (char *)malloc (strlen (pszTrayName));
               strcpy (pszTray, pszTrayName);
            }
         }

         if (pszDefaultFeature)
         {
            free (pszDefaultFeature);
         }
         if (pszDefaultSetting)
         {
            free (pszDefaultSetting);
         }
      }

      nodeItem = nextNode (nodeItem);
   }

   *pofstream << "  <DefaultJobProperties>" << endl
              << "    <orientation>" << pszOrientation << "</orientation>" << endl
              << "    <form>" << pszForm << "</form>" << endl
              << "    <tray>" << pszTray << "</tray>" << endl
              << "    <media>" << pszMedia << "</media>" << endl
              << "    <resolution>" << pszResolution << "</resolution>" << endl
              << "    <dither>" << pszDither << "</dither>" << endl
              << "    <printmode>" << pszPrintMode << "</printmode>" << endl;

   nodeItem = firstNode (rootNode->xmlChildrenNode);

   while (nodeItem != 0)
   {
      if (0 == strcmp ((char *)nodeItem->name, "LocaleDefault"))
      {
         char *pszDefaultFeature = (char *)xmlGetProp (nodeItem, (xmlChar *)"Default_Feature");
         char *pszDefaultSetting = (char *)xmlGetProp (nodeItem, (xmlChar *)"Default_Setting");

         if (fDebugOutput) cerr << "pszDefaultFeature = " << pszDefaultFeature << endl;
         if (fDebugOutput) cerr << "pszDefaultSetting = " << pszDefaultSetting << endl;

/////////if (  0 != strcmp (pszDefaultFeature, "MediaSize.MediaSize_Global_ID")
/////////   && 0 != strcmp (pszDefaultFeature, "MediaType.MediaType_Global_ID")
/////////   && 0 != strcmp (pszDefaultFeature, "MediaSource.MediaSource_Global_ID")
/////////   )
/////////{
            *pofstream << "    <other>" << pszDefaultFeature << "=" << pszDefaultSetting << "</other>" << endl;
/////////}

         if (pszDefaultFeature)
         {
            free (pszDefaultFeature);
         }
         if (pszDefaultSetting)
         {
            free (pszDefaultSetting);
         }
      }

      nodeItem = nextNode (nodeItem);
   }

   *pofstream << "  </DefaultJobProperties>" << endl;

   if (pszForm)
   {
      free (pszForm);
   }
   if (pszTray)
   {
      free (pszTray);
   }
   if (pszMedia)
   {
      free (pszMedia);
   }
}

char *
findOmniLanguage (xmlNodePtr rootNode)
{
   typedef struct _OmniLanguageMapping {
      char *pszUPDF;
      char *pszOmni;
   } OMNILANGUAGEMAPPING, *POMNILANGUAGEMAPPING;
   OMNILANGUAGEMAPPING aOmniLanguages[] = {
      { "PCL5e", "PCL5E" }
   };

   xmlNodePtr nodeCurrent;

   if (  ((nodeCurrent = findEntry (rootNode,    "Device_Features")) != 0)
      && ((nodeCurrent = findEntry (nodeCurrent, "PrintCapabilities")) != 0)
      && ((nodeCurrent = findEntry (nodeCurrent, "Header")) != 0)
      && ((nodeCurrent = findEntry (nodeCurrent, "Languages")) != 0)
      )
   {
      xmlNodePtr nodeItem;

      nodeItem = findEntry (nodeCurrent, "PageDescriptionLanguage");

      if (nodeItem == 0)
      {
         cerr << "Error: Could not findEntry @ " << __LINE__ << endl;
         return 0;
      }

      char *pszLanguage = (char *)xmlGetProp (nodeItem, (xmlChar *)"PDLMajorLevel");

      if (pszLanguage)
      {
         if (fDebugOutput) cerr  << "Found language " << pszLanguage << endl;

         for (int i = 0; i < dimof (aOmniLanguages); i++)
         {
            if (0 == strcmp (pszLanguage, aOmniLanguages[i].pszUPDF))
            {
               free (pszLanguage);

               return aOmniLanguages[i].pszOmni;
            }
         }

         free (pszLanguage);
      }

//////nodeItem = findEntry (nodeCurrent, "JobControlLanguage");
   }
   else
   {
      cerr << "Error: Could not findEntry @ " << __LINE__ << endl;
   }

   return 0;
}

char *
findOmniDriverName (xmlNodePtr rootNode)
{
   xmlNodePtr nodeCurrent;

   if ((nodeCurrent = findEntry (rootNode, "Device_Header")) != 0)
   {
      char *pszDriverName = (char *)xmlGetProp (nodeCurrent, (xmlChar *)"Manufacturer");

      return pszDriverName;
   }
   else
   {
      cerr << "Error: Could not findEntry @ " << __LINE__ << endl;
   }

   return 0;
}

char *
findOmniDeviceName (xmlNodePtr rootNode)
{
   xmlNodePtr nodeCurrent;

   if ((nodeCurrent = findEntry (rootNode, "Device_Header")) != 0)
   {
      char *pszDeviceName = (char *)xmlGetProp (nodeCurrent, (xmlChar *)"ProductName");

      return pszDeviceName;
   }
   else
   {
      cerr << "Error: Could not findEntry @ " << __LINE__ << endl;
   }

   return 0;
}

bool UPDFDomParser::
parse ()
{
   const char *pszXMLFile = pstrXMLFile_d->c_str ();
   xmlValidCtxt validator = {
      this,
      MyValidityErrorFunc,
      MyValidityWarningFunc
   };

   if (fDebugOutput) cerr << "DOM Parsing: " << pszXMLFile << endl;

   doc_d = xmlParseFile (pszXMLFile);

   if (fDebugOutput) cerr << "doc_d = 0x" << hex << (int)doc_d << dec << endl;

   if (doc_d)
   {
      char achFileName[512];

      xmlValidateDocument (&validator, doc_d);

      if (fDebugOutput) cerr << "Document validated." << endl;

      xmlNodePtr nodeRoot          = xmlDocGetRootElement (doc_d);
      char      *pszOmniLanguage   = findOmniLanguage (nodeRoot);
      char      *pszOmniDriverName = findOmniDriverName (nodeRoot);
      char      *pszOmniDeviceName = findOmniDeviceName (nodeRoot);

      if (!pszOmniLanguage)
      {
         cerr << "Error: Missing language." << endl;
         fErrorCondition_d = true;
      }
      if (!pszOmniDriverName)
      {
         cerr << "Error: Missing driver name." << endl;
         fErrorCondition_d = true;
      }
      if (!pszOmniDeviceName)
      {
         cerr << "Error: Missing device name." << endl;
         fErrorCondition_d = true;
      }

      if (fErrorCondition_d)
         goto done;

      if (  pszOmniLanguage
         && pszOmniDriverName
         && pszOmniDeviceName
         )
      {
         cout << "The Omni language is " << pszOmniLanguage << endl;
         cout << "The Omni driver name is " << pszOmniDriverName << endl;
         cout << "The Omni device name is " << pszOmniDeviceName << endl;
      }

      sprintf (achFileName, "%s.xml", pszOmniDeviceName);

      ofstream ofstreamDevice (achFileName);

      outputHeader (&ofstreamDevice, "Device");

      ofstreamDevice << "<Device name=\""
                     << pszOmniDeviceName
                     << "\">"
                     << endl;

      ofstreamDevice << "  <DriverName>"
                     << pszOmniDriverName
                     << "</DriverName>"
                     << endl;

      ofstreamDevice << "  <Capability type=\"MONOCHROME\"/>"
                     << endl;

      ofstreamDevice << "  <RasterCapabilities type=\"TOP_TO_BOTTOM\"/>"
                     << endl;

      ofstreamDevice << "  <PDL level=\"PDL_PCL\" sublevel=\"LEVEL_"
                     << pszOmniLanguage
                     << "\" major=\"1\" minor=\"0\"/>"
                     << endl;

      sprintf (achFileName, "%s Commands.xml", pszOmniDeviceName);

      ofstream ofstreamCommands (achFileName);

      outputHeader (&ofstreamCommands, "deviceCommands");

      ofstreamCommands << "<deviceCommands>"
                       << endl;
      ofstreamCommands << "<command name=\"cmdNull\">_NUL_</command>"
                       << endl;
      ofstreamCommands << "</deviceCommands>"
                       << endl;

      ofstreamDevice << "  <Uses>"
                     << achFileName
                     << "</Uses>"
                     << endl;

      sprintf (achFileName, "%s Forms.xml", pszOmniDeviceName);

      ofstream ofstreamForms (achFileName);

      processForms  (nodeRoot, &ofstreamForms);

      ofstreamDevice << "  <Has>"
                     << achFileName
                     << "</Has>"
                     << endl;

      sprintf (achFileName, "%s Medias.xml", pszOmniDeviceName);

      ofstream ofstreamMedias (achFileName);

      processMedias (nodeRoot, &ofstreamMedias);

      ofstreamDevice << "  <Has>"
                     << achFileName
                     << "</Has>"
                     << endl;

      sprintf (achFileName, "%s Trays.xml", pszOmniDeviceName);

      ofstream ofstreamTrays (achFileName);

      processTrays (nodeRoot, &ofstreamTrays);

      ofstreamDevice << "  <Has>"
                     << achFileName
                     << "</Has>"
                     << endl;

      sprintf (achFileName, "%s Resolutions.xml", pszOmniDeviceName);

      ofstream ofstreamResolutions (achFileName);

      processResolutions (nodeRoot, &ofstreamResolutions);

      ofstreamDevice << "  <Has>"
                     << achFileName
                     << "</Has>"
                     << endl;

      sprintf (achFileName, "%s Orientations.xml", pszOmniDeviceName);

      ofstream ofstreamOrientations (achFileName);

      processOrientations (nodeRoot, &ofstreamOrientations);

      ofstreamDevice << "  <Has>"
                     << achFileName
                     << "</Has>"
                     << endl;

      sprintf (achFileName, "%s Connections.xml", pszOmniDeviceName);

      ofstream ofstreamConnections (achFileName);

      processConnections (nodeRoot, &ofstreamConnections);

      ofstreamDevice << "  <Has>"
                     << achFileName
                     << "</Has>"
                     << endl;

      sprintf (achFileName, "%s Data.xml", pszOmniDeviceName);

      ofstream ofstreamData (achFileName);

      processData (nodeRoot, &ofstreamData);

      ofstreamDevice << "  <Has>"
                     << achFileName
                     << "</Has>"
                     << endl;

      sprintf (achFileName, "%s PrintModes.xml", pszOmniDeviceName);

      ofstream ofstreamPrintModes (achFileName);

      ofstreamPrintModes << "<devicePrintModes>" << endl
                         << "  <devicePrintMode>" << endl
                         << "    <name>PRINT_MODE_1_ANY</name>" << endl
                         << "    <printModePhysicalCount>1</printModePhysicalCount>" << endl
                         << "    <printModeLogicalCount>1</printModeLogicalCount>" << endl
                         << "    <printModePlanes>1</printModePlanes>" << endl
                         << "  </devicePrintMode>" << endl
                         << "</devicePrintModes>" << endl;

      ofstreamDevice << "  <Has>"
                     << achFileName
                     << "</Has>"
                     << endl;

      ofstreamDevice << "  <Instance>UPDF Instance.hpp</Instance>" << endl
                     << "  <Instance>UPDF Instance.cpp</Instance>" << endl
                     << "  <Blitter>UPDF Blitter.hpp</Blitter>" << endl
                     << "  <Blitter>UPDF Blitter.cpp</Blitter>" << endl;

      processLocale (nodeLocale_d,
                     &ofstreamDevice);

      ofstreamDevice << "</Device>" << endl;

      if (pszOmniDriverName)
      {
         free (pszOmniDriverName);
      }
      if (pszOmniDeviceName)
      {
         free (pszOmniDeviceName);
      }
   }
   else
   {
      cerr << "Error: File not found \"" << pszXMLFile << "\"" << endl;

      fErrorCondition_d = true;
   }

done:
   if (!fErrorCondition_d)
   {
      return 1;
   }
   else
   {
      return 0;
   }
}
